<?php
/**************************************************
  Coppermine 1.5.x Plugin - gallery_merge
**************************************************/

/* Configuration variables
      'plugin_gallery_merge_config_link'
      'plugin_gallery_merge_hide_buttons'
      'plugin_gallery_merge_source_local'
      'plugin_gallery_merge_source_filepath'
      'plugin_gallery_merge_source_dbserver'
      'plugin_gallery_merge_source_dbuser'
      'plugin_gallery_merge_source_dbpass'
      'plugin_gallery_merge_source_dbname'
      'plugin_gallery_merge_source_tableprefix'
      'plugin_gallery_merge_remove_subfolder'
      'plugin_gallery_merge_add_subfolder'
      'plugin_gallery_merge_merge_categories'
      'plugin_gallery_merge_merge_albums'
      'plugin_gallery_merge_folder_permissions'
      'plugin_gallery_merge_folder_perm_value'
      'plugin_gallery_merge_file_permissions'
      'plugin_gallery_merge_file_perm_value'
      'plugin_gallery_merge_copy_languages'
      'plugin_gallery_merge_copy_plugins'
      'plugin_gallery_merge_copy_themes'
      'plugin_gallery_merge_replace_files'
      'plugin_gallery_merge_checkpoint_freq'
      'plugin_gallery_merge_checkpoint_quiesce'
      'plugin_gallery_merge_chkpt_quiesce_files'
      'plugin_gallery_merge_message_level'
      'plugin_gallery_merge_debug'
      'plugin_gallery_merge_pictures_process'
*/


// create Inspekt supercage
$superCage = Inspekt::makeSuperCage();

// Make sure that this file can't be accessed directly, but only from within the Coppermine user interface
if (!defined('IN_COPPERMINE')) {
    die('Not in Coppermine...');
}

if (!GALLERY_ADMIN_MODE) {
    cpg_die(ERROR, $lang_errors['access_denied'], __FILE__, __LINE__);
}

include_once('functions.php');                                                  // include plugin subfunctions
// get language files
require "./plugins/gallery_merge/lang/english.php";
if ($CONFIG['lang'] != 'english' && file_exists("./plugins/gallery_merge/lang/{$CONFIG['lang']}.php")) {
    require "./plugins/gallery_merge/lang/{$CONFIG['lang']}.php";
}

// get sanitized POST parameters
if (($superCage->post->keyExists('submit')) && ($superCage->post->getRaw('submit') == $lang_plugin_gallery_merge['submit'])) {
    //Check if the form token is valid
    if(!checkFormToken()){
        cpg_die(ERROR, $lang_errors['invalid_form_token'], __FILE__, __LINE__);
    }

  // Define the sanitization patterns
  $sanitization_array = array(
      'plugin_gallery_merge_config_link' => array('type' => 'checkbox', 'min' => '0', 'max' => '1'),                           // checkbox - yes|no
      'plugin_gallery_merge_hide_buttons' => array('type' => 'checkbox', 'min' => '0', 'max' => '1'),                          // checkbox - yes|no
      'plugin_gallery_merge_source_local' => array('type' => 'checkbox', 'min' => '0', 'max' => '1'),                          // checkbox - yes|no
      'plugin_gallery_merge_source_filepath' => array('type' => 'raw', 'regex' => '/([^n]*+[^V\.]++)*+/'),                     // text - A-Z a-z - _ . /
      'plugin_gallery_merge_source_dbserver' => array('type' => 'raw', 'regex' => '/([^n]*+[^V\.]++)*+/'),                     // text - A-Z a-z - _ : . /
      'plugin_gallery_merge_source_dbuser' => array('type' => 'raw', 'regex' => '/([^n]*+[^V\.]++)*+/'),                       // text - A-Z a-z - _ . /
      'plugin_gallery_merge_source_dbpass' => array('type' => 'raw', 'regex' => '/([^n]*+[^V\.]++)*+/'),                       // text - A-Z a-z - _ . /
      'plugin_gallery_merge_source_dbname' => array('type' => 'raw', 'regex' => '/([^n]*+[^V\.]++)*+/'),                       // text - A-Z a-z - _ . /
      'plugin_gallery_merge_source_table_prefix' => array('type' => 'raw', 'regex' => '/([^n]*+[^V\.]++)*+/'),                 // text - A-Z a-z - _ . /
      'plugin_gallery_merge_remove_subfolder' => array('type' => 'raw', 'regex' => '/([^n]*+[^V\.]++)*+/'),                    // text - A-Z a-z - _ . /
      'plugin_gallery_merge_add_subfolder' => array('type' => 'raw', 'regex' => '/([^n]*+[^V\.]++)*+/'),                       // text - A-Z a-z - _ . /
      'plugin_gallery_merge_merge_categories' => array('type' => 'checkbox', 'min' => '0', 'max' => '1'),                      // checkbox - yes|no
      'plugin_gallery_merge_merge_albums' => array('type' => 'checkbox', 'min' => '0', 'max' => '1'),                          // checkbox - yes|no
      'plugin_gallery_merge_folder_permissions' => array('type' => 'checkbox', 'min' => '0', 'max' => '3'),                    // radio - 0|1|2
      'plugin_gallery_merge_folder_perm_value' => array('type' => 'raw', 'regex' => '/([^n]*+[^V\.]++)*+/'),                   // text - 0-9 repeating (NOT int)
      'plugin_gallery_merge_file_permissions' => array('type' => 'checkbox', 'min' => '0', 'max' => '3'),                      // radio - 0|1|2
      'plugin_gallery_merge_file_perm_value' => array('type' => 'raw', 'regex' => '/([^n]*+[^V\.]++)*+/'),                     // text - 0-9 repeating (NOT int)
      'plugin_gallery_merge_copy_languages' => array('type' => 'checkbox', 'min' => '0', 'max' => '1'),                        // checkbox - yes|no
      'plugin_gallery_merge_copy_plugins' => array('type' => 'checkbox', 'min' => '0', 'max' => '1'),                          // checkbox - yes|no
      'plugin_gallery_merge_copy_themes' => array('type' => 'checkbox', 'min' => '0', 'max' => '1'),                           // checkbox - yes|no
      'plugin_gallery_merge_replace_files' => array('type' => 'checkbox', 'min' => '0', 'max' => '1'),                         // checkbox - yes|no
      'plugin_gallery_merge_checkpoint_freq' => array('type' => 'int', 'min' => '250', 'max' => '5000'),                       // integer - 250 -> 5000
      'plugin_gallery_merge_checkpoint_quiesce' => array('type' => 'int', 'min' => '5000', 'max' => '20000'),                  // integer - 5000 -> 20000
      'plugin_gallery_merge_chkpt_quiesce_files' => array('type' => 'int', 'min' => '5000', 'max' => '50000'),                 // integer - 5000 -> 20000
      'plugin_gallery_merge_message_level' => array('type' => 'checkbox', 'min' => '1', 'max' => '2'),                         // radio - 1|2
      'plugin_gallery_merge_debug' => array('type' => 'checkbox', 'min' => '0', 'max' => '3'),                                 // radio - 0|1|2|3
      'plugin_gallery_merge_pictures_process' => array('type' => 'checkbox', 'min' => '0', 'max' => '1'),                      // checkbox - yes|no
  );
  $config_changes_counter = 0;
  $plugin_gallery_merge_execute = FALSE;
  $plugin_gallery_merge_execute_files = FALSE;
  foreach ($sanitization_array as $san_key => $san_value) {
      if (isset($CONFIG[$san_key]) == TRUE) { // only loop if config value is set --- start
          if ($san_value['type'] == 'checkbox') { // type is checkbox --- start
            if ($superCage->post->getInt($san_key) <= $san_value['max'] && $superCage->post->getInt($san_key) >= $san_value['min'] && $superCage->post->getInt($san_key) != $CONFIG[$san_key]) {   // **GMC
                $CONFIG[$san_key] = $superCage->post->getInt($san_key);
                cpg_db_query("UPDATE {$CONFIG['TABLE_CONFIG']} SET value='{$CONFIG[$san_key]}' WHERE name='$san_key'");
                $config_changes_counter++;
            } elseif($superCage->post->keyExists($san_key) != TRUE && $CONFIG[$san_key] != '0') {
                $CONFIG[$san_key] = 0;
                cpg_db_query("UPDATE {$CONFIG['TABLE_CONFIG']} SET value='{$CONFIG[$san_key]}' WHERE name='$san_key'");
                $config_changes_counter++;
            }
          } // type is checkbox --- end
          if ($san_value['type'] == 'int') { // type is integer --- start
              if ($superCage->post->getInt($san_key) <= $san_value['max'] && $superCage->post->getInt($san_key) >= $san_value['min'] && $superCage->post->getInt($san_key) != $CONFIG[$san_key]) {
                  $CONFIG[$san_key] = $superCage->post->getInt($san_key);
                  cpg_db_query("UPDATE {$CONFIG['TABLE_CONFIG']} SET value='{$CONFIG[$san_key]}' WHERE name='$san_key'");
                  $config_changes_counter++;
              }
          } // type is integer --- end
          if ($san_value['type'] == 'raw') { // type is raw --- start
              // added code to:
              // - retain current value if variable is not passed in form (some raw elements set to disabled - so not passed in form input..)
              // - trim whitespace from beginning/end of string
              // - convert plugin_gallery_merge_source_filepath to realpath if new...
              if($superCage->post->keyExists($san_key) == TRUE) {
                // only if variable was passed in form
                $trim_var = trim($superCage->post->getRaw($san_key));             // pull in var from superCage and trim whitespace
                if (isset($san_value['regex']) == TRUE && preg_match($san_value['regex'], $trim_var) && $trim_var != $CONFIG[$san_key]) {
                  if ($san_key == 'plugin_gallery_merge_source_filepath') {
                    $trim_var = realpath($trim_var);                            // translate new filepath to realpath (eliminate relative address ./ or ../)
                  }
                  $CONFIG[$san_key] = $trim_var;
                  if ($trim_var == 'none') {
                    $CONFIG[$san_key] = '';
                  }
                  cpg_db_query("UPDATE {$CONFIG['TABLE_CONFIG']} SET value='{$CONFIG[$san_key]}' WHERE name='$san_key'");
                  $config_changes_counter++;
                }
              }
          } // type is raw --- end
          if ($san_value['type'] == 'array') { // type is array --- start              
          $evaluate_value = $superCage->post->getRaw($san_key);
              if (is_array($evaluate_value) && isset($san_value['regex']) == TRUE && isset($san_value['delimiter']) == TRUE) {
                  $temp = '';
                  for ($i = 0; $i <= count($evaluate_value); $i++) {
                      if (preg_match($san_value['regex'], $evaluate_value[$i])) {
                          $temp .= $evaluate_value[$i] . $san_value['delimiter'];
                      }
                  }
                  unset($evaluate_value);
                  $evaluate_value = rtrim($temp, $san_value['delimiter']);
                  unset($temp);
              }
              if ($evaluate_value != $CONFIG[$san_key]) {
                  $CONFIG[$san_key] = $evaluate_value;
                  cpg_db_query("UPDATE {$CONFIG['TABLE_CONFIG']} SET value='{$CONFIG[$san_key]}' WHERE name='$san_key'");
                  $config_changes_counter++;
              }
          } // type is array --- end
      } // only loop if config value is set --- end
  }
}


// display config page

// Set the option output stuff
if ($CONFIG['plugin_gallery_merge_adminmode'] == '1') {
    $option_output['plugin_gallery_merge_adminmode'] = 'checked="checked"';
} else { 
    $option_output['plugin_gallery_merge_adminmode'] = '';
}

if ($CONFIG['plugin_gallery_merge_config_link'] == '1') {
    $option_output['plugin_gallery_merge_config_link'] = 'checked="checked"';
} else { 
    $option_output['plugin_gallery_merge_config_link'] = '';
}

if ($CONFIG['plugin_gallery_merge_hide_buttons'] == '1') {
    $option_output['plugin_gallery_merge_hide_buttons'] = 'checked="checked"';
} else { 
    $option_output['plugin_gallery_merge_hide_buttons'] = '';
}

if ($CONFIG['plugin_gallery_merge_source_local'] == '1') {
    $option_output['plugin_gallery_merge_source_local'] = 'checked="checked" onchange="gallery_merge_protect_fields();"';
} else { 
    $option_output['plugin_gallery_merge_source_local'] = 'onchange="gallery_merge_protect_fields();"';
}

if ($CONFIG['plugin_gallery_merge_merge_categories'] == '1') {
    $option_output['plugin_gallery_merge_merge_categories'] = 'checked="checked"';
} else { 
    $option_output['plugin_gallery_merge_merge_categories'] = '';
}

if ($CONFIG['plugin_gallery_merge_merge_albums'] == '1') {
    $option_output['plugin_gallery_merge_merge_albums'] = 'checked="checked"';
} else { 
    $option_output['plugin_gallery_merge_merge_albums'] = '';
}

if ($CONFIG['plugin_gallery_merge_folder_permissions'] == '0') {
    $option_output['plugin_gallery_merge_folder_permissions'][0] = 'checked="checked"';
    $option_output['plugin_gallery_merge_folder_permissions'][1] = '';
    $option_output['plugin_gallery_merge_folder_permissions'][2] = '';
    $option_output['plugin_gallery_merge_folder_permissions'][3] = '';
} elseif ($CONFIG['plugin_gallery_merge_folder_permissions'] == '1') {
    $option_output['plugin_gallery_merge_folder_permissions'][0] = '';
    $option_output['plugin_gallery_merge_folder_permissions'][1] = 'checked="checked"';
    $option_output['plugin_gallery_merge_folder_permissions'][2] = '';
    $option_output['plugin_gallery_merge_folder_permissions'][3] = '';
} elseif ($CONFIG['plugin_gallery_merge_folder_permissions'] == '2') {
    $option_output['plugin_gallery_merge_folder_permissions'][0] = '';
    $option_output['plugin_gallery_merge_folder_permissions'][1] = '';
    $option_output['plugin_gallery_merge_folder_permissions'][2] = 'checked="checked"';
    $option_output['plugin_gallery_merge_folder_permissions'][3] = '';
} elseif ($CONFIG['plugin_gallery_merge_folder_permissions'] == '3') {
    $option_output['plugin_gallery_merge_folder_permissions'][0] = '';
    $option_output['plugin_gallery_merge_folder_permissions'][1] = '';
    $option_output['plugin_gallery_merge_folder_permissions'][2] = '';
    $option_output['plugin_gallery_merge_folder_permissions'][3] = 'checked="checked"';
}

if ($CONFIG['plugin_gallery_merge_file_permissions'] == '0') {
    $option_output['plugin_gallery_merge_file_permissions'][0] = 'checked="checked"';
    $option_output['plugin_gallery_merge_file_permissions'][1] = '';
    $option_output['plugin_gallery_merge_file_permissions'][2] = '';
    $option_output['plugin_gallery_merge_file_permissions'][3] = '';
} elseif ($CONFIG['plugin_gallery_merge_file_permissions'] == '1') {
    $option_output['plugin_gallery_merge_file_permissions'][0] = '';
    $option_output['plugin_gallery_merge_file_permissions'][1] = 'checked="checked"';
    $option_output['plugin_gallery_merge_file_permissions'][2] = '';
    $option_output['plugin_gallery_merge_file_permissions'][3] = '';
} elseif ($CONFIG['plugin_gallery_merge_file_permissions'] == '2') {
    $option_output['plugin_gallery_merge_file_permissions'][0] = '';
    $option_output['plugin_gallery_merge_file_permissions'][1] = '';
    $option_output['plugin_gallery_merge_file_permissions'][2] = 'checked="checked"';
    $option_output['plugin_gallery_merge_file_permissions'][3] = '';
} elseif ($CONFIG['plugin_gallery_merge_file_permissions'] == '2') {
    $option_output['plugin_gallery_merge_file_permissions'][0] = '';
    $option_output['plugin_gallery_merge_file_permissions'][1] = '';
    $option_output['plugin_gallery_merge_file_permissions'][2] = '';
    $option_output['plugin_gallery_merge_file_permissions'][3] = 'checked="checked"';
}

if ($CONFIG['plugin_gallery_merge_copy_languages'] == '1') {
    $option_output['plugin_gallery_merge_copy_languages'] = 'checked="checked"';
} else { 
    $option_output['plugin_gallery_merge_copy_languages'] = '';
}

if ($CONFIG['plugin_gallery_merge_copy_plugins'] == '1') {
    $option_output['plugin_gallery_merge_copy_plugins'] = 'checked="checked"';
} else { 
    $option_output['plugin_gallery_merge_copy_plugins'] = '';
}

if ($CONFIG['plugin_gallery_merge_copy_themes'] == '1') {
    $option_output['plugin_gallery_merge_copy_themes'] = 'checked="checked"';
} else { 
    $option_output['plugin_gallery_merge_copy_themes'] = '';
}

if ($CONFIG['plugin_gallery_merge_replace_files'] == '1') {
    $option_output['plugin_gallery_merge_replace_files'] = 'checked="checked"';
} else { 
    $option_output['plugin_gallery_merge_replace_files'] = '';
}

if ($CONFIG['plugin_gallery_merge_message_level'] == '0') {
    $option_output['plugin_gallery_merge_message_level'][0] = 'checked="checked"';
    $option_output['plugin_gallery_merge_message_level'][1] = '';
    $option_output['plugin_gallery_merge_message_level'][2] = '';
} elseif ($CONFIG['plugin_gallery_merge_message_level'] == '1') {
    $option_output['plugin_gallery_merge_message_level'][0] = '';
    $option_output['plugin_gallery_merge_message_level'][1] = 'checked="checked"';
    $option_output['plugin_gallery_merge_message_level'][2] = '';
} elseif ($CONFIG['plugin_gallery_merge_message_level'] == '2') {
    $option_output['plugin_gallery_merge_message_level'][0] = '';
    $option_output['plugin_gallery_merge_message_level'][1] = '';
    $option_output['plugin_gallery_merge_message_level'][2] = 'checked="checked"';
}

if ($CONFIG['plugin_gallery_merge_debug'] == '0') {
    $option_output['plugin_gallery_merge_debug'][0] = 'checked="checked"';
    $option_output['plugin_gallery_merge_debug'][1] = '';
    $option_output['plugin_gallery_merge_debug'][2] = '';
    $option_output['plugin_gallery_merge_debug'][3] = '';
} elseif ($CONFIG['plugin_gallery_merge_debug'] == '1') {
    $option_output['plugin_gallery_merge_debug'][0] = '';
    $option_output['plugin_gallery_merge_debug'][1] = 'checked="checked"';
    $option_output['plugin_gallery_merge_debug'][2] = '';
    $option_output['plugin_gallery_merge_debug'][3] = '';
} elseif ($CONFIG['plugin_gallery_merge_debug'] == '2') {
    $option_output['plugin_gallery_merge_debug'][0] = '';
    $option_output['plugin_gallery_merge_debug'][1] = '';
    $option_output['plugin_gallery_merge_debug'][2] = 'checked="checked"';
    $option_output['plugin_gallery_merge_debug'][3] = '';
} elseif ($CONFIG['plugin_gallery_merge_debug'] == '3') {
    $option_output['plugin_gallery_merge_debug'][0] = '';
    $option_output['plugin_gallery_merge_debug'][1] = '';
    $option_output['plugin_gallery_merge_debug'][2] = '';
    $option_output['plugin_gallery_merge_debug'][3] = 'checked="checked"';
}

if ($CONFIG['plugin_gallery_merge_pictures_process'] == '1') {
    $option_output['plugin_gallery_merge_pictures_process'] = 'checked="checked"';
} else { 
    $option_output['plugin_gallery_merge_pictures_process'] = '';
}

// following option always displayed as 0 - value not stored in DB - used as a trigger
//$option_output['plugin_gallery_merge_execute'] = '';

pageheader(sprintf("{$lang_plugin_gallery_merge['configure_x']} V{$CONFIG['plugin_gallery_merge_version']}", $lang_plugin_gallery_merge['plugin_name']));
list($timestamp, $form_token) = getFormToken();
echo <<< EOT
<form action="index.php?file=gallery_merge/admin" method="post" name="gallery_merge_settings">
EOT;
starttable('100%', sprintf("{$lang_plugin_gallery_merge['configure_x']} V{$CONFIG['plugin_gallery_merge_version']}", $lang_plugin_gallery_merge['plugin_name']), 3, 'cpg_zebra');

// add condition to be sure the submit button was for config updates before displaying messagebox
if ($superCage->post->keyExists('submit') && $superCage->post->getRaw('submit') == $lang_plugin_gallery_merge['submit']) {
    if ($config_changes_counter > 0) {
        gallery_merge_checkpoint('config', 'timestamp', 'update');              // note our config update timestamp
        echo <<< EOT
    <tr>
        <td class="tablef" colspan="2" >
EOT;
        msg_box('', $lang_plugin_gallery_merge['update_success'], '', '', 'success');
    } else {
        msg_box('', $lang_plugin_gallery_merge['no_changes'], '', '', 'validation');
    }
        echo <<< EOT
        </td>
    </tr>
EOT;
}

// setup of variable data needed below
$CONFIG['TABLE_RESTART'] = $CONFIG['TABLE_PREFIX'].'plugin_gallery_merge_restart_data';  // add our table name for easy access

//  Only add in this code if we have multiple sections to show/hide...
/*
echo <<< EOT
    <tr>
      <td class="tableh2" colspan="2">
        <button type="button" class="button detail_expand_all" name="expandalltop" id="expandalltop" onclick="return false">
          <img src="images/tree/plus.gif" width="9" height="9" border="0" alt="" title="Expand All" />
          Expand All
        </button>
        <button type="button" class="button detail_collapse_all" name="collapsealltop" id="collapsealltop" style="display:none;" onclick="return false">
          <img src="images/tree/minus.gif" width="9" height="9" border="0" alt="" title="Collapse all" />
          Collapse all
        </button>
        <button type="button" class="button detail_toggle_all" name="togglealltop" id="togglealltop" onclick="return false">
          <img src="images/tree/toggle.gif" width="9" height="9" border="0" alt="" title="Toggle All" />
          Toggle All
        </button>
      </td>
    </tr>
EOT;
*/

echo <<< EOT
  <!-- insert config option form code start -->

    <tr>                                                                                                                               
      <td class="tableb" colspan="2">
        <span class="detail_head_collapsed">
          Show / Hide Plugin Settings
        </span>
        <div id="section0" class="detail_body">
          <table align="center" width="100%" cellspacing="1" cellpadding="0" class="maintable" border="0">
            <tr>
              <td>
                &nbsp;
              </td>                                                                                                                            
            </tr>                                                                                                                              

            <tr>
              <td>
                <label for="plugin_gallery_merge_config_link" class="clickable_option">{$lang_plugin_gallery_merge['config_link']}</label>
              </td>
              <td>
                <input type="checkbox" name="plugin_gallery_merge_config_link" id="plugin_gallery_merge_config_link" class="checkbox" value="1" {$option_output['plugin_gallery_merge_config_link']} />
                <label for="plugin_gallery_merge_config_link" class="clickable_option">{$lang_common['yes']}</label>
              </td>
            </tr>

            <tr>
              <td>
                <label for="plugin_gallery_merge_hide_buttons" class="clickable_option">{$lang_plugin_gallery_merge['hide_buttons']}</label>
              </td>
              <td>
                <input type="checkbox" name="plugin_gallery_merge_hide_buttons" id="plugin_gallery_merge_hide_buttons" class="checkbox" value="1" {$option_output['plugin_gallery_merge_hide_buttons']} />
                <label for="plugin_gallery_merge_hide_buttons" class="clickable_option">{$lang_common['yes']}</label> &nbsp;&nbsp;&nbsp; ({$lang_plugin_gallery_merge['hide_buttons_comment']})
              </td>
            </tr>

            <tr>
              <td>
                <label for="plugin_gallery_merge_source_local" class="clickable_option">{$lang_plugin_gallery_merge['source_local']}</label>
              </td>
              <td>
                <input type="checkbox" name="plugin_gallery_merge_source_local" id="plugin_gallery_merge_source_local" class="checkbox" value="1" {$option_output['plugin_gallery_merge_source_local']} />
                <label for="plugin_gallery_merge_source_local" class="clickable_option">{$lang_common['yes']}</label> &nbsp;&nbsp;&nbsp; ({$lang_plugin_gallery_merge['source_local_comment']})
              </td>
            </tr>

            <tr>
              <td valign="top">
                &nbsp;
              </td>
              <td>
                {$lang_plugin_gallery_merge['source_filepath_comment']}
              </td>
            </tr>

            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['source_filepath']}
              </td>
              <td>
                <input type="text" name="plugin_gallery_merge_source_filepath" id="plugin_gallery_merge_source_filepath" class="textinput" size="75" maxlength="120" value="{$CONFIG['plugin_gallery_merge_source_filepath']}" {$option_output['plugin_gallery_merge_source_filepath']} /> {$lang_plugin_gallery_merge['source_filepath_comment2']}
              </td>
            </tr>

            <tr>
              <td valign="top">
                &nbsp;
              </td>
              <td>
                {$lang_plugin_gallery_merge['source_dbserver_comment']}
              </td>
            </tr>

            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['source_dbserver']}
              </td>
              <td>
                <input type="text" name="plugin_gallery_merge_source_dbserver" id="plugin_gallery_merge_source_dbserver" class="textinput" size="40" maxlength="60" value="{$CONFIG['plugin_gallery_merge_source_dbserver']}" {$option_output['plugin_gallery_merge_source_dbserver']} />
              </td>
            </tr>

            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['source_dbuser']}
              </td>
              <td>
                <input type="text" name="plugin_gallery_merge_source_dbuser" id="plugin_gallery_merge_source_dbuser" class="textinput" size="40" maxlength="60" value="{$CONFIG['plugin_gallery_merge_source_dbuser']}" {$option_output['plugin_gallery_merge_source_dbuser']} />
              </td>
            </tr>

            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['source_dbpass']}
              </td>
              <td>
                <input type="password" name="plugin_gallery_merge_source_dbpass" id="plugin_gallery_merge_source_dbpass" class="textinput" size="40" maxlength="60" value="{$CONFIG['plugin_gallery_merge_source_dbpass']}" {$option_output['plugin_gallery_merge_source_dbpass']} />
              </td>
            </tr>

            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['source_dbname']}
              </td>
              <td>
                <input type="text" name="plugin_gallery_merge_source_dbname" id="plugin_gallery_merge_source_dbname" class="textinput" size="40" maxlength="60" value="{$CONFIG['plugin_gallery_merge_source_dbname']}" {$option_output['plugin_gallery_merge_source_dbname']} />
              </td>
            </tr>

            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['source_tableprefix']}
              </td>
              <td>
                <input type="text" name="plugin_gallery_merge_source_table_prefix" id="plugin_gallery_merge_source_table_prefix" class="textinput" size="40" maxlength="60" value="{$CONFIG['plugin_gallery_merge_source_table_prefix']}" {$option_output['plugin_gallery_merge_source_table_prefix']} />
              </td>
            </tr>

            <tr>
              <td valign="top">
                &nbsp;
              </td>
              <td>
                {$lang_plugin_gallery_merge['subfolder_comment']}
              </td>
            </tr>

            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['remove_subfolder']}
              </td>
              <td>
                <input type="text" name="plugin_gallery_merge_remove_subfolder" id="plugin_gallery_merge_remove_subfolder" class="textinput" size="40" maxlength="60" value="{$CONFIG['plugin_gallery_merge_remove_subfolder']}" {$option_output['plugin_gallery_merge_remove_subfolder']} />
                {$lang_plugin_gallery_merge['subfolder_comment2']}
              </td>
            </tr>

            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['add_subfolder']}
              </td>
              <td>
                <input type="text" name="plugin_gallery_merge_add_subfolder" id="plugin_gallery_merge_add_subfolder" class="textinput" size="40" maxlength="60" value="{$CONFIG['plugin_gallery_merge_add_subfolder']}" {$option_output['plugin_gallery_merge_add_subfolder']} />
                {$lang_plugin_gallery_merge['subfolder_comment2']}
              </td>
            </tr>

            <tr>
              <td valign="top">
                &nbsp;
              </td>
              <td>
                {$lang_plugin_gallery_merge['merge_cat_alb_comment']}
              </td>
            </tr>

            <tr>
              <td>
                <label for="plugin_gallery_merge_merge_categories" class="clickable_option">{$lang_plugin_gallery_merge['merge_categories']}</label>
              </td>
              <td>
                <input type="checkbox" name="plugin_gallery_merge_merge_categories" id="plugin_gallery_merge_merge_categories" class="checkbox" value="1" {$option_output['plugin_gallery_merge_merge_categories']} />
                <label for="plugin_gallery_merge_merge_categories" class="clickable_option">{$lang_common['yes']}</label> &nbsp;&nbsp;&nbsp; ({$lang_plugin_gallery_merge['merge_categories_comment']})
              </td>
            </tr>

            <tr>
              <td>
                <label for="plugin_gallery_merge_merge_albums" class="clickable_option">{$lang_plugin_gallery_merge['merge_albums']}</label>
              </td>
              <td>
                <input type="checkbox" name="plugin_gallery_merge_merge_albums" id="plugin_gallery_merge_merge_albums" class="checkbox" value="1" {$option_output['plugin_gallery_merge_merge_albums']} />
                <label for="plugin_gallery_merge_merge_albums" class="clickable_option">{$lang_common['yes']}</label>  &nbsp;&nbsp;&nbsp; ({$lang_plugin_gallery_merge['merge_albums_comment']})
              </td>
            </tr>

            <tr>
              <td valign="top">
                &nbsp;
              </td>
              <td>
                {$lang_plugin_gallery_merge['file_folder_permissions_comment']}
              </td>
            </tr>

            <tr>
              <td style="vertical-align:top;">
                {$lang_plugin_gallery_merge['folder_permissions']}
              </td>
              <td>
                <input type="radio" name="plugin_gallery_merge_folder_permissions" id="plugin_gallery_merge_folder_permissions_0" class="radio" value="0" {$option_output['plugin_gallery_merge_folder_permissions'][0]} />
                <label for="plugin_gallery_merge_folder_permissions_0" class="clickable_option">{$lang_plugin_gallery_merge['default']} ({$CONFIG['default_dir_mode']})</label>
                <input type="radio" name="plugin_gallery_merge_folder_permissions" id="plugin_gallery_merge_file_permissions_1" class="radio" value="1" {$option_output['plugin_gallery_merge_folder_permissions'][1]} />
                <label for="plugin_gallery_merge_folder_permissions_1" class="clickable_option">{$lang_plugin_gallery_merge['obtain_source']}</label>
                <input type="radio" name="plugin_gallery_merge_folder_permissions" id="plugin_gallery_merge_file_permissions_2" class="radio" value="2" {$option_output['plugin_gallery_merge_folder_permissions'][2]} />
                <label for="plugin_gallery_merge_folder_permissions_2" class="clickable_option">{$lang_plugin_gallery_merge['use_specified']}</label>&nbsp;&nbsp;&nbsp;&nbsp;
                <input type="text" name="plugin_gallery_merge_folder_perm_value" id="plugin_gallery_merge_folder_perm_value" class="textinput" size="4" maxlength="4" value="{$CONFIG['plugin_gallery_merge_folder_perm_value']}" {$option_output['plugin_gallery_merge_folder_perm_value']} />
                <input type="radio" name="plugin_gallery_merge_folder_permissions" id="plugin_gallery_merge_file_permissions_3" class="radio" value="3" {$option_output['plugin_gallery_merge_folder_permissions'][3]} />
                <label for="plugin_gallery_merge_folder_permissions_3" class="clickable_option">{$lang_plugin_gallery_merge['none']}</label>
                ({$lang_plugin_gallery_merge['folder_permissions_comment']})
              </td>
            </tr>

            <tr>
              <td style="vertical-align:top;">
                {$lang_plugin_gallery_merge['file_permissions']}
              </td>
              <td>
                <input type="radio" name="plugin_gallery_merge_file_permissions" id="plugin_gallery_merge_file_permissions_0" class="radio" value="0" {$option_output['plugin_gallery_merge_file_permissions'][0]} />
                <label for="plugin_gallery_merge_file_permissions_0" class="clickable_option">{$lang_plugin_gallery_merge['default']}  ({$CONFIG['default_file_mode']})</label>
                <input type="radio" name="plugin_gallery_merge_file_permissions" id="plugin_gallery_merge_file_permissions_1" class="radio" value="1" {$option_output['plugin_gallery_merge_file_permissions'][1]} />
                <label for="plugin_gallery_merge_file_permissions_1" class="clickable_option">{$lang_plugin_gallery_merge['obtain_source']}</label>
                <input type="radio" name="plugin_gallery_merge_file_permissions" id="plugin_gallery_merge_file_permissions_2" class="radio" value="2" {$option_output['plugin_gallery_merge_file_permissions'][2]} />
                <label for="plugin_gallery_merge_file_permissions_2" class="clickable_option">{$lang_plugin_gallery_merge['use_specified']}</label>&nbsp;&nbsp;&nbsp;&nbsp;
                <input type="text" name="plugin_gallery_merge_file_perm_value" id="plugin_gallery_merge_file_perm_value" class="textinput" size="4" maxlength="4" value="{$CONFIG['plugin_gallery_merge_file_perm_value']}" {$option_output['plugin_gallery_merge_file_perm_value']} />
                <input type="radio" name="plugin_gallery_merge_file_permissions" id="plugin_gallery_merge_file_permissions_3" class="radio" value="3" {$option_output['plugin_gallery_merge_file_permissions'][3]} />
                <label for="plugin_gallery_merge_file_permissions_3" class="clickable_option">{$lang_plugin_gallery_merge['none']}</label>
                ({$lang_plugin_gallery_merge['file_permissions_comment']})
              </td>
            </tr>

            <tr>
              <td valign="top">
                &nbsp;
              </td>
              <td>
                {$lang_plugin_gallery_merge['copy_replace_comment']}
              </td>
            </tr>

            <tr>
              <td>
                <label for="plugin_gallery_merge_copy_languages" class="clickable_option">{$lang_plugin_gallery_merge['copy_languages']}</label>
              </td>
              <td>
                <input type="checkbox" name="plugin_gallery_merge_copy_languages" id="plugin_gallery_merge_copy_languages" class="checkbox" value="1" {$option_output['plugin_gallery_merge_copy_languages']} />
                <label for="plugin_gallery_merge_copy_languages" class="clickable_option">{$lang_common['yes']}</label> &nbsp;&nbsp;&nbsp; ({$lang_plugin_gallery_merge['copy_languages_comment']})
              </td>
            </tr>

            <tr>
              <td>
                <label for="plugin_gallery_merge_copy_plugins" class="clickable_option">{$lang_plugin_gallery_merge['copy_plugins']}</label>
              </td>
              <td>
                <input type="checkbox" name="plugin_gallery_merge_copy_plugins" id="plugin_gallery_merge_copy_plugins" class="checkbox" value="1" {$option_output['plugin_gallery_merge_copy_plugins']} />
                <label for="plugin_gallery_merge_copy_plugins" class="clickable_option">{$lang_common['yes']}</label> &nbsp;&nbsp;&nbsp; ({$lang_plugin_gallery_merge['copy_folders_comment']})
              </td>
            </tr>

            <tr>
              <td>
                <label for="plugin_gallery_merge_copy_themes" class="clickable_option">{$lang_plugin_gallery_merge['copy_themes']}</label>
              </td>
              <td>
                <input type="checkbox" name="plugin_gallery_merge_copy_themes" id="plugin_gallery_merge_copy_themes" class="checkbox" value="1" {$option_output['plugin_gallery_merge_copy_themes']} />
                <label for="plugin_gallery_merge_copy_themes" class="clickable_option">{$lang_common['yes']}</label> &nbsp;&nbsp;&nbsp; ({$lang_plugin_gallery_merge['copy_folders_comment']})
              </td>
            </tr>

            <tr>
              <td>
                <label for="plugin_gallery_merge_replace_files" class="clickable_option">{$lang_plugin_gallery_merge['replace_files']}</label>
              </td>
              <td>
                <input type="checkbox" name="plugin_gallery_merge_replace_files" id="plugin_gallery_merge_replace_files" class="checkbox" value="1" {$option_output['plugin_gallery_merge_replace_files']} />
                <label for="plugin_gallery_merge_replace_files" class="clickable_option">{$lang_common['yes']}</label> &nbsp;&nbsp;&nbsp; ({$lang_plugin_gallery_merge['replace_files_comment']})
              </td>
            </tr>

            <tr>
              <td valign="top">
                &nbsp;
              </td>
              <td>
                {$lang_plugin_gallery_merge['checkpoint_comment']}
              </td>
            </tr>

            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['checkpoint_freq']}
              </td>
              <td>
                <input type="text" name="plugin_gallery_merge_checkpoint_freq" id="plugin_gallery_merge_checkpoint_freq" class="textinput" size="4" maxlength="4" value="{$CONFIG['plugin_gallery_merge_checkpoint_freq']}" {$option_output['plugin_gallery_merge_checkpoint_freq']} />
                {$lang_plugin_gallery_merge['inserts']} (250 thru 5000) {$lang_plugin_gallery_merge['checkpoint_freq_comment']}
              </td>
            </tr>
           
            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['checkpoint_quiesce']}
              </td>
              <td>
                <input type="text" name="plugin_gallery_merge_checkpoint_quiesce" id="plugin_gallery_merge_checkpoint_quiesce" class="textinput" size="5" maxlength="5" value="{$CONFIG['plugin_gallery_merge_checkpoint_quiesce']}" {$option_output['plugin_gallery_merge_checkpoint_quiesce']} />
                {$lang_plugin_gallery_merge['inserts']} (5000 thru 20000) {$lang_plugin_gallery_merge['checkpoint_quiesce_comment']}
              </td>
            </tr>

            <tr>
              <td valign="top">
                {$lang_plugin_gallery_merge['checkpoint_quiesce_files']}
              </td>
              <td>
                <input type="text" name="plugin_gallery_merge_chkpt_quiesce_files" id="plugin_gallery_merge_chkpt_quiesce_files" class="textinput" size="5" maxlength="5" value="{$CONFIG['plugin_gallery_merge_chkpt_quiesce_files']}" {$option_output['plugin_gallery_merge_chkpt_quiesce_files']} />
                {$lang_plugin_gallery_merge['copies']} (5000 thru 50000) {$lang_plugin_gallery_merge['checkpoint_quiesce_files_comment']}
              </td>
            </tr>

            <tr>
              <td valign="top">
                &nbsp;
              </td>
              <td>
                &nbsp;
              </td>
            </tr>

            <tr>
              <td style="vertical-align:top;">
                {$lang_plugin_gallery_merge['message_level']}
              </td>
              <td>
                <input type="radio" name="plugin_gallery_merge_message_level" id="plugin_gallery_merge_message_level_1" class="radio" value="1" {$option_output['plugin_gallery_merge_message_level'][1]} />
                <label for="plugin_gallery_merge_message_level_1" class="clickable_option">{$lang_plugin_gallery_merge['summary']}</label>
                <input type="radio" name="plugin_gallery_merge_message_level" id="plugin_gallery_merge_message_level_2" class="radio" value="2" {$option_output['plugin_gallery_merge_message_level'][2]} />
                <label for="plugin_gallery_merge_message_level_2" class="clickable_option">{$lang_plugin_gallery_merge['detail']}</label>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
                ({$lang_plugin_gallery_merge['message_level_comment']})
              </td>
            </tr>

            <tr>
              <td style="vertical-align:top;">
                {$lang_plugin_gallery_merge['debug']}
              </td>
              <td>
                <input type="radio" name="plugin_gallery_merge_debug" id="plugin_gallery_merge_debug_0" class="radio" value="0" {$option_output['plugin_gallery_merge_debug'][0]} />
                <label for="plugin_gallery_merge_debug_0" class="clickable_option">{$lang_plugin_gallery_merge['none']}</label>
                <input type="radio" name="plugin_gallery_merge_debug" id="plugin_gallery_merge_debug_1" class="radio" value="1" {$option_output['plugin_gallery_merge_debug'][1]} />
                <label for="plugin_gallery_merge_debug_1" class="clickable_option">{$lang_plugin_gallery_merge['summary']}</label>
                <input type="radio" name="plugin_gallery_merge_debug" id="plugin_gallery_merge_debug_2" class="radio" value="2" {$option_output['plugin_gallery_merge_debug'][2]} />
                <label for="plugin_gallery_merge_debug_2" class="clickable_option">{$lang_plugin_gallery_merge['sql']}</label>&nbsp;&nbsp;&nbsp;&nbsp;
                <input type="radio" name="plugin_gallery_merge_debug" id="plugin_gallery_merge_debug_3" class="radio" value="3" {$option_output['plugin_gallery_merge_debug'][3]} />
                <label for="plugin_gallery_merge_debug_3" class="clickable_option">{$lang_plugin_gallery_merge['detail']}</label>&nbsp;&nbsp;&nbsp;&nbsp;
                ({$lang_plugin_gallery_merge['debug_comment']})
              </td>
            </tr>
            <tr>
              <td>
                &nbsp;
              </td>                                                                                                                            
              <td>
                {$lang_plugin_gallery_merge['memory_comment']}
              </td>
            </tr>                                                                                                                              
            <tr>
              <td>
                <label for="plugin_gallery_merge_pictures_process" class="clickable_option">{$lang_plugin_gallery_merge['pictures_process']}</label>
              </td>
              <td>
                <input type="checkbox" name="plugin_gallery_merge_pictures_process" id="plugin_gallery_merge_pictures_process" class="checkbox" value="1" {$option_output['plugin_gallery_merge_pictures_process']} />
                <label for="plugin_gallery_merge_pictures_process" class="clickable_option">{$lang_common['yes']}</label> &nbsp;&nbsp;&nbsp; ({$lang_plugin_gallery_merge['pictures_process_comment']})
              </td>
            </tr>

          </table>
        </div>
      </td>
    </tr>

    <!-- insert config option form code end -->
    <tr>
        <td class="tablef" colspan="3">
            <input type="hidden" name="form_token" value="{$form_token}" />
            <input type="hidden" name="timestamp" value="{$timestamp}" />
            <button type="submit" class="button" name="submit" value="{$lang_plugin_gallery_merge['submit']}">{$lang_plugin_gallery_merge['submit']}</button>
        </td>
    </tr>
EOT;

endtable();
// leave form open into merge code...
//echo <<< EOT
//</form>
//EOT;
// include merge code
require "./plugins/gallery_merge/merge.php";

pagefooter();

?>