<?php
/**************************************************
  Coppermine 1.5.x Plugin - gallery_merge
  Plugin Support for {plugin_name}
**************************************************/

/*
   page included into merge_db.php to handle simulate and execution of merges for a specific plugin
   page should be named 'merge_plugin_{pluginfoldername}.php and placed in the gallery_merge/data directory
*/

// Make sure that this file can't be accessed directly, but only from within the Coppermine user interface
if (!defined('IN_COPPERMINE')) {
    die('Not in Coppermine...');
}

if (!GALLERY_ADMIN_MODE) {
    cpg_die(ERROR, $lang_errors['access_denied'], __FILE__, __LINE__);
}

// ********************
// PLUGIN Merge Support
// ********************
/*
  This skeleton is provided to add support for a plugin's data tables to Gallery Merge.

  When to use this script:
  * If a Plugin does NOT create it's own data tables, this script is not needed.
  * This is intended to support merge of plugin defined tables - taking advantage of the mapping already done by gallery merge,
    and extending those capabilities to a plugin's tables.
    In defining the support for Coppermine's core tables, a range of cases can be easily handled.
*/

/*
  See readme_plugin_doc.txt for detailed comments and parameters
  Unused variables or functions sections can be deleted.
*/

// ***********************
// Define Plugin Variables
// ***********************
// if any variables are required with GLOBAL scope, add them here - named $plugin_gallery_merge_{plugin_folder}_{name}
//   used to preserve/pass data between pre_merge and post_merge functions

$plugin_gallery_merge_{plugin_folder}_{varname} = 'global var for use in functions below...';

// ***********************
// Define Plugin Functions
// ***********************
// if any functions are required, add them here - named gallery_merge_plugin_{plugin_folder}_{name}
// to call function around processing for a given table (process_type = 'process_table') or instead of normal functions (process_type = 'custom'),
//   specify function name in 'pre_merge_function' field of plugin_table_array below to call BEFORE normal processing.
//   specify function name in 'post_merge_function' field of plugin_table_array below to call AFTER normal processing.
//   (for custom, either or both can be specified... no processingwill occur between them.)
// replace values in {} with applicable values
// to pass data between 'pre_merge' and 'post_merge' functions, specify global variables above and reference with global statement in function
// to save data across restarts, place data in array, and use sample checkpoint logic..
//   (typically should only be necessary in 'pre_merge' in event merge or 'post_merge' are interrupted - and process cannot just be re-run)


function gallery_merge_plugin_{plugin_folder}_{name}() {
  global $CONFIG, $SOURCE, $message_level, $debug_level, $debug_trace;          // core variables and settings
  global $restart_flag                                                          // checkpoint restart setting
  global $plugin_gallery_merge_execute;                                         // execution flags - only perform UPDATES if set to TRUE
  global $source_mapping_{tablesuffix}, $source_mapping_{tablesuffix}_key;      // access to tables mapping table and key
  global $plugin_gallery_merge_{plugin_folder}_{varname};                       // access to global variables defined above
  // common initialization
  $func_name = substr(__FUNCTION__, 21);                                        // strip off gallery_merge_plugin
  if ($debug_level >= 3) { $debug_trace[] = "Plugin Function $func_name Entered"; }  // debug footprint
  $process_name = "PROCESS_PLUGIN_$func_name";                                  // build processs name based on this function

  // if checkpointing data we need back on restart...
  if ($restart_flag == 'RESTART') {
    // this is a restart - see if we have data...
    $data = gallery_merge_checkpoint_load($process_name);                       // request our checkpoint data
    // if returned data is array - we have our data back (if only one chackpoint taken at end - should mean we are done processing)
    // if returned data == TRUE - checkpoint existed, but no data included
    // if returned data == FALSE - no checkpoint found, process as never run
  }

  // normal message - default should be high level status and exceptions
  // no provision currently for language support - possible future enhancement to include...
  $col1 = "&nbsp;";                                                             // column 1 in message output typically used only for section start - 'blank' or "&nbsp;" for others
  $col2 = "Message text here";
  gallery_merge_output_status($col1, $col2);
  // detail message for normal events (like row will be inserted)
  if ($message_level == 2) {
    $col1 = "&nbsp;";                                                           // column 1 in message output typically used only for section start - 'blank' or "&nbsp;" for others
    $col2 = "Message text here";
    gallery_merge_output_status($col1, $col2);
  }

  // issue warning - typically used during simulation for serious items that will cause failure in execution.. but we want to continue simulation
  // examples: plugin in source gallery that alters cpg tables is not in target... our inserts will have mismatch at execution, but want to complete simulation
  //           mapping that finds requested fields or results missing - resulting in data integrity issue if executed
  // execution will not be allowed until a simulation runs without warnings
  $message = "WARNING Message text here";
  gallery_merge_error('db', $process_name, 'warning', $message);

  // issue error - item serious enough to halt simulation.. we can't or shouldn't continue until corrected
  // error flag will bypass all remaining processing after this function - but still allow messages and debug information to be displayed.
  // avoid use of die - as user won't get to see other output.
  // execution will not be allowed until a simulation runs without errors
  $message = "ERROR Message text here";
  gallery_merge_error('db', $process_name, 'error', $message);

  // read sql
  $sql = "SELECT {columns} FROM {$CONFIG['{table_constant}']} WHERE {conditions) ORDER BY {columns}";
  $result = cpg_db_query($sql);                                                 // execute against TARGET DB 
  $result = cpg_db_query($sql, $SOURCE['LINK_ID']);                             // execute against SOURCE DB

  // update sql
  $update_sql = "UPDATE {remaining sql}";                                       // set sql
  gallery_merge_debug_sql($update_sql);                                         // send to debug
  if ($plugin_gallery_merge_execute) {
    // execute mode!
    cpg_db_query($update_sql);                                                  // execute against TARGET DB if execute mode only
                                                                                // NO updates should be made to SOURCE DB
  }

  // if all ok and data to save for restart... in array $data
  gallery_merge_checkpoint('db', 'checkpoint', '', "$process_name:COMPLETE", $data);
  // common termination
  if ($debug_level >= 3) { $debug_trace[] = "Plugin Function $func_name Completed"; }  // debug footprint
} // end function


// *******************
// Define Plugin Array
// *******************
//** Update variables to support merge of this plugin's data tables
//** see readme_plugin_doc.txt for detailed information on parameters
$plugin_name = '{Plugin Name}';                                                 // name as it appears in cpg_plugins table and data/plugin_data.php
// array of plugin's tables - repeat as needed for each plugin table for this plugin
// tables will be processed in the order specified... make sure to properly sequence those tables with dependencies..
// (if table b needs to translate values from table a mappings... then process (and map) a before processing b)
$plugin_table_array = array(
                            'table1_suffix' => array(                                                         // table name that follows the cpg table prefix
                                        'process_type' => 'compare_values',                                   // see detailed options in readme_plugin_doc.txt
                                                       // ****** following fields only required if 'process_type' = 'compare_values' ******
                                 'compare_name_column' => 'name',                                             // name of column containing the variable names
                                'compare_value_column' => 'value'                                            // name of the column containing the variable setting
                                                   ),  // end table definition


                            'table2_suffix' => array(                                                         // table name that follows the cpg table prefix
                                        'process_type' => 'process_table',                                    // see detailed options in readme_plugin_doc.txt
                                                       // ****** following fields only required if 'process_type' = 'process_table' or 'custom' ******
                                           'table_key' => 'aid',                                              // key field name
                                             'orderby' => '',                                                 // order by key field is default to assist restart - override here if needed
                                      'compare_fields' => 'title',                                            // field(s) - comma separated - to match source to target (identify duplicates/inserts)
                                        'compare_type' => 'UNIQUE',                                           // type of compare: UNIQUE, IGNORE, MERGE
                                           'skip_keys' => array('aid'),                                       // set keys/columns to bypass for insert (typically auto-increment fields)
                                       'obsolete_keys' => array(''),                                          // set keys/columns that are now obsolete - but might still exist in older installs
                                           'zero_keys' => array('thumb'),                                     // set keys/columns to 0 value on insert (typically data not yet available... to be updated later)
                                          'map_values' => array('/title', 'thumb'),                           // set mapping values to set fields to build array for later processing
                                'skip_keys_difference' => array('aid', 'thumb'),                              // set keys to bypass for comparing like rows in source and target (like passwords...)
                                        'translations' => array('visibility' => 'usergroups.visibility.group_id',
                                                                     'owner' => 'users.owner.user_id',
                                                                  'category' => 'categories.category.cid',
                                                                       'pos' => 'categories.category.incr_subalbum_pos'),  // set translations to be done... fieldname => table.keyvalue.field_to_use.optional_condition  
                                                       // ****** following fields (at least one) only required if 'process_type' = 'custom' - Optional for 'process_type = 'process_table' ******
                                  'pre_merge_function' => '',                                                 // function to call before execution of merge script... (process types of 'process_table' and 'custom' only)
                                 'post_merge_function' => ''                                                  // function to call following execution of merge script... (process types of 'process_table' and 'custom' only)
                                                   ),  // end table definition


                            'table3_suffix' => array(                                                         // table name that follows the cpg table prefix
                                        'process_type' => 'custom',                                            // see detailed options in readme_plugin_doc.txt
                                                       // ****** following fields (at least one) only required if 'process_type' = 'custom' - Optional for 'process_type = 'process_table' ******
                                  'pre_merge_function' => '',                                                 // function to call before execution of merge script... (process types of 'process_table' and 'custom' only)
                                 'post_merge_function' => ''                                                  // function to call following execution of merge script... (process types of 'process_table' and 'custom' only)
                                                   ),  // end table definition


                            'table4_suffix' => array(                                                         // table name that follows the cpg table prefix
                                        'process_type' => 'transient'                                         // see detailed options in readme_plugin_doc.txt
                                                   ),  // end table definition


                            'table5_suffix' => array(                                                         // table name that follows the cpg table prefix
                                        'process_type' => 'static'                                            // see detailed options in readme_plugin_doc.txt
                                                   ),  // end table definition


                            'table6_suffix' => array(                                                         // table name that follows the cpg table prefix
                                        'process_type' => 'unsupported'                                       // see detailed options in readme_plugin_doc.txt
                                                   ),  // end table definition


                           ); // end all tables definitions

//** add this script name to the merge_script variable for this plugin in data/plugin_data.php to have it invoked when plugin is in use in source and target galleries.
?>