<?php
/**************************************************
  Coppermine 1.5.x Plugin - gallery_merge
**************************************************/

// Make sure that this file can't be accessed directly, but only from within the Coppermine user interface
if (!defined('IN_COPPERMINE')) {
    die('Not in Coppermine...');
}

// base install, uninstall, and direct plugin called functions are in codebase.php
// sub functions may be shared between codebase, admin, and viewgalleryphoto pages and reside here.

// ***********************************
// CPG Core function changes/additions
// ***********************************

// alternate version of cpg_db_connect to support an alternate $CONFIG variable
if (!function_exists('cpg_db_connect_alt')) {
  function cpg_db_connect_alt($use_alt_config = 0)
  {
    global $CONFIG;
                                                                                            
    if ($use_alt_config) {
        $db_config =& $use_alt_config;
    } else {                              
        $db_config =& $CONFIG;
    }                                     

    $result = @mysql_connect($db_config['dbserver'], $db_config['dbuser'], $db_config['dbpass']);
                                                                                            
    if (!$result) {                                                                         
        return false;                                                                       
    }                                                                                       
                                                                                            
    if (!mysql_select_db($db_config['dbname'])) {
        return false;                                                                       
    }                                                                                       
                                                                                            
    if (!empty($db_config['dbcharset'])) {
        mysql_query("SET NAMES '{$db_config['dbcharset']}'", $result);
    }                                                                                       
                                                                                            
    return $result;                                                                         
  }
}

// cpg missing function to surround mysql_escape_string
if (!function_exists('cpg_db_escape_string')) {
  function cpg_db_escape_string($string)
  {
      return mysql_escape_string($string);
  }
}


// cpg missing function to surround mysql_real_escape_string
if (!function_exists('cpg_db_real_escape_string')) {
  function cpg_db_real_escape_string($string, $use_link_id = 0) {
      global $CONFIG;

      if ($use_link_id) {
          $link_id = $use_link_id;
      } else {
          $link_id = $CONFIG['LINK_ID'];
      }
      return mysql_real_escape_string($string, $link_id);
  }
}


// cpg missing function to surround mysql_free_result
// $result passed by reference in this case
if (!function_exists('cpg_db_free_result')) {
  function cpg_db_free_result(&$result)
  {
      mysql_free_result($result);
  }
}


// cpg missing function to surround mysql_result
if (!function_exists('cpg_db_result')) {
  function cpg_db_result($result, $row = 0, $field = 0)
  {
      return mysql_result($result, $row, $field);
  }
}

                                                                                                                                                          
// cpg missing function to surround mysql_num_rows
if (!function_exists('cpg_db_num_rows')) {
  function cpg_db_num_rows($result) {
      return mysql_num_rows($result);
  }
}


// cpg missing function to surround mysql_affected_rows
if (!function_exists('cpg_db_affected_rows')) {
  function cpg_db_affected_rows($use_link_id = 0) {
      global $CONFIG;

      if ($use_link_id) {
          $link_id = $use_link_id;
      } else {
          $link_id = $CONFIG['LINK_ID'];
      }
      return mysql_affected_rows($link_id);
  }
}


// ***************************
// Validate access to $GLOBALS
// ***************************

// validate $GLOBALS is accessible
function gallery_merge_validate_globals() {
  // if register_globals is ON - CPG executes code that impacts ability to use $GLOBALS - even for non-input variables.
  // reported on forum with workaround. This code will validate $GLOBALS are accessible (only invoked if register_globals is ON - though could run always)
  global $plugin_gallery_merge_globals_test1;
  if ($plugin_gallery_merge_globals_test1 == $GLOBALS['plugin_gallery_merge_globals_test2']) {
    return TRUE;
  } else {
    return FALSE;
  }
}


// *********************************
// Execution Options and Validations
// *********************************

// determine if the provided option is valid based on CONFIG statuses
function gallery_merge_validate_option($option) {
  // see if the passed option is a valid choice based on config timestamps and existance of checkpoints (in some cases)
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  switch ($option) {
    case 'sim_db' :
      // for simulation db merge, config must have been completed at least once
      if ($CONFIG['plugin_gallery_merge_ts_config_update'] > '0000-00-00 00:00:00') {
        return TRUE;                                                            // allow simulation as long as config was executed once
      }
    break;

    case 'sim_restart_db' :
      // for simulating db RESTART, simulation must have started, NOT ended (successfully or with warnings), and checkpoints must exist
      if (($CONFIG['plugin_gallery_merge_ts_sim_db_start'] > $CONFIG['plugin_gallery_merge_ts_sim_db_end']) &&
          ($CONFIG['plugin_gallery_merge_ts_sim_db_start'] > $CONFIG['plugin_gallery_merge_ts_sim_db_warn']) &&
          (gallery_merge_checkpoint_exists($CONFIG['plugin_gallery_merge_ts_sim_db_start']))) {
        return TRUE;                        // allow SIM RESTART if sim start > sim end - and checkpoints exist
      }
    break;

    case 'exec_db' :
      // for executing db merge, simulation must have started/ended without warnings after last config update - and after last execution.. exec restart DB must NOT be an option
      if (($CONFIG['plugin_gallery_merge_ts_sim_db_end'] >= $CONFIG['plugin_gallery_merge_ts_sim_db_start']) &&
          ($CONFIG['plugin_gallery_merge_ts_sim_db_end'] > $CONFIG['plugin_gallery_merge_ts_sim_db_warn']) &&
          ($CONFIG['plugin_gallery_merge_ts_sim_db_end'] > $CONFIG['plugin_gallery_merge_ts_config_update']) &&
          ($CONFIG['plugin_gallery_merge_ts_sim_db_end'] > $CONFIG['plugin_gallery_merge_ts_exec_db_end']) &&
          (!gallery_merge_validate_option('exec_restart_db'))) {
        return TRUE;                                                            // allow DB execution if sim started/ended after last config update
      }
    break;

    case 'exec_restart_db' :
      // for executing db RESTART, execution must have started, NOT ended, BACKOUT must not have completed since execution, and checkpoints must exist
      // (backout could have started... partial backouts are handled during checkpoint restart)
      // (use start time for comparison against backout - as end times could be 0 or from prev execution.
      if (($CONFIG['plugin_gallery_merge_ts_exec_db_start'] > $CONFIG['plugin_gallery_merge_ts_exec_db_end']) &&
          ($CONFIG['plugin_gallery_merge_ts_exec_db_start'] > $CONFIG['plugin_gallery_merge_ts_exec_dbbko_end']) &&
          (gallery_merge_checkpoint_exists($CONFIG['plugin_gallery_merge_ts_exec_db_start']))) {
        return TRUE;                        // allow EXEC RESTART if sim completed, exec start > exec end - and checkpoints exist
      }
    break;

    case 'sim_backout_db' :
      // for simulating db BACKOUT, execution must have started, and checkpoints must exist, backout execution must not have completed since execution
      // backout of last started execution will remain an option until sucessful - but obviously database updates made since the restore point was taken will be lost!!
      if (($CONFIG['plugin_gallery_merge_ts_exec_db_start'] > '0000-00-00 00:00:00') &&
          ($CONFIG['plugin_gallery_merge_ts_exec_db_start'] > $CONFIG['plugin_gallery_merge_ts_exec_dbbko_end']) &&
          (gallery_merge_checkpoint_exists($CONFIG['plugin_gallery_merge_ts_exec_db_start']))) {
        return TRUE;                        // allow SIM BACKOUT if execution started - and checkpoints exist
      }
    break;

    case 'exec_backout_db' :
      // for executing db BACKOUT, execution must have started, backout must not have completed since,
      //   a backout simulation must have started/completed after exec start, and checkpoints must exist
      // backout of last started execution will remain an option until sucessful - but obviously database updates made since the restore point was taken will be lost!!
      if (($CONFIG['plugin_gallery_merge_ts_exec_db_start'] > '0000-00-00 00:00:00') &&
          ($CONFIG['plugin_gallery_merge_ts_exec_db_start'] > $CONFIG['plugin_gallery_merge_ts_exec_dbbko_end']) &&
          ($CONFIG['plugin_gallery_merge_ts_sim_dbbko_start'] >= $CONFIG['plugin_gallery_merge_ts_exec_db_start']) &&
          ($CONFIG['plugin_gallery_merge_ts_sim_dbbko_end'] >= $CONFIG['plugin_gallery_merge_ts_sim_dbbko_start']) &&
          (gallery_merge_checkpoint_exists($CONFIG['plugin_gallery_merge_ts_exec_db_start']))) {
        return TRUE;                        // allow EXEC RESTART if sim completed, exec start > exec end - and checkpoints exist
      }
    break;

    case 'emergency_backout':
      // a non-displayed config option allows emergency backout to any timestamp (assuming matching backup files exist of course...)
      if ($CONFIG['plugin_gallery_merge_ts_emer_backout'] != '0000-00-00 00:00:00') {
        return TRUE;                        // allow EMER BACKOUT if config setting is non-zero
      }
    break;

    case 'reset_db' :
      // if either sim restart or exec restart is available - so is reset...
      if ((gallery_merge_validate_option('sim_restart_db')) ||
          (gallery_merge_validate_option('exec_restart_db'))) {
        return TRUE;                                                            // allow reset if either sim restart or exec restart is available
      }
    break;

    case 'sim_files' :
      // for simulation files merge, config must have been completed at least once, and filesystem must be local
      if (($CONFIG['plugin_gallery_merge_source_local'] == '1') &&
          ($CONFIG['plugin_gallery_merge_ts_config_update'] > '0000-00-00 00:00:00')) {
        return TRUE;                                                            // allow simulation as long as config was executed once
      }
    break;

    case 'exec_files' :
      // for executing file merge, simulation must have started/ended after last config update - and after last execution, and filesystem must be local.. exec restart files must NOT be an option
      if (($CONFIG['plugin_gallery_merge_source_local'] == '1') &&
          ($CONFIG['plugin_gallery_merge_ts_sim_files_end'] >= $CONFIG['plugin_gallery_merge_ts_sim_files_start']) &&
          ($CONFIG['plugin_gallery_merge_ts_sim_files_end'] > $CONFIG['plugin_gallery_merge_ts_config_update']) &&
          ($CONFIG['plugin_gallery_merge_ts_sim_files_end'] > $CONFIG['plugin_gallery_merge_ts_exec_files_end']) &&
          (!gallery_merge_validate_option('exec_restart_files'))) {

        return TRUE;                        // allow FILE execution if sim started/ended after last config update
      }
    break;

    case 'exec_restart_files' :
    case 'reset_files' :
      // for executing files RESTART, execution must have started, NOT ended, checkpoints must exist, and filesystem must be local
      // if restart is an option - so is reset...
      if (($CONFIG['plugin_gallery_merge_source_local'] == '1') &&
          ($CONFIG['plugin_gallery_merge_ts_exec_files_start'] > $CONFIG['plugin_gallery_merge_ts_exec_files_end']) &&
          (gallery_merge_checkpoint_exists($CONFIG['plugin_gallery_merge_ts_exec_files_start']))) {
        return TRUE;                        // allow EXEC RESTART if sim completed, exec start > exec end - and checkpoints exist
      }
    break;

    case 'compare_db' :
      // for executing files compare, execution must have ended afer confg was last set... and backout must not have fun since execution
      if (($CONFIG['plugin_gallery_merge_ts_exec_db_end'] > $CONFIG['plugin_gallery_merge_ts_config_update']) &&
          ($CONFIG['plugin_gallery_merge_ts_exec_db_end'] > $CONFIG['plugin_gallery_merge_ts_sim_dbbko_end'])) {
        return TRUE;                        // allow db compare if exec completed after config, and no backout since
      }
    break;

    case 'compare_files' :
      // for executing files compare, execution must have ended afer confg was last set
      if ($CONFIG['plugin_gallery_merge_ts_exec_files_end'] > $CONFIG['plugin_gallery_merge_ts_config_update']) {
        return TRUE;                        // allow files compare if exec completed after config
      }
    break;

    case 'compare_gallery' :
      // for executing gallery compare, either of compare_files or compare_db must be true
      if ((gallery_merge_validate_option('compare_files')) ||
          (gallery_merge_validate_option('compare_db'))) {
        return TRUE;                        // allow gallery compare if files compare or db ccompare is allowed
      }
    break;

    case 'set_offline' :
      // for setting gallery online, it must be offline currently
      if (!$CONFIG['offline']) {
        return TRUE;                        // allow set offline if gallery is online
      }
    break;

    case 'set_online' :
      // for setting gallery online, it must be offline currently
      if ($CONFIG['offline']) {
        return TRUE;                        // allow set online if gallery is offline
      }
    break;

    default:
      // undefined case
      return FALSE;
    break;
  }
  return FALSE;
}


// **********************
// Gallery Access Control
// **********************

function gallery_merge_offline($action) {
  // view/alter gallery offline/online status
  // $action  - check - return gallery status
  //          - warnoffline  - issue warning if gallery is offline
  //          - warnonline   - issue warning if gallery is online
  //          - offline      - set gallery offline
  //          - online       - set gallery online
  //          - reset        - set gallery to initial saved status
  global $CONFIG;                                                                       // CONFIG parms
  global $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;        // language and debug
  static $save_gallery_status = 'unknown';

  if ($save_gallery_status == 'unknown') {
    // save initial status on first invocation
    $save_gallery_status = $CONFIG['offline'];
    $debug_trace[] = "Gallery Status: status ($save_gallery_status) saved";
  }
  switch ($action) {
    case 'check':
      return $CONFIG['offline'];
    break;
    case 'warnoffline':
      if ($CONFIG['offline']) {
        $col1 = $lang_plugin_gallery_merge['gallery_status'];
        $col2 = $lang_plugin_gallery_merge['gallery_warn_offline'];
        gallery_merge_output_status($col1, $col2);
      } else {
        $col1 = $lang_plugin_gallery_merge['gallery_status'];
        $col2 = $lang_plugin_gallery_merge['gallery_online'];
        gallery_merge_output_status($col1, $col2);
      }
    break;
    case 'warnonline':
      if (!$CONFIG['offline']) {
        $col1 = $lang_plugin_gallery_merge['gallery_status'];
        $col2 = $lang_plugin_gallery_merge['gallery_warn_online'];
        gallery_merge_output_status($col1, $col2);
      } else {
        $col1 = $lang_plugin_gallery_merge['gallery_status'];
        $col2 = $lang_plugin_gallery_merge['gallery_offline'];
        gallery_merge_output_status($col1, $col2);
      }
    break;
    case 'offline':
      if (!$CONFIG['offline']) {
        $CONFIG['offline'] = '1';
        $sql = "UPDATE {$CONFIG['TABLE_CONFIG']} SET value = '{$CONFIG['offline']}' WHERE name = 'offline'";
        $result = cpg_db_query($sql);                                           // write to DB
        $debug_trace[] = "Gallery Status: Gallery set offline";
        $col1 = $lang_plugin_gallery_merge['gallery_status'];
        $col2 = $lang_plugin_gallery_merge['gallery_set_offline'];
        gallery_merge_output_status($col1, $col2);
      } else {
        $col1 = $lang_plugin_gallery_merge['gallery_status'];
        $col2 = $lang_plugin_gallery_merge['gallery_offline'];
        gallery_merge_output_status($col1, $col2);
      }
    break;
    case 'online':
      if ($CONFIG['offline']) {
        $CONFIG['offline'] = '0';
        $sql = "UPDATE {$CONFIG['TABLE_CONFIG']} SET value = '{$CONFIG['offline']}' WHERE name = 'offline'";
        $result = cpg_db_query($sql);                                           // write to DB
        $debug_trace[] = "Gallery Status: Gallery set online";
        $col1 = $lang_plugin_gallery_merge['gallery_status'];
        $col2 = $lang_plugin_gallery_merge['gallery_set_online'];
        gallery_merge_output_status($col1, $col2);
      } else {
        $col1 = $lang_plugin_gallery_merge['gallery_status'];
        $col2 = $lang_plugin_gallery_merge['gallery_online'];
        gallery_merge_output_status($col1, $col2);
      }
    break;
    case 'reset':
      if ($CONFIG['offline'] != $save_gallery_status) {
        $CONFIG['offline'] = $save_gallery_status;
        $sql = "UPDATE {$CONFIG['TABLE_CONFIG']} SET value = '{$CONFIG['offline']}' WHERE name = 'offline'";
        $result = cpg_db_query($sql);                                           // write to DB
        if ($save_gallery_status) {
          $debug_trace[] = "Gallery Status: Gallery status reset to offline";
          $col1 = $lang_plugin_gallery_merge['gallery_status'];
          $col2 = $lang_plugin_gallery_merge['gallery_reset_offline'];
          gallery_merge_output_status($col1, $col2);
        } else {
          $col1 = $lang_plugin_gallery_merge['gallery_status'];
          $col2 = $lang_plugin_gallery_merge['gallery_reset_online'];
          gallery_merge_output_status($col1, $col2);
        }
      }
    break;
    default:
    break;
  } // end switch
}


// *************************
// Checkpoint and Timestamps
// *************************

function gallery_merge_checkpoint_settings() {
  // return array with key plugin CONFIG settings - used for 'INITIATED' checkpoint calls
  global $CONFIG, $debug_trace;
  $settings = array();
  $prefix = 'plugin_gallery_merge_';
  $length = strlen($prefix);
  foreach($CONFIG AS $name => $value) {
    if (substr($name, 0, $length) == $prefix) {
      // capture just our settings
      if ($name == 'plugin_gallery_merge_source_dbpass') {
        $value = '*SUPPRESSED*';
      }
      $settings[$name] = $value;
    }
  }
  return $settings;
}


function gallery_merge_checkpoint($runtype, $calltype, $record, $text = '', $data = '') {
  // common checkpoint and time recording routine
  // $runtype  - config, db, or files
  // $calltype - timestamp, checkpoint, or restart
  // $record   - key to record ts under - update, start, warn, end, restart*, backout*, or none/null
  // $text     - text to place in checkpoint record (will be prepended by standard timestamp)
  // $data     - array of data to be recorded (optional) typically used for data to be reloaded on checkpoint restart
  // Invocations:
  //   gallery_merge_checkpoint('config', 'timestamp', 'update');
  //   gallery_merge_checkpoint('db', 'timestamp', 'start');
  //   gallery_merge_checkpoint('db', 'timestamp', 'end');
  //   gallery_merge_checkpoint('db', 'timestamp', 'restart', 'start timestamp');
  //   gallery_merge_checkpoint('files', 'timestamp', 'start');
  //   gallery_merge_checkpoint('files', 'timestamp', 'end');
  //   gallery_merge_checkpoint('db', 'checkpoint', '', 'checkpoint text');     (typically table name completed)
  //   gallery_merge_checkpoint('files', 'checkpoint', '', 'checkpoint text');  (typically filepath completed)
  global $CONFIG;                                                                       // CONFIG parms
  global $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;        // language and debug
  global $plugin_gallery_merge_simulate, $plugin_gallery_merge_execute;                 // DB merge flags
  global $plugin_gallery_merge_backout_simulate, $plugin_gallery_merge_backout_execute, $plugin_gallery_merge_backout_emergency; // DB backout flags
  global $plugin_gallery_merge_simulate_files, $plugin_gallery_merge_execute_files;     // File merge flags
  static $tsprefix = '';
  static $chkptfield = '';
  static $starttime = '';
  static $checkpoint = FALSE;
  // initialization first time through...
  if ($tsprefix == '') {
    // first time thru - set up prefix
    if ($runtype == 'config') {
      $tsprefix = 'plugin_gallery_merge_ts_config_';                            // timestamp variable prefix for config
    } elseif (($runtype == 'db') && ($plugin_gallery_merge_execute)) {
      $tsprefix = 'plugin_gallery_merge_ts_exec_db_';                           // timestamp variable prefix for db execution
    } elseif (($runtype == 'db') && ($plugin_gallery_merge_simulate)) {
      $tsprefix = 'plugin_gallery_merge_ts_sim_db_';                            // timestamp variable prefix for db simulation
    } elseif (($runtype == 'dbbko') && ($plugin_gallery_merge_backout_execute || $plugin_gallery_merge_backout_emergency)) {
      $tsprefix = 'plugin_gallery_merge_ts_exec_dbbko_';                        // timestamp variable prefix for db backout execution
      $tsexec_db = 'plugin_gallery_merge_ts_exec_db_start';                     // timestamp variable prefix for db execution to validate
    } elseif (($runtype == 'dbbko') && ($plugin_gallery_merge_backout_simulate)) {
      $tsprefix = 'plugin_gallery_merge_ts_sim_dbbko_';                         // timestamp variable prefix for db backout simulation
      $tsexec_db = 'plugin_gallery_merge_ts_exec_db_start';                     // timestamp variable prefix for db execution to validate
    } elseif (($runtype == 'files') && ($plugin_gallery_merge_execute_files)) {
      $tsprefix = 'plugin_gallery_merge_ts_exec_files_';                        // timestamp variable prefix for files execution
    } elseif (($runtype == 'files') && ($plugin_gallery_merge_simulate_files)) {
      $tsprefix = 'plugin_gallery_merge_ts_sim_files_';                         // timestamp variable prefix for files simulation
    } elseif ($runtype == 'util') {
      $tsprefix = 'unused';                                                     // timestamp variable prefix unused for util
    }
  }
  if ($chkptfield == '') {
    // first time thru - set up prefix
    if ($plugin_gallery_merge_execute || $plugin_gallery_merge_execute_files || $plugin_gallery_merge_reset_db || $plugin_gallery_merge_reset_files ||
        $plugin_gallery_merge_backout_execute || $plugin_gallery_merge_backout_emergency) {
      $chkptfield = 'plugin_gallery_merge_ts_exec_chkpttext';                   // checkpoint variable for execution
    } else {
      $chkptfield = 'plugin_gallery_merge_ts_sim_chkpttext';                    // checkpoint variable for simulation
    }
  }
  // initialization complete - continue on...
  $ts = date("Y-m-d H:i:s");
  if ($calltype == 'timestamp') {
    if ($record == 'restart') {
      // for restart processing, we want to re-establish original start time - so checkpointing can continue.
      // no CONFIG updates needed... CONFIG array and database will already contain proper time - which we will validate
      // log our validated restart to TABLE_RESTART fo tracking/audit
      // $text - restart timestamp passed from form
      $tsfield = $tsprefix.'start';
      if ($CONFIG[$tsfield] == $text) {
        // valid restart checkpoint
        $starttime = $CONFIG[$tsfield];                                         // retain starttime from original start for future checkpoints
        $chkpt_sql = "INSERT INTO  {$CONFIG['TABLE_RESTART']} (`timestamp`, `text`) VALUES ('$ts', 'RESTARTING TIMESTAMP: {$starttime}')";
        $result = cpg_db_query($chkpt_sql);
        $debug_trace[] = "Checkpoint Restart: Valid restart request for $starttime";
      } else {
        $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_restart_time'], $text, $CONFIG[$tsfield]);
        gallery_merge_error('init', 'PROCESS_CHECKPOINT', 'error', $error_message);
      }
    } elseif ($record == 'backout') {
      // for backout processing, we want to re-establish original start time - to find correct backup files and invalidate a later attempted restart.
      // no CONFIG updates needed... CONFIG array and database will already contain proper time - which we will validate
      // log our validated backout to TABLE_RESTART fo tracking/audit
      // $text - backout timestamp passed from form
      $tsfield = $tsprefix.'start';
      if (($CONFIG[$tsexec_db] == $text) || $plugin_gallery_merge_backout_emergency) {
        // valid backout checkpoint
        $starttime = $CONFIG[$tsfield] = $text;                                 // retain starttime from original start for future checkpoints (or use emergency override)
        if ($plugin_gallery_merge_backout_execute) {
          $backout_type = 'EXECUTION';
        } elseif ($plugin_gallery_merge_backout_emergency) {
          $backout_type = 'EMERGENCY EXECUTION';
        } else {
          $backout_type = 'SIMULATION';
        }
        $chkpt_sql = "INSERT INTO  {$CONFIG['TABLE_RESTART']} (`timestamp`, `text`) VALUES ('$ts', 'BACKOUT $backout_type TO TIMESTAMP: {$starttime}')";
        $result = cpg_db_query($chkpt_sql);
        $chkpt_sql = "UPDATE {$CONFIG['TABLE_CONFIG']} SET value = '{$CONFIG[$tsfield]}' WHERE name = '$tsfield'";
        $result = cpg_db_query($chkpt_sql);                                       // write to DB
        $debug_trace[] = "Checkpoint Restart: Valid backout $backout_type request for $starttime";
      } else {
        $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_backout_time'], $text, $CONFIG[$tsexec_db]);
        gallery_merge_error('init', 'PROCESS_CHECKPOINT', 'error', $error_message);
        $debug_trace[] = "Checkpoint Restart: Valididation failed - comparing config record '$tsexec_db' contents '{$CONFIG[$tsexec_db]}' with input '$text'";
      }

    } else {
      $tsfield = $tsprefix.$record;
      $CONFIG[$tsfield] = $ts;                                                  // set time in CONFIG array
      $chkpt_sql = "UPDATE {$CONFIG['TABLE_CONFIG']} SET value = '{$CONFIG[$tsfield]}' WHERE name = '$tsfield'";
      $result = cpg_db_query($chkpt_sql);                                       // write to DB
      if ($record == 'start') {
        $starttime = $ts;                                                       // starttime will be used for all data written to restart table
      }
    }
    return $starttime;
  } elseif ($calltype == 'checkpoint') {
    $CONFIG[$chkptfield] = "$ts $text";
    $chkpt_sql = "UPDATE {$CONFIG['TABLE_CONFIG']} SET value = '{$CONFIG[$chkptfield]}' WHERE name = '$chkptfield'";
    $result = cpg_db_query($chkpt_sql);
    if ((preg_match('/(INITIATED|COMPLETED)/', $text, $matches)) && ($data == '')) {
      // initiated or completed message - capture settings for checkpoint (note completed vs complete)
      $data = gallery_merge_checkpoint_settings();
    }
    if ((strpos($text,'TABLE_') !== FALSE) && ($data == '')) {
      // checkpoint for a table without data...     see if mapping array exists (specifically for plugin table merges... where we don't know array names to global...)
      $textdata =  explode(':', $text);
      $table_constant = $textdata[0];                                           // process name if first word of checkpoint text
      $table_name_uc = substr($table_constant, 6);                              // uppercase table constant for messages
      $table_name = strtolower($table_name_uc);                                 // lowercase table constant for messages
      $table_status = $checkpoint[1];
      if (is_array($GLOBALS["source_mapping_$table_name"])) {
        $data =& $GLOBALS["source_mapping_$table_name"];                        // establish global reference to table
      }
    }
    if ($runtype == 'util') {
      $starttime = $ts;                                                         // use current time for util execution
    }
    if ($data != '') {
      $datastring = mysql_escape_string(serialize($data));
      $chkpt_sql = "INSERT INTO  {$CONFIG['TABLE_RESTART']} (`timestamp`, `text`, `data`) VALUES ('$starttime', '$text', '$datastring')";
    } else {
      $chkpt_sql = "INSERT INTO  {$CONFIG['TABLE_RESTART']} (`timestamp`, `text`) VALUES ('$starttime', '$text')";
    }
    $result = cpg_db_query($chkpt_sql);
  }
}


// Checkpoints exist?
function gallery_merge_checkpoint_exists($timestamp) {
  // see if checkpoint records exist for given timestamp (indicates restart is possible)
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  $result = cpg_db_query("SELECT timestamp, COUNT( * ) AS count FROM {$CONFIG['TABLE_RESTART']} WHERE timestamp = '$timestamp' GROUP BY timestamp");
  $row = cpg_db_fetch_row($result);
  cpg_db_free_result($result);
  if ($row['count'] > 0) {
    if ($debug_level == 4) { $debug_trace[] = "Checkpoint: {$row['count']} checkpoint records found for $timestamp"; }
    return TRUE;
  } else {
    return FALSE;
  }
}


// Load Checkpoint data
function gallery_merge_checkpoint_load($text_mask) {
  // load last checkpoint for this process at timestamp
  // return checkpointed data as array if exists
  // return TRUE if no data, but checkpoint found
  // return FALSE if no checkpoint found
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $starttime;
  $result = cpg_db_query("SELECT * FROM {$CONFIG['TABLE_RESTART']} WHERE timestamp = '$starttime' AND text LIKE '{$text_mask}%' ORDER BY id DESC LIMIT 1");
  $row = cpg_db_fetch_row($result);
  cpg_db_free_result($result);
  if ($row) {
    if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Load: Checkpoint record loaded for $text_mask at $starttime"; }
      if ($row['data'] != '') {
        $data = unserialize($row['data']);
        if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Load: Data loaded for process $text_mask"; }
        return $data;
      } else {
        return TRUE;
      }
  } else {
    return FALSE;
  }
}


// Load Checkpoint data
function gallery_merge_checkpoint_load_id($record_id) {
  // load specified record id
  // return checkpointed data as array if exists
  // return TRUE if no data, but checkpoint found
  // return FALSE if no checkpoint found
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $starttime;
  $result = cpg_db_query("SELECT * FROM {$CONFIG['TABLE_RESTART']} WHERE id = '$record_id' LIMIT 1");
  $row = cpg_db_fetch_row($result);
  cpg_db_free_result($result);
  if ($row) {
    if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Load: Checkpoint record loaded for record id $record_id for {$row['timestamp']} - {$row['text']}"; }
      if ($row['data'] != '') {
        $data = unserialize($row['data']);
        if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Load: Data loaded for record id $record_id"; }
        return $data;
      } else {
        return TRUE;
      }
  } else {
    return FALSE;
  }
}


// Checkpoint restart for DB processing
function gallery_merge_checkpoint_restart_db($timestamp) {
  // process restart request - reloading mapping arrays from restart table and setting resume point
  global $CONFIG, $RESTART, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_warning, $plugin_gallery_merge_warning_text;     // warning flags - needed to reload/re-establish warnings from previous failed attempt

  $result = cpg_db_query("SELECT * FROM {$CONFIG['TABLE_RESTART']} WHERE timestamp = '$timestamp' ORDER BY id DESC");
  $first = TRUE;
  $first_error = '';
  $warning_loaded = FALSE;
  $runtype = 'db';
  while ($row = cpg_db_fetch_row($result)) {
    // process checkpoints taken by the earlier execution(s) in reverse order - capture most recent data for each map
    // update $RESTART array for each table to indicate 'COMPLETE'
    // $row['text'] contains checkpoint text
    // $row['data'] contains mapping array checkpointed
    $load_map = FALSE;
    $load_warning = FALSE;
    $checkpoint =  explode(':', $row['text']);
    $table_constant = $checkpoint[0];                                           // target table name is first word of checkpoint text
    $table_indicator = substr($table_constant, 0, 6);                           // extract prefix - if TABLE_ - this is a table checkpoint
    $process_indicator = substr($table_constant, 0, 8);                         // extract prefix - if PROCESS_ - this is a process checkpoint
    $table_name_uc = substr($table_constant, 6);                                // uppercase table constant for messages
    $table_name = strtolower($table_name_uc);                                   // lowercase table constant for messages
    $table_status = $checkpoint[1];
    if ($first) {
      // first record read...
      if (($table_indicator == 'TABLE_') || ($process_indicator == 'PROCESS_') || ($RESTART[$table_constant] == '0')) {
        // record for a table - or record for a process - or we have specifically noted in RESTART array
        // other log records like process initiation (failed restart) or log data might be encountered first
        if ($table_status == 'ERROR') {
          // last execution was terminated by error
          $debug_trace[] = "Checkpoint Restart: Last execution ended with ERROR in $table_constant";
          $first = FALSE;
          $first_error = $table_constant;
        } elseif ($table_status == 'STARTED') {
          // last execution was unexpectedly ended
          $debug_trace[] = "Checkpoint Restart: Last execution ended unexpectedly in $table_constant";
          $RESTART[$table_constant] = '-1';
          $first = FALSE;
        } elseif ($table_status == 'WARNING') {
          // last execution issued warning - don't consider this a 'first' record - keep looking...
          $debug_trace[] = "Checkpoint Restart: Last execution issued WARNING in $table_constant";
          if (!$warning_loaded) {
            $load_warning = TRUE;
          }
        } else {
          if ($table_status == 'QUIESCE') {
            $lastkey = $checkpoint[4];
            $debug_trace[] = "Checkpoint Restart: Last execution ended with Quiesce in table $table_name at key $lastkey";
          } elseif ($table_status == 'INSERTED') {
            $lastkey = $checkpoint[4];
            $debug_trace[] = "Checkpoint Restart: Last execution ended unexpectedly in table $table_name after key $lastkey";
          } elseif ($table_status == 'COMPLETE') {
            $debug_trace[] = "Checkpoint Restart: Last execution ended unexpectedly after completing $table_constant";
            $lastkey = 'COMPLETE';
          }
          $RESTART[$table_constant] = $lastkey;
          if ($row['data'] != '') {
            $load_map = TRUE;
          }
          $first = FALSE;
        }
      } // endif a table we care about
    } else {
      // second and subsequent records... may be earlier checkpoints for same table process - or previous table.
      // if this is for a table or process... and either RESTART doesn't contain key - or key = 0...
      if ((($table_indicator == 'TABLE_') || ($process_indicator == 'PROCESS_')) && (!array_key_exists($table_constant, $RESTART) || ($RESTART[$table_constant] == '0'))) {
        // first record for a table/process we haven't seen yet
        if ($table_status == 'BACKOUT') {
          // last operation on this table was backout - need to reprocess
            $RESTART[$table_constant] = '-1';                                   // set to ignore prior checkpoint records for this table
        } elseif ($table_status == 'STARTED') {
          // last operation on this table was start of processing - need to reprocess
            $RESTART[$table_constant] = '-1';                                   // set to ignore prior checkpoint records for this table
        } elseif ($table_status == 'COMPLETE') {
          // good checkpoint
          $RESTART[$table_constant] = $table_status;
          if ($row['data'] != '') {
            $load_map = TRUE;
          }
        } elseif (($table_status == 'ERROR') && ($first_error == $table_constant)) {
          // multiple errors for same process - ignore and continue
        } else {
          // shouldn't happen... first record seen for any table except first encountered should be COMPLETE or BACKOUT...
          $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_checkpoint'], $table_constant, $table_status, $RESTART[$table_constant]);
          gallery_merge_error('init', 'PROCESS_RESTART', 'error', $error_message);
          break;
        }
      } elseif ($table_status == 'WARNING') {
        // either a table or process we've already seen - or not one we are tracking... but still check for warning...
        // if warning - and we haven't already loaded warning data (it is cumulative... so we want the last - which we processed first...)
        $debug_trace[] = "Checkpoint Restart: Last execution issued WARNING in $table_constant";
        if (!$warning_loaded) {
          $load_warning = TRUE;
        }
      }
    }
    if ($load_map) {
      // have a current map to load
      $map_var = "source_mapping_{$table_name}";                                    // map used for checkpointing
      if (is_array($GLOBALS[$map_var])) {
        // if array is defined (will be empty, but exist) - load data...
        $GLOBALS[$map_var] = unserialize($row['data']);
        if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Restart: Map array $map_var loaded for $table_constant"; }
      } else {
        // data found but array doesn't exist... no sense loading as it will vanish at end of function
        if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Restart: Checkpoint data for $table_constant exists, but NOT loaded."; }
      }
    }
    if ($load_warning) {
      $plugin_gallery_merge_warning_text = unserialize($row['data']);           // reload warning message array
      $plugin_gallery_merge_warning = TRUE;                                     // and show we have a warning
      if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Restart: Warning message data loaded."; }
    }
  }
  cpg_db_free_result($result);
}


// Checkpoint restart for File processing
function gallery_merge_checkpoint_restart_files($timestamp) {
  // process restart request - reloading mapping arrays from restart table and setting resume point
  global $CONFIG, $SOURCE, $RESTART, $RESTART_DIR, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  // no warning message or process started support in files

  $result = cpg_db_query("SELECT * FROM {$CONFIG['TABLE_RESTART']} WHERE timestamp = '$timestamp' ORDER BY id DESC");
  $first = TRUE;
  $first_error = '';
  $runtype = 'files';
  while ($row = cpg_db_fetch_row($result)) {
    // process checkpoints taken by the earlier execution(s) in reverse order - capture most recent data for each directory
    // update $RESTART array for each process to indicate 'COMPLETE'
    // update $RESTART_DIR with each directory completed (since we can't know these in advance...)
    // process differs from DB in that for incomplete processes we need data from every related checkpoint to built RESTART_DIR array.
    // $row['text'] contains checkpoint text
    // $row['data'] contains mapping array checkpointed
    $load_map = FALSE;
    $checkpoint =  explode(':', $row['text']);
    $table_constant = $checkpoint[0];                                           // process name if first word of checkpoint text
    $table_indicator = substr($table_constant, 0, 6);                           // extract prefix - if TABLE_ - this is a table checkpoint
    $process_indicator = substr($table_constant, 0, 8);                         // extract prefix - if PROCESS_ - this is a process checkpoint
    $process_name_uc = substr($table_constant, 8);                              // uppercase process constant for messages
    $process_name = strtolower($process_name_uc);                               // lowercase process constant for messages
    $table_status = $checkpoint[1];

    if ($first) {
      // first record read...
      if (($table_indicator == 'TABLE_') || ($process_indicator == 'PROCESS_') || ($RESTART[$table_constant] == '0')) {
        // record for a table - or record for a process - or we have specifically noted in RESTART array
        // other log records like process initiation (failed restart) or log data might be encountered first
        if ($table_status == 'ERROR') {
          // last execution was terminated by error
          $debug_trace[] = "Checkpoint Restart: Last execution ended with ERROR in $process_name";
          $first = FALSE;
          $first_error = $table_constant;
        } else {
          if ($table_status == 'QUIESCE') {
            $lastkey = $checkpoint[4];
            $debug_trace[] = "Checkpoint Restart: Last execution ended with Quiesce in process $process_name at key $lastkey";
          } elseif ($table_status == 'COPIED') {
            $lastkey = $checkpoint[4];
            $debug_trace[] = "Checkpoint Restart: Last execution ended unexpectedly in table $process_name after key $lastkey";
          } elseif ($table_status == 'COMPLETE') {
            $debug_trace[] = "Checkpoint Restart: Last execution ended unexpectedly after completing process $process_name";
            $lastkey = 'COMPLETE';
          }
          $RESTART[$table_constant] = $lastkey;
          if ($row['data'] != '') {
            $load_map = TRUE;
          }
          $first = FALSE;
        }
      } // endif a table we care about
    } else {
      // second and subsequent records... may be earlier checkpoints for same table process - or previous table.
      // if this is for a table or process... and either RESTART doesn't contain key - or key = 0...
      if ((($table_indicator == 'TABLE_') || ($process_indicator == 'PROCESS_')) && (!array_key_exists($table_constant, $RESTART) || ($RESTART[$table_constant] == '0'))) {
        // first record for a table/process we haven't seen yet
        if ($table_status == 'COMPLETE') {
          // good checkpoint
          $RESTART[$table_constant] = $table_status;
          if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Restart: Process: {$table_constant} Status: {$table_status} added to RESTART table"; }
          if ($row['data'] != '') {
            $load_map = TRUE;
          }
        } elseif (($table_status == 'ERROR') && ($first_error == $table_constant)) {
          // multiple errors for same process - ignore and continue
        } else {
          // shouldn't happen... first record seen for any table except first encountered should be COMPLETE...
          $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_checkpoint'], $table_constant, $table_status, $RESTART[$table_constant]);
          gallery_merge_error('init', 'PROCESS_RESTART', 'error', $error_message);
          break;
        }
      } elseif ((array_key_exists($table_constant, $RESTART)) && ($RESTART[$table_constant] != 'COMPLETE')) {
        // we've seen this process before - but it didn't complete... we need some info (unique to files restart)
        if ($table_status == 'COPIED') {
          // good checkpoint
          // $SOURCE['album_filepath'] is not set yet... array will be updated later
          $RESTART_DIR[$checkpoint[4]] = $checkpoint[4];
          if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Restart: {$SOURCE['album_filepath']}{$checkpoint[4]} added to RESTART DIRECTORY table"; }
        } elseif ($table_status == 'QUIESCE') {
          // previous QUIESCE for same process, no data needed, continue
        } elseif ($table_status == 'ERROR') {
          // previous ERROR for same process, no data needed, continue
        } else {
          // shouldn't happen... unknown record type for a process we care about...
          $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_checkpoint'], $table_constant, $table_status, $RESTART[$table_constant]);
          gallery_merge_error('init', 'PROCESS_RESTART', 'error', $error_message);
          break;
        }
      }
    }
  }
  cpg_db_free_result($result);
}


// Checkpoint bypass
function gallery_merge_checkpoint_bypass() {
  // process any bypass requests
  // processes AFTER restart to allow any maps to be reloaded.
  global $CONFIG, $RESTART, $BYPASS_WARNING, $BYPASS_ERROR, $BYPASS, $BYPASS_DIR, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  $prohibited = array('TABLE_USERGROUPS', 'TABLE_USERS', 'TABLE_CATEGORIES', 'TABLE_ALBUMS', 'TABLE_PICTURES');

  // process warning bypasses
  foreach ($BYPASS_WARNING AS $bypass_process => $bypass_action) {
    if ($bypass_action != 'bypass') {
      // action MUST be 'bypass' or we ignore...
      $debug_trace[] = "Invalid WARNING bypass request for '$bypass_process' with action '$bypass_action' - IGNORED";
      continue;
    }
    $col2 = sprintf($lang_plugin_gallery_merge['bypass_warning_requested'], $bypass_process);
    gallery_merge_output_status('blank', $col2);
  }
  // process error bypasses
  foreach ($BYPASS_ERROR AS $bypass_process => $bypass_action) {
    if ($bypass_action != 'bypass') {
      // action MUST be 'bypass' or we ignore...
      $debug_trace[] = "Invalid ERROR bypass request for '$bypass_process' with action '$bypass_action' - IGNORED";
      continue;
    }
    $col2 = sprintf($lang_plugin_gallery_merge['bypass_error_requested'], $bypass_process);
    gallery_merge_output_status('blank', $col2);
  }
  // process step bypasses
  foreach ($BYPASS AS $bypass_process => $bypass_action) {
    if ($bypass_action != 'bypass') {
      // action MUST be 'bypass' or we ignore...
      $debug_trace[] = "Invalid STEP bypass request for '$bypass_process' with action '$bypass_action' - IGNORED";
      continue;
    }
    if (in_array($bypass_process, $prohibited)) {
      // we can't bypass these tables - too many cascade failures in mapping and references...
      $debug_trace[] = "Invalid STEP bypass request for '$bypass_process' - unsupported for this table/process - IGNORED";
      continue;
    }
    if (!key_exists($bypass_process, $RESTART)) {
      // no match in restart array - add as bypass
      $RESTART[$bypass_process] = 'BYPASS';
      $col2 = sprintf($lang_plugin_gallery_merge['bypass_step_requested'], $bypass_process);
      gallery_merge_output_status('blank', $col2);
    } elseif ($RESTART[$bypass_process] == 'COMPLETE') {
      // step already completed - no need for bypass
      $debug_trace[] = "Bypass requested for '$bypass_process' - step already marked COMPLETE";
    } elseif ($RESTART[$bypass_process] == '-1') {
      // step already completed - but was backed out
      $RESTART[$bypass_process] = 'BYPASS';
      $col2 = sprintf($lang_plugin_gallery_merge['bypass_step_requested_backout'], $bypass_process);
      gallery_merge_output_status('blank', $col2);
    } else {
      // step initiated, but didn't complete... allow bypass - but doth protest with messages!
      $col2 = sprintf($lang_plugin_gallery_merge['bypass_step_requested_failure'], $bypass_process, $RESTART[$bypass_process]);
      gallery_merge_output_status('blank', $col2);
      $RESTART[$bypass_process] = 'BYPASS';
    }
  }
  $bypass_list = array('warning' => $BYPASS_WARNING, 'error' => $BYPASS_ERROR, 'step' => $BYPASS);

  // process directory bypasses
  if (is_array($BYPASS_DIR)) {
    foreach ($BYPASS_DIR AS $bypass_process) {
      if (key_exists($bypass_process, $RESTART_DIR)) {
        // directory already completed - no need for bypass
        $debug_trace[] = "Bypass requested for '$bypass_process' - directory already marked COMPLETE";
      } else {
        $col2 = sprintf($lang_plugin_gallery_merge['bypass_dir_requested'], $bypass_process);
        gallery_merge_output_status('blank', $col2);
      }
    }
  $bypass_list['directory'] = $BYPASS_DIR;
  }
  return $bypass_list;
}


function gallery_merge_checkpoint_restart_table_update(&$value, $key, $prefix) {
  // called via array_walk to update array adding prefix to each element
  // used to update RESTART_DIR once SOURCE album_filepath is known
  $value = $prefix.$value;
}


// Process given table on restart?
function gallery_merge_checkpoint_restart_table($table_constant) {
  // called to determine if a given table needs to be processed on restart
  // added 'bypass' capability - if $CONFIG["plugin_gallery_merge_bypass_$table_constant"] is set - bypass this step
  global $CONFIG, $RESTART, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $restart_flag;
  // see if this is a restart - and if table/process is already complete
  if (($restart_flag == 'RESTART') && (key_exists($table_constant, $RESTART)) && ($RESTART[$table_constant] == 'COMPLETE')) {
    // restart - and this table shows complete...
    if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Restart: Status for table $table_constant: {$RESTART[$table_constant]} - COMPLETE"; }
    return FALSE;
  }
  // no restart - or table constant != COMPLETE... check for bypass
  if ((key_exists($table_constant, $RESTART)) && ($RESTART[$table_constant] == 'BYPASS')) {
    // bypass specified
    if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Restart: Status for table $table_constant: {$RESTART[$table_constant]} - BYPASS"; }
    return FALSE;
  }
  if ($debug_level >= 3) { $debug_trace[] = "Checkpoint Restart: Status for table $table_constant: {$RESTART[$table_constant]} - PROCESS"; }
  return TRUE;
}


// *******************
// Gallery Connections
// *******************

// Obtain source DB connecton information...
function gallery_merge_obtain_source_config() {
  // based on local or remote gallery request - obtain the info we need to connect to DB
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;

  // determine if local or remote gallery by CONFIG
  if ($CONFIG['plugin_gallery_merge_source_local'] == '1') {
    // *********************
    // local gallery request
    // *********************
    $col1 = sprintf("$restart_flag {$lang_plugin_gallery_merge['request_merge_local']}", $CONFIG['plugin_gallery_merge_source_filepath'], $CONFIG['gallery_name']);
    gallery_merge_output_status($col1, 'blank', 'heading');
    // validate path to gallery for local
    $debug_trace[] = "Source Filepath: {$CONFIG['plugin_gallery_merge_source_filepath']}";
    if (strlen($CONFIG['plugin_gallery_merge_source_filepath']) <= 1) {
      // source filepath not valid.. too short
      $error_message = $lang_plugin_gallery_merge['e_invalid_filepath_length'];
      gallery_merge_error('init', 'PROCESS_CONNECT', 'error', $error_message);
    } else {
      $source_config_filepath = "{$CONFIG['plugin_gallery_merge_source_filepath']}/include/config.inc.php";
      $debug_trace[] = "Source Config Filepath: $source_config_filepath";
      if (file_exists($source_config_filepath)) {
        // found config file..
        // cpg uses $CONFIG for all config data... Using $SOURCE for consistency
        $source_config_contents = file_get_contents($source_config_filepath);
        $source_config_contents = str_replace('$CONFIG', '$SOURCE', $source_config_contents);
        $source_config_contents = str_replace('<?php', '', $source_config_contents);
        $source_config_contents = str_replace('?>', '', $source_config_contents);
        eval ($source_config_contents);
        $debug_trace[] = "Source Connection String (dbserver): {$SOURCE['dbserver']}";
        $debug_trace[] = "Source User: {$SOURCE['dbuser']} - Source DB: {$SOURCE['dbname']}";
        $debug_trace[] = "Source Table Prefix: {$SOURCE['TABLE_PREFIX']}";
        unset ($source_config_contents);
      } else {
        // source filepath not valid.. cannot locate coppermine config
        $error_message = $lang_plugin_gallery_merge['e_unable_locate_config'];
        gallery_merge_error('init', 'PROCESS_CONNECT', 'error', $error_message);
      }
    }
  } else {
    // **********************
    // remote gallery request
    // **********************
    $col1 = sprintf("$restart_flag {$lang_plugin_gallery_merge['request_merge_remote']}", $CONFIG['plugin_gallery_merge_source_dbserver'], $CONFIG['gallery_name']);
    gallery_merge_output_status($col1, 'blank', 'heading');
    // move variables to $SOURCE for consistency...
    $SOURCE['dbserver'] = $CONFIG['plugin_gallery_merge_source_dbserver'];
    $SOURCE['dbuser'] = $CONFIG['plugin_gallery_merge_source_dbuser'];
    $SOURCE['dbpass'] = $CONFIG['plugin_gallery_merge_source_dbpass'];
    $SOURCE['dbname'] = $CONFIG['plugin_gallery_merge_source_dbname'];
    $SOURCE['TABLE_PREFIX'] = $CONFIG['plugin_gallery_merge_source_table_prefix'];
    // provided database connection info will be validated in another function
  }
}


// Connect to Source DB...
function gallery_merge_connect_source() {
  // establish SOURCE database connection with provided information (global $SOURCE)
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;

  // For local or remote... Open database and set up database variables
  $col1 = $lang_plugin_gallery_merge['open_source_db'];
  gallery_merge_output_status($col1, 'blank');
  if (($CONFIG['dbserver'] == $SOURCE['dbserver']) &&
      ($CONFIG['dbname'] == $SOURCE['dbname']) &&
      ($CONFIG['TABLE_PREFIX'] == $SOURCE['TABLE_PREFIX'])) {
    // source and target are the same... not allowed
    $error_message = $lang_plugin_gallery_merge['e_source_target_same'];
    gallery_merge_error('init', 'PROCESS_CONNECT', 'error', $error_message);
  } else {
    // Using 'alternate' added CPG function to connect - as CPG functions don't accept parms for connection
    $SOURCE['LINK_ID'] = cpg_db_connect_alt($SOURCE);
    if (!$SOURCE['LINK_ID']) {
      // unable to connect to source DB
      $error_message = $lang_plugin_gallery_merge['e_unable_connect_source'];
      gallery_merge_error('init', 'PROCESS_CONNECT', 'error', $error_message);
    } else {
      // open connection!!
      // we'll continue to use CPG functions for accessing TARGET and SOURCE database from here, passing LINK_ID for SOURCE (TARGET LINK_ID is assumed)
      $debug_trace[] = "Source Database {$CONFIG['plugin_gallery_merge_source_dbname']} Open";
    }
  }
}


// ******************
// Backup and Restore
// ******************

// Backup requested file...
function gallery_merge_backup_table($table_constant, $starttime) {
  // backup the requested table to a new table in DB
  // timestamp in restart_data for PROCESS_BACKUP:COMPLETE matches backup to execution for recovery
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;
  static $length_error_issued = FALSE;
  static $length_warning_issued = FALSE;
  $table_name_uc = substr($table_constant, 6);                                  // uppercase table constant for messages
  $table_name = strtolower($table_name_uc);                                     // lowercase table constant for messages
  $timestamp = str_replace(' ', '_', $starttime);                               // remove spaces in timestamp
  $backup_table = "{$CONFIG['TABLE_PREFIX']}{$CONFIG['plugin_gallery_merge_backout_prefix']}{$timestamp}_{$table_name}";
  if (strlen($backup_table) < 65) {
    if ($plugin_gallery_merge_execute) {
      $result = cpg_db_query("CREATE TABLE `$backup_table` LIKE {$CONFIG[$table_constant]}");
      $result = cpg_db_query("INSERT INTO  `$backup_table` SELECT * FROM {$CONFIG[$table_constant]}");
      $col2 = sprintf($lang_plugin_gallery_merge['backup_complete'], $table_name_uc);
      gallery_merge_output_status('blank', $col2);
    } else {
      $col2 = sprintf($lang_plugin_gallery_merge['backup_bypassed'], $table_name_uc);
      gallery_merge_output_status('blank', $col2);
    }
  } else {
    // table name exceeds 64 characters - create table will fail (adjust 'plugin_gallery_merge_backout_prefix' in cpg_config to shorter value...)
    if ($plugin_gallery_merge_execute) {
      if (!$length_error_issued) {
        $length_error_issued = TRUE;
        $error_message = $lang_plugin_gallery_merge['e_tablename_length'];
        gallery_merge_error('db', $table_constant, 'error', $error_message);
      }
    } else {
      if (!$length_warning_issued) {
        $length_warning_issued = TRUE;
        $warning_message = $lang_plugin_gallery_merge['w_tablename_length'];
        gallery_merge_error('db', $table_constant, 'warning', $warning_message);
      }
    }
  }
  if ($debug_level >= 3) { $debug_trace[] = "Table Backup: Backup for $table_constant: {$CONFIG[$table_constant]} => $backup_table complete"; }
}


// Restore requested file...
function gallery_merge_restore_table($table_constant, $starttime) {
  // backup the requested table to a new table in DB
  // timestamp in restart_data for PROCESS_BACKUP:COMPLETE matches backup to execution for recovery
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_backout_execute;
  $table_name_uc = substr($table_constant, 6);                                  // uppercase table constant for messages
  $table_name = strtolower($table_name_uc);                                     // lowercase table constant for messages
  $timestamp = str_replace(' ', '_', $starttime);                               // remove spaces in timestamp
  $backup_table = "{$CONFIG['TABLE_PREFIX']}plugin_gallery_merge_bkup_{$timestamp}_{$table_name}";
  $backup_result = cpg_db_query("SELECT COUNT( * ) FROM `{$backup_table}`");
  $backup_count = cpg_db_result($backup_result, 0, 0);
  cpg_db_free_result($backup_result);
  // table must exist or we would have gotten error (0 rows may be valid - so can't base on that...)
  if ($plugin_gallery_merge_backout_execute) {
    $result = cpg_db_query("TRUNCATE TABLE {$CONFIG[$table_constant]}");
    $result = cpg_db_query("INSERT INTO {$CONFIG[$table_constant]} SELECT * FROM `{$backup_table}`");
    $col2 = sprintf($lang_plugin_gallery_merge['restore_complete'], $table_name_uc);
    gallery_merge_output_status('blank', $col2);
    gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:BACKOUT");
  } else {
    $col2 = sprintf($lang_plugin_gallery_merge['restore_bypassed'], $table_name_uc);
    gallery_merge_output_status('blank', $col2);
  }
  if ($debug_level >= 3) { $debug_trace[] = "Table Restore: Restore for $table_constant from $starttime backup ($backup_table) complete"; }
}


// **************
// Error Handling
// **************

// Debug SQL Output
function gallery_merge_error($mode, $constant, $level, $message, $nobypass = '') {
  global $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_error, $plugin_gallery_merge_error_text;
  global $plugin_gallery_merge_warning, $plugin_gallery_merge_warning_text;
  global $plugin_gallery_merge_info, $plugin_gallery_merge_info_text;
  global $BYPASS_WARNING, $BYPASS_ERROR;
  // common warning and error handling routine
  // $mode = db|files
  // $constant = table or process constant
  // $level = 'info'|'warning'|'error'
  // appropriate info/warning/error flag and text set. checkpoint issued to record for restart/review
  $status = strtoupper($level);         // upper case for checkpoint
  if ($status == 'ERROR') {
    // error processing
    if (key_exists($constant, $BYPASS_ERROR) && ($nobypass == '')) {
      // error bypass requested
      $status = 'INFO';
      $message = $lang_plugin_gallery_merge['bypass_error_prefix'].$message;
    } else {
      $plugin_gallery_merge_error = TRUE;
      $plugin_gallery_merge_error_text[] = $message;
      $data = $plugin_gallery_merge_error_text;
    }
    // warning processing
  } elseif ($status == 'WARNING') {
    if (key_exists($constant, $BYPASS_WARNING) && ($nobypass == '')) {
      // warning bypass requested
      $status = 'INFO';
      $message = $lang_plugin_gallery_merge['bypass_warning_prefix'].$message;
    } else {
    // warning processing
    $plugin_gallery_merge_warning = TRUE;
    $plugin_gallery_merge_warning_text[] = $message;
    $data = $plugin_gallery_merge_warning_text;
    }
  }
  // above bypasses may have changed error or warning to info - so check separately (also catch all for any other status)
  if ($status != 'ERROR' && $status != 'WARNING') {
    // info processing
    $plugin_gallery_merge_info = TRUE;
    $plugin_gallery_merge_info_text[] = $message;
    $data = $plugin_gallery_merge_info_text;
  }
  if ($mode != 'init') {
    // checkpoint the info/warning/error if we have already initialized checkpointing...
    // early errors like unable to connect to DB or checkpoint restart errors not logged
  gallery_merge_checkpoint($mode, 'checkpoint', '', "$constant:$status", $data);
  }
}


function gallery_merge_error_output($level) {
  global $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_error, $plugin_gallery_merge_error_text;
  global $plugin_gallery_merge_warning, $plugin_gallery_merge_warning_text;
  global $plugin_gallery_merge_info, $plugin_gallery_merge_info_text;
  global $plugin_gallery_merge_modetext;
  // common warning and error handling output routine
  // $level = 'info'|'warning'|'error'
  // appropriate info/warning/error messages outputted in message box.
  $status = strtolower($level);         // lower case for msg_box
  $data = '';
  if ($status == 'error') {
    // error processing
    foreach ($plugin_gallery_merge_error_text as $message) {
      // extract messages... should really only be one for error status - but just in case
      $data .= "$message<br>";
    }
  } elseif ($status == 'warning') {
    // warning processing
    foreach ($plugin_gallery_merge_warning_text as $message) {
      // extract messages... multiple warnings possible
      $data .= "$message<br>";
    }
  } elseif ($status == 'info') {
    // info processing
    foreach ($plugin_gallery_merge_info_text as $message) {
      // extract messages... multiple info messages possible
      $data .= "$message<br>";
    }
  } elseif ($status == 'success') {
    // success processing
    // construct message
    $data = "{$lang_plugin_gallery_merge[$plugin_gallery_merge_modetext]} {$lang_plugin_gallery_merge['completed']}";
  } elseif ($status == 'success_warn') {
    // success processing - but with warnings
    // construct message
    $data = "{$lang_plugin_gallery_merge[$plugin_gallery_merge_modetext]} {$lang_plugin_gallery_merge['completed']} with warnings";
    $status = 'warning';
  }
  msg_box('', $data, '', '', $status);
}


// ***************
// Output Routines
// ***************

// Debug SQL Output
function gallery_merge_debug_sql($debug_text, $close = FALSE) {
  // SQL is highest volume of debug output...
  // For simulation - if $debug_level >= 2, add to $debug_trace array for normal sysout
  // For execution - in ALL cases, write SQL to a log file so it is retained (usable to import into MySQL)
  //   filename:  logs/log_{starttime}.sql
  //              starttime is time gallery merge started. restarts of process will have same starttime and append to same file
  //              timestamp is unique to this run - generated on first call - and placed in initial comment
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;    
  global $plugin_gallery_merge_execute, $starttime;
  static $sql_filehandle = FALSE;
  static $sql_filepath = '';

  if ($plugin_gallery_merge_execute || $CONFIG['plugin_gallery_merge_sim_output']) {
    // executing merge or override set
    if (!$sql_filehandle) {
      // first time through - calculate timestamps and open log file
      $file_starttime = (str_replace(' ', '_', $starttime));
      $file_starttime = (str_replace(':', '-', $file_starttime));
      $ts = date("Y-m-d H:i:s");
      $sql_filepath = $CONFIG['cpg_filepath']."/plugins/gallery_merge/logs/log_{$file_starttime}.sql";
      $sql_filehandle = fopen($sql_filepath, 'a');
      if (!$sql_filehandle) {
        $debug_trace[] = "Log file $sql_filepath failed to open - reverting to in memory debug for SQL";
        $sql_filehandle = 'failed';
      } else {
        $debug_trace[] = "Log file $sql_filepath open - SQL will be logged to this file.";
        fwrite($sql_filehandle, "# Log file opened at $ts\r\n");
      }
    }
    // write to file...
    if ($sql_filehandle != 'failed') {
      // ok to write...
      fwrite($sql_filehandle, "$debug_text;\n");
    } else {
      $debug_trace[] = $debug_text;
    }
  } else {
    // not executing... base on $debug_level and write to $debug_trace
    if ($debug_level >= 2) {
      $debug_trace[] = 'SQL: '.$debug_text;
    }
  } // end else

  if (($close) && ($sql_filehandle) && ($sql_filehandle != 'failed')) {
    fwrite($sql_filehandle, "$debug_text;\n");
    fclose($sql_filehandle);
  }
}                                                                   


// Output status display
function gallery_merge_output_status($col1, $col2, $type = '') {
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute, $plugin_gallery_merge_execute_files, $plugin_gallery_merge_backout_execute, $starttime;
  static $message_filehandle = FALSE;
  static $debug_filehandle = FALSE;
  static $message_filepath = '';
  static $debug_filepath = '';

  // file close request?
  if ($type == 'close') {
    if (($message_filehandle) && ($message_filehandle != 'failed')) {
      fwrite($message_filehandle, "<!-- Message Log file closed -->\n");
      fclose($message_filehandle);
    }
    if (($debug_filehandle) && ($debug_filehandle != 'failed')) {
      fwrite($debug_filehandle, "<!-- Debug Log file closed -->\n");
      fclose($debug_filehandle);
    }
  return;
  }

  // construct message html
  if ($type == 'heading') {
    // $col2 unused for heading... $col1 shouldn't be blank...
    $message = <<< EOT
      <tr>
        <td class="tablef" colspan="3">
          {$col1}
        </td>
      </tr>
EOT;
  } else {
    if ($col1 == 'blank') {
      // normal status messages during process have blank for $col1
      $col1 = "&nbsp;";
    }
    if ($col2 == 'blank') {
      // some non-heading status messages leave $col2 blank.
      $col2 = "&nbsp;";
    }
    $message = <<< EOT
      <tr>
        <td valign="top">
          {$col1}
        </td>
        <td>
          {$col2}
        </td>
      </tr>
EOT;
  }
  // write out message
  echo $message;
  // need to log to file?
  if ($plugin_gallery_merge_execute || $plugin_gallery_merge_execute_files || $plugin_gallery_merge_backout_execute || $CONFIG['plugin_gallery_merge_sim_output']) {
    // execute mode for db, files, or db backout, or override... use external logging
    if ($type == 'debug') {
      $filehandle = 'debug_filehandle';
      $filepath = 'debug_filepath';
      $filenamepref = 'debug';
    } else {
      $filehandle = 'message_filehandle';
      $filepath = 'message_filepath';
      $filenamepref = 'message';
    }
    if (!$$filehandle) {
      // first time through for this type - calculate timestamps and open log file
      $file_starttime = (str_replace(' ', '_', $starttime));
      $file_starttime = (str_replace(':', '-', $file_starttime));
      $ts = date("Y-m-d H:i:s");
      $$filepath = $CONFIG['cpg_filepath']."/plugins/gallery_merge/logs/{$filenamepref}_{$file_starttime}.html";
      $$filehandle = fopen($$filepath, 'a');
      if (!$$filehandle) {
        $debug_trace[] = "Log file {$$filepath} failed to open - No external logging for $filenamepref";
        $$filehandle = 'failed';
      } else {
        $debug_trace[] = "Log file {$$filepath} open - $filenamepref will be logged to this file.";
        fwrite($$filehandle, "<!-- $filenamepref log file opened at $ts -->\n");
      }
    }
    // write to file...
    if ($$filehandle != 'failed') {
      // ok to write...
      fwrite($$filehandle, "$message\n");
    }
  }
}


// *************
// Documentation
// *************

// process database doc request
function gallery_merge_db_doc($table_name) {
  global $CONFIG, $lang_plugin_gallery_merge;
  // map start and end labels in cpg doc for each table description
  $doc_map = array('db_albums' => 'db_banned',
                   'db_banned' => 'db_bridge',
                   'db_bridge' => 'db_categories',
                   'db_categories' => 'db_categorymap',
                   'db_categorymap' => 'db_comments',
                   'db_comments' => 'db_config',
                   'db_config' => 'db_dict',
                   'db_dict' => 'db_ecards',
                   'db_ecards' => 'db_exif',
                   'db_exif' => 'db_favpics',
                   'db_favpics' => 'db_filetypes',
                   'db_filetypes' => 'db_hit_stats',
                   'db_hit_stats' => 'db_languages',
                   'db_languages' => 'db_pictures',
                   'db_pictures' => 'db_plugins',
                   'db_plugins' => 'db_sessions',
                   'db_sessions' => 'db_temp_messages',
                   'db_temp_messages' => 'db_usergroups',
                   'db_usergroups' => 'db_users',
                   'db_users' => 'db_votes',
                   'db_votes' => 'db_vote_stats',
                   'db_vote_stats' => 'db_end');
  // address names where doc tags don't match table constant...
  if ($table_name == 'catmap') {
    $docstart = 'db_categorymap';
  } elseif ($table_name == 'language') {
    $docstart = "db_languages";
  } else {
    // the rest of the names match
    $docstart = "db_{$table_name}";
  }
  $docend = $doc_map[$docstart];
  return cpg_display_help("f=dev_database.htm&amp;as={$docstart}&amp;ae={$docend}", '400', '200');
}


// ******************
// Merge DB Functions
// ******************

// compare SOURCE and TARGET table structure - report difference, account for obsolete keys (if any)
function gallery_merge_version_check($table_constant, $target_version) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;

  if ($CONFIG['plugin_gallery_merge_source_local'] == '1') {
    // local gallery request - so we can validate source version/maintenance
    $source_init_filepath = "{$CONFIG['plugin_gallery_merge_source_filepath']}/include/init.inc.php";
    $source_version = '?.?.??';
    if (file_exists($source_init_filepath)) {
      // found init file..
      $filehandle = fopen($source_init_filepath, 'r');
      $found_version = FALSE;
      while ($line = fgets($filehandle)) {
        if (strpos($line, 'COPPERMINE_VERSION')) {
          // found the line we want...
          $found_version = TRUE;
          break;
        }
      }
      fclose($filehandle);
      if ($found_version) {
        // this should be the line we found...
        // define('COPPERMINE_VERSION', '1.5.24');
        $version_info = explode("'", $line);
        $source_version = $version_info[3];
        if ($target_version == $source_version) {
          $col2 = sprintf($lang_plugin_gallery_merge['version_match'], $source_verson);
          gallery_merge_output_status('blank', $col2);
        } else {
          $col2 = sprintf($lang_plugin_gallery_merge['version_mismatch'], $source_version, $target_version);
          gallery_merge_output_status('blank', $col2);
          if ($plugin_gallery_merge_execute) {
            // abort if execution
            $error_message = $lang_plugin_gallery_merge['e_version_mismatch'];
            gallery_merge_error('db', $table_constant, 'error', $error_message);
          } else {
            // warn if simulation
            $warning_message = $lang_plugin_gallery_merge['w_version_mismatch']."<br />- ".$lang_plugin_gallery_merge['w_version_upgrade']." ".$lang_plugin_gallery_merge['w_rerun_simulation'];
            gallery_merge_error('db', $table_constant, 'warning', $warning_message);
          }
        }
      } else {
        // version not found in init.inc.php
        $col2 = sprintf($lang_plugin_gallery_merge['version_notfound'], $target_version);
        gallery_merge_output_status('blank', $col2);
        if (!$plugin_gallery_merge_execute) {
          // warn if simulation
          $warning_message = $lang_plugin_gallery_merge['w_version_unknown'];
          gallery_merge_error('db', $table_constant, 'warning', $warning_message);
        }
      }
    } else {
      // cannot locate coppermine init.inc.php...
      $col2 = sprintf($lang_plugin_gallery_merge['version_noinit'], $target_version);
      gallery_merge_output_status('blank', $col2);
      if (!$plugin_gallery_merge_execute) {
        // warn if simulation
        $warning_message = $lang_plugin_gallery_merge['w_version_unknown'];
        gallery_merge_error('db', $table_constant, 'warning', $warning_message);
      }
    }
  } else {
    // source not local - cannot verify...
    $col2 = $lang_plugin_gallery_merge['version_remote'];
    gallery_merge_output_status('blank', $col2);
  }
  $debug_trace[] = "Target version: $target_version - Source version: $source_version";
}


// compare SOURCE and TARGET table structure - report difference, account for obsolete keys (if any)
function gallery_merge_compare_structure($table_constant, $source_table, $target_table, $obsolete_keys, &$skip_keys) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;

  $structure_warning_issued = FALSE;
  $structure_error_issued = FALSE;
  $target_array = array();
  $where = '';
  if (!empty($obsolete_keys)) {
    // process any obsolete keys first...
    $first = TRUE;
    $where = 'WHERE `Field` IN (';
    foreach ($obsolete_keys AS $key) {
      if (!$first) {
        $where .= ', ';
      } else {
        $first = FALSE;
      }
      $where .= "'$key'";
    }
    $where .= ')';
    $source_result = cpg_db_query("SHOW COLUMNS FROM {$source_table} $where", $SOURCE['LINK_ID']);
    // see if any obsolete fields where found...
    $source_numrows = cpg_db_num_rows($source_result);
    if ($source_numrows > 0) {
      // have some obsolete fields found... get same info from target
      $target_result = cpg_db_query("SHOW COLUMNS FROM {$target_table} $where");
      $target_numrows = cpg_db_num_rows($target_result);
      if ($target_numrows > 0) {
        // both source and target have obsolete fields... match them up...
        while ($row = cpg_db_fetch_row($target_result)) {
          $target_array[$row['Field']] = $row;
        }
        while ($row = cpg_db_fetch_row($source_result)) {
          if (array_key_exists($row['Field'], $target_array)) {
            // field in both source and target
            $col2 = sprintf($lang_plugin_gallery_merge['structure_obsolete_both'], $source_table, $row['Field']);
            gallery_merge_output_status('blank', $col2);
          } else {
            // field in source only
            $col2 = sprintf($lang_plugin_gallery_merge['structure_obsolete_source'], $source_table, $row['Field']);
            gallery_merge_output_status('blank', $col2);
            $skip_keys[] = $row['Field'];
          }
        }
      } else {
        // obsolete fields in source only - issue messages and add to skip keys...
        while ($row = cpg_db_fetch_row($source_result)) {
          $col2 = sprintf($lang_plugin_gallery_merge['structure_obsolete_source'], $source_table, $row['Field']);
          gallery_merge_output_status('blank', $col2);
          $skip_keys[] = $row['Field'];
        }
      }
      cpg_db_free_result($target_result);
    } else {
      // no obsolete fields found in source... continue on...
    }
    cpg_db_free_result($source_result);
  } // end if obsolete keys specified

  // check remaining structure...
  // same approach as elsewhere... load target results into array... process source results and compare...
  $target_array = array();
  $where_not = str_replace('WHERE `Field` IN (', 'WHERE `Field` NOT IN (', $where);
  $target_result = cpg_db_query("SHOW COLUMNS FROM {$target_table} $where_not");
  while ($row = cpg_db_fetch_row($target_result)) {
    $target_array[$row['Field']] = $row;
  }
  cpg_db_free_result($target_result);
  // get source columns
  $source_result = cpg_db_query("SHOW COLUMNS FROM {$source_table} $where_not", $SOURCE['LINK_ID']);
  while ($row = cpg_db_fetch_row($source_result)) {
    // compare to target array - message on any from source NOT in target
    if (!array_key_exists($row['Field'], $target_array)) {
      // field in source only - oops...
      $col2 = sprintf($lang_plugin_gallery_merge['structure_not_matched'], $source_table, $row['Field']);
      gallery_merge_output_status('blank', $col2);
      // issue  - error if execution - warning if simualte...
      if ($plugin_gallery_merge_execute) {
        if (!$structure_error_issued) {
          $structure_error_issued = TRUE;
          $error_message = $lang_plugin_gallery_merge['e_table_structure_mismatch'];
          gallery_merge_error('db', $table_constant, 'error', $error_message);
        }
      } else {
        if (!$structure_warning_issued) {
          $structure_warning_issued = TRUE;
          $warning_message = $lang_plugin_gallery_merge['w_table_structure_mismatch'];
          gallery_merge_error('db', $table_constant, 'warning', $warning_message);
        }
      }

    }
  }
  cpg_db_free_result($source_result);
  return;
}


// compare SOURCE and TARGET arrays - report difference
function gallery_merge_compare_arrays($source_row, $target_row, $skip_keys = array(), $partial_data = FALSE) {
  if (!$partial_data) {
    // all fields from source and target available - compare using source as primary (we want any fields in source not in target too!
    $differences = array_diff_assoc($source_row, $target_row);
  } else {
    // only some of the target fields are available (reduced memory usage) - compare using target as primary - we know we are missing fields
    $differences = array_diff_assoc($target_row, $source_row);
  }
  $difference = '';
  foreach ($differences AS $key => $value) {
    if (!in_array($key, $skip_keys)) {
      // if not a key to bypass
      $difference .= "&nbsp;&nbsp;&nbsp;&nbsp;Difference: {$lang_plugin_gallery_merge['source_uc']} '$key' = $value ({$lang_plugin_gallery_merge['target_uc']} = {$target_row[$key]}) <br />";
    }
  }
  return $difference;
}


// generate insert SQL
function gallery_merge_generate_sql($table, &$source_row, $skip_keys, $zero_keys, $insert_ignore) {
  // sample SQL:
  // $merge_sql = "INSERT INTO `cpg_usergroups`(`group_name`, `group_quota`, `has_admin_access`, `can_rate_pictures`, `can_send_ecards`,
  //                                            `can_post_comments`, `can_upload_pictures`, `can_create_albums`, `pub_upl_need_approval`,
  //                                            `priv_upl_need_approval`, `access_level`)
  //                      VALUES ({$source_row['group_name']},{$source_row['group_quota']},{$source_row['has_admin_access']},{$source_row['can_rate_pictures']},{$source_row['can_send_ecards']},
  //                              {$source_row['can_post_comments']},{$source_row['can_upload_pictures']},{$source_row['can_create_albums']},{$source_row['pub_upl_need_approval']},
  //                              {$source_row['priv_upl_need_approval']},{$source_row['access_level']})";
  $insert = "INSERT $insert_ignore INTO `$table` ";
  $fields = '(';
  $values = ' VALUES (';
  $end = ')';
  $first = TRUE;
  foreach ($source_row AS $key => $value) {
    if (!in_array($key, $skip_keys)) {
      // if not a key to bypass
      if (in_array($key, $zero_keys)) {
        $source_row[$key] = 0;
      }
      if ($first) {
        $first = FALSE;    // reset flag..
      } else {
        $fields .= ', ';   // names/values after first need a ,
        $values .= ', ';
      }
      $fields .= "`$key`";
      $values .= "'".cpg_db_real_escape_string($value)."'";
    }
  }
  $sql = $insert.$fields.$end.$values.$end;
  return $sql;
}


// initialize translation arrays for mapping pictures
function gallery_merge_map_prime_categories($restart_key = 0) {
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $source_mapping_users,      $source_mapping_users_key;                 // access to users mapping table and key
  global $source_mapping_categories, $source_mapping_categories_key;            // access to categories mapping table and key
  // user albums are in categories > 10000* - that do not exist in the categories table. Their value is typically 10000* + user_id value.
  // prime array with the keys we care about - possible categories for those users we've mapped
  // The actual increment for user categories is defined in FIRST_USER_CAT in init.inc.php
  if ($restart_key > '0') {
    // we've already been here on previous execution..
    return;
  }
  // process users to set up possible user categories
  foreach ($source_mapping_users AS $user) {
    $usergallery_id = FIRST_USER_CAT + $user['user_id'];
    $target_usergallery_id = $FIRST_USER_CAT + $user['target_user_id'];
    $source_mapping_categories[$usergallery_id]['cid'] = $usergallery_id;
    $source_mapping_categories[$usergallery_id]['target_cid'] = $target_usergallery_id;
    $source_mapping_categories[$usergallery_id]['inserted'] = $user['inserted'];
    $source_mapping_categories[$usergallery_id]['name'] = '';
    $source_mapping_categories[$usergallery_id]['target_subcat_depth'] = 1;
    $source_mapping_categories[$usergallery_id]['incr_subcat_pos'] = 0;
    // obtain max position from albums within category. Could be none... if only categories exist at this level
    // convert to an increment value... albums start at pos = 100
    $result = cpg_db_query("SELECT category, COUNT( * ) AS count FROM {$CONFIG['TABLE_ALBUMS']} WHERE category = '$target_usergallery_id' GROUP BY category");
    $row = cpg_db_fetch_row($result);
    cpg_db_free_result($result);
    if ($row['count'] > 0) {
      $result = cpg_db_query("SELECT MAX(pos) AS maxpos FROM {$CONFIG['TABLE_ALBUMS']} WHERE category = '$target_usergallery_id'");
      $row = cpg_db_fetch_row($result);
      $source_mapping_categories[$usergallery_id]['incr_subalbum_pos'] = $row['maxpos'] - 99;
      cpg_db_free_result($result);
    } else {
      $source_mapping_categories[$usergallery_id]['incr_subalbum_pos'] = 0;
    }
    if ($debug_level >= 3) { $debug_trace[] = "Map build: 'categories' id: $usergallery_id (User Gallery) translates to $target_usergallery_id - target subcat depth: {$source_mapping_categories[$usergallery_id]['target_subcat_depth']} - incr subcat pos: {$source_mapping_categories[$usergallery_id]['incr_subcat_pos']} - incr subalbum pos: {$source_mapping_categories[$usergallery_id]['incr_subalbum_pos']}"; }
  } // end foreach
}


// initialize translation arrays for mapping pictures
function gallery_merge_map_prime_pictures($restart_key = 0) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $source_mapping_categories, $source_mapping_categories_key;            // access to categories mapping table and key
  global $source_mapping_albums,     $source_mapping_albums_key;                // access to albums mapping table and key
  global $source_mapping_pictures,   $source_mapping_pictures_key;              // access to pictures mapping table and key
  global $source_mapping_plugins;                                                // access to plugins mapping table
  // pictures array would be potentially too large if all pictures tracked
  // prime array with the keys we care about - those referenced elsewhere that need to be mapped:
  // Categories - thumb references > 0  (use array)
  // Albums - thumb references > 0      (use array)
  // Comments - unique pid reference    (access SOURCE table)
  // Favpics - unique pid references imbedded in serialized array (access SOURCE table)

  if ($restart_key > '0') {
    // we've already been here on previous execution..
    return;
  }
  // process categories
  foreach ($source_mapping_categories AS $category) {
    if ($category['thumb'] > 0) {
      $source_mapping_pictures[$category['thumb']]['map'] = 'cat';
    } // end if
  } // end foreach

  // process albums
  foreach ($source_mapping_albums AS $album) {
    if ($album['thumb'] > 0) {
      $source_mapping_pictures[$album['thumb']]['map'] .= 'alb';
    } // end if
  } // end foreach

  // process comments
  $result = cpg_db_query("SELECT DISTINCT pid FROM {$SOURCE['TABLE_COMMENTS']}", $SOURCE['LINK_ID']);
  while ($row = cpg_db_fetch_row($result)) {
    $source_mapping_pictures[$row['pid']]['map'] .= 'cmt';
  }
  cpg_db_free_result($result);

  // process favpics
  $result = cpg_db_query("SELECT * FROM {$SOURCE['TABLE_FAVPICS']}", $SOURCE['LINK_ID']);
  while ($row = cpg_db_fetch_row($result)) {
    $favpics = @unserialize(@base64_decode($row['user_favpics']));
    foreach ($favpics AS $pid) {
      $source_mapping_pictures[$pid]['map'] .= 'fav';
    }
  }
  cpg_db_free_result($result);

  // process plugins
  foreach ($source_mapping_plugins AS $plugin) {
    // see if plugin provided SQL to extract pid's
    if ($plugin['pid_sql'] != '') {
      $sql = str_replace('{cpg_table_prefix}', $SOURCE['TABLE_PREFIX'], $plugin['pid_sql']);
      $result = cpg_db_query($sql, $SOURCE['LINK_ID']);
      while ($row = cpg_db_fetch_row($result)) {
        $source_mapping_pictures[$row['pid']]['map'] .= 'plugin';
      }
      cpg_db_free_result($result);
    } // end if
  } // end foreach

}


// build translation arrays for mapping
function gallery_merge_map_build($source_row, $source_row_updated, $target_row, $translate_table, $map_values, $insert_id) {
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;
  static $initialized = FALSE;                                                  // initialization flag
  //   We don't know which map we will need to update - nor even all the possible map names (as support for plugin table support can add new maps..       
  //   Determine correct map and establish global reference dynamically.                                                                                  

  // maps are 2 dimensional arrays used to re-map keys from source to target
  // - either to existing target entry in the case of duplicates between galleries, or to newly inserted row.
  // first key is the $source_row[$map_key] - ie the user_id of the source row if mapping users
  // second key is the field name from source row being saved, with either 'target_field' for the new id to map to or 'incr_field' for value to be added to current value.
  //   various additional fields are saved as requested in $map_values array passed in

  if (!$initialized) {
    // initialization first time through only (or on restart)- set some values for base (0) usergroups, users, categories...
    // These values don't exist in tables, but needed to map future rows referring to 0 as usergroup, user_id, category (parent)
    // RESTART NOTE: no harm if values already in array... but will initialize any arrays we didn't get to checkpoint - and reset 'max' values to account for previous run

    // set usergroup mapping for user_id = 0 to retain 0 (used in `visibility` in albums - public albums have value of 0)
    $source_mapping_usergroups =& $GLOBALS['source_mapping_usergroups'];        // establish global reference
    $source_mapping_usergroups[0]['group_id'] = 0;
    $source_mapping_usergroups[0]['target_group_id'] = 0;
    $source_mapping_usergroups[0]['inserted'] = 0;
    $source_mapping_usergroups[0]['group_name'] = 'System';
    $source_mapping_usergroups[0]['target_group_name'] = 'System';
    if ($debug_level >= 3) { $debug_trace[] = "Map build: 'usersgroups' - id: 0 (System) translates to 0"; }

    // set user mapping for user_id = 0 to retain 0
    $source_mapping_users =& $GLOBALS['source_mapping_users'];                  // establish global reference
    $source_mapping_users[0]['user_id'] = 0;
    $source_mapping_users[0]['target_user_id'] = 0;
    $source_mapping_users[0]['inserted'] = 0;
    $source_mapping_users[0]['user_name'] = 'System';
    if ($debug_level >= 3) { $debug_trace[] = "Map build: 'users' - id: 0 (System) translates to 0"; }

    // set category mapping for category 0 to retain 0 (used in `parent` in categories - highest level has parent = 0
    $source_mapping_categories =& $GLOBALS['source_mapping_categories'];        // establish global reference
    $source_mapping_categories[0]['cid'] = 0;
    $source_mapping_categories[0]['target_cid'] = 0;
    $source_mapping_categories[0]['inserted'] = 0;
    $source_mapping_categories[0]['name'] = 'System';
    $source_mapping_categories[0]['target_subcat_depth'] = 1;
    // obtain max position from categories with parent = 0 (these will be all at depth 1) Could be none... if only albums exist at this level
    $result = cpg_db_query("SELECT parent, COUNT( * ) AS count FROM {$CONFIG['TABLE_CATEGORIES']} WHERE parent = '0' GROUP BY parent");
    $row = cpg_db_fetch_row($result);
    cpg_db_free_result($result);
    if ($row['count'] > 0) {
      $result = cpg_db_query("SELECT MAX(pos) AS maxpos FROM {$CONFIG['TABLE_CATEGORIES']} WHERE parent = '0'");
      $row1 = cpg_db_fetch_row($result);
      $source_mapping_categories[0]['incr_subcat_pos'] = $row1['maxpos'] + 1;    // pos starts at 0 - so add 1 to result
      cpg_db_free_result($result);
    } else {
      $source_mapping_categories[0]['incr_subcat_pos'] = 0;
    }
    // obtain max position from albums within category of 0 (those with category = 0)  Could be none... if only categories exist at this level
    // convert to an increment value... albums start at pos = 100
    $result = cpg_db_query("SELECT category, COUNT( * ) AS count FROM {$CONFIG['TABLE_ALBUMS']} WHERE category = '0' GROUP BY category");
    $row = cpg_db_fetch_row($result);
    cpg_db_free_result($result);
    if ($row['count'] > 0) {
      $result = cpg_db_query("SELECT MAX(pos) AS maxpos FROM {$CONFIG['TABLE_ALBUMS']} WHERE category = '0'");
      $row = cpg_db_fetch_row($result);
      $source_mapping_categories[0]['incr_subalbum_pos'] = $row['maxpos'] - 99;
      cpg_db_free_result($result);
    } else {
      $source_mapping_categories[0]['incr_subalbum_pos'] = 0;
    }
    if ($debug_level >= 3) { $debug_trace[] = "Map build: 'categories' id: 0 (System) translates to 0 - target subcat depth: {$source_mapping_categories[0]['target_subcat_depth']} - incr subcat pos: {$source_mapping_categories[0]['incr_subcat_pos']} - incr subalbum pos: {$source_mapping_categories[0]['incr_subalbum_pos']}"; }

    $initialized = TRUE;
  }

  // determine correct tables and keys
  $translate_array_var = "source_mapping_{$translate_table}";
  $translate_array =& $GLOBALS[$translate_array_var];                           // establish global reference...
  $translate_array_key_var = "source_mapping_{$translate_table}_key";
  $translate_array_key = $GLOBALS[$translate_array_key_var];                    // no reference needed - static value
  $debug_text = "Map build: '$translate_table' Key Field: $translate_array_key: ";

  // start out array entry with the key field...
  $translate_array[$source_row[$translate_array_key]][$translate_array_key] = $source_row[$translate_array_key];
  if ($insert_id == '*TGT*') {
    // row already existed in target
    $insert_id = $translate_array[$source_row[$translate_array_key]]["target_$translate_array_key"] = $target_row[$translate_array_key];
    $insert = $translate_array[$source_row[$translate_array_key]]['inserted'] = 0;
    } else {
    // row was just inserted...
    $translate_array[$source_row[$translate_array_key]]["target_$translate_array_key"] = $insert_id;
    $insert = $translate_array[$source_row[$translate_array_key]]['inserted'] = 1;
    }
  if ($debug_level >= 3) { $debug_text .= "{$translate_array[$source_row[$translate_array_key]]["$translate_array_key"]} => {$translate_array[$source_row[$translate_array_key]]["target_$translate_array_key"]} "; }

  foreach ($map_values AS $value) {
    // build mapping array

    // in all cases - set default mappings
    // if value begins with / - field is for doc/debug - target not required (save space while still making array human readable
    if (substr($value, 0, 1) == '/') {
      $value = str_replace('/', '', $value);
      $translate_array[$source_row[$translate_array_key]][$value] = $source_row[$value];
    } else {
      $translate_array[$source_row[$translate_array_key]][$value] = $source_row[$value];
      $translate_array[$source_row[$translate_array_key]]["target_$value"] = $source_row_updated[$value];
    }
    if ($debug_level >= 3) { $debug_text .= "- Field: '$value' Value: {$source_row[$value]} => {$source_row_updated[$value]}"; }

    // some special processing for categories
    if (($translate_table == 'categories') && ($value == 'depth')) {
      // special processing for categories `depth` field... set target_subcat_depth to our new depth +1
      $translate_array[$source_row[$translate_array_key]]["target_subcat_$value"] = $source_row[$value] + 1;
      if ($debug_level >= 3) { $debug_text .= "- Field 'target_subcat_$value': => {$translate_array[$source_row[$translate_array_key]]["target_subcat_$value"]}"; }
    } elseif (($translate_table == 'categories') && ($value == 'pos')) {
      // special processing for categories `pos` field... set incr_subcat_pos to max pos of categories under this cat..
      // obtain max position from categories within this parent. Could be none... if only albums exist at this level - or if cat was just inserted...
      $result = cpg_db_query("SELECT parent, COUNT( * ) AS count FROM {$CONFIG['TABLE_CATEGORIES']} WHERE parent = '$insert_id' GROUP BY parent");
      $row = cpg_db_fetch_row($result);
      cpg_db_free_result($result);
      if ($row['count'] > 0) {
        $result = cpg_db_query("SELECT MAX(pos) AS maxpos FROM {$CONFIG['TABLE_CATEGORIES']} WHERE parent = '{$insert_id}'");
        $row = cpg_db_fetch_row($result);
        cpg_db_free_result($result);
        $translate_array[$source_row[$translate_array_key]]['incr_subcat_pos'] = $row['maxpos'] + 1;    // pos starts at 0 - so add 1 to result
      } else {
        $translate_array[$source_row[$translate_array_key]]['incr_subcat_pos'] = 0;
      }
      if ($debug_level >= 2) { $debug_text .= "- Field 'incr_subcat_$value': => {$translate_array[$source_row[$translate_array_key]]["incr_subcat_$value"]}"; }
      // obtain max position from albums within this category. Could be none... if only categories exist at this level - or if cat was just inserted...
      // convert to an increment value... albums start at pos = 100
      $result = cpg_db_query("SELECT category, COUNT( * ) AS count FROM {$CONFIG['TABLE_ALBUMS']} WHERE category = '$insert_id' GROUP BY category");
      $row = cpg_db_fetch_row($result);
      cpg_db_free_result($result);
      if ($row['count'] > 0) {
        $result = cpg_db_query("SELECT MAX(pos) AS maxpos FROM {$CONFIG['TABLE_ALBUMS']} WHERE category = '{$insert_id}'");
        $row = cpg_db_fetch_row($result);
        cpg_db_free_result($result);
        $translate_array[$source_row[$translate_array_key]]['incr_subalbum_pos'] = $row['maxpos'] - 99;
      } else {
        $translate_array[$source_row[$translate_array_key]]['incr_subalbum_pos'] = 0;
      }
      if ($debug_level >= 3) { $debug_text .= "- Field 'incr_subalbum_$value': => {$translate_array[$source_row[$translate_array_key]]["incr_subalbum_$value"]}"; }

    }
  } // end foreach
  // added all requested values...
  if ($debug_level >= 3) { $debug_trace[] = $debug_text; }
}


// perform data translations
function gallery_merge_map_translate($source_row, $translations) {
  global $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  $map_warning_issued = FALSE;
  //   We don't know which map we will need to read - nor even all the possible map names (as support for plugin table support can add new maps..
  //   Determine correct map and establish global reference dynamically.

  foreach ($translations AS $translate_fieldname => $translate_table_key_field) {
    if (substr_count($translate_table_key_field, '.') == '3') {
      // 4 parms provided
      // key translate_fieldname needs to be extracted - has .# suffix
      list($translate_fieldname, $suffix) = explode('.', $translate_fieldname);
      // extract table, key, field, and condition from $translate_table_key_field
      list($translate_table, $translate_key, $translate_field, $translate_condition) = explode('.', $translate_table_key_field);
      // extract the condition details
      list($translate_condition_field, $translate_condition_value) = explode('=', $translate_condition);
      settype($translate_condition_value, "string");                            // force value to a string
      if ($debug_level >= 3) { $debug_trace[] = "Translate condition: '$translate_condition_field' = '$translate_condition_value' - Source: '{$source_row[$translate_condition_field]}'"; }
      // and test
      if ($source_row[$translate_condition_field] != $translate_condition_value) {
        continue;      // bypass this row...
      }
    } else {
      // 3 parms is default - extract table, key, field
      list($translate_table, $translate_key, $translate_field) = explode('.', $translate_table_key_field);
      $translate_condition = '';                                                // clear condition - not present here
    }
    // handle special cases of $translate_table = 'append'
    // $translate_key = start|end - where to 'append'
    // $translate_field = value to 'append'
    // code currently unused by core files - but available if converting a plugin table needs it...
    if ($$translate_table == 'append') {
      if ($translate_key == 'start') {
        $source_row[$translate_fieldname] = $translate_field.$source_row[$translate_fieldname];
      } elseif ($translate_key == 'end') {
        $source_row[$translate_fieldname] .= $translate_field;
      } else {
        // oops.. shouldn't be here - invalid $translate_key...
        if ($debug_level >= 1) {
          $debug_trace[] = "Request to append with invalid key '$translate_key' - validate input";
        }
        if (!$map_warning_issued) {
          $map_warning_issued = TRUE;
          $warning_message = $lang_plugin_gallery_merge['w_unexpected_mapping'].'<br>'.$lang_plugin_gallery_merge['w_unexpected_mapping2'];
          gallery_merge_error('db', $plugin_table_constant, 'warning', $warning_message);
        }
      }
      // move on to next entry...
      continue;
    }

    // determine correct tables and keys
    $translate_array_var = "source_mapping_{$translate_table}";
    $translate_array =& $GLOBALS[$translate_array_var];                         // establish global reference...
    $translate_array_key_var = "source_mapping_{$translate_table}_key";
    $translate_array_key = $GLOBALS[$translate_array_key_var];                  // no reference needed - static value

    if ($debug_level >= 2) { $debug_text = "Translation for '$translate_fieldname' Value: '{$source_row[$translate_fieldname]}' using Translate table '$translate_table' - key '$translate_key' ({$source_row[$translate_key]}) - field '$translate_field'"; }
    $save_source_translate_value = $source_row[$translate_fieldname];           // save for diagnostics
    // special processing for target fields beginning with incr_ use addition instead of assignment
    if (substr($translate_field, 0, 5) == 'incr_') {
      $source_row[$translate_fieldname] += $translate_array[$source_row[$translate_key]]["$translate_field"];         // add value for field
    } else {
    // normal processing - assign requested value from mapping table
      $source_row[$translate_fieldname] = $translate_array[$source_row[$translate_key]]["target_$translate_field"];   // assign value for all others
    }

    if ((is_numeric($source_row[$translate_fieldname])) && (is_numeric($save_source_translate_value)) && (is_numeric($source_row[$translate_key]))) {
      // all as expected - don't output unless level 3 debug
      if ($debug_level >= 3) { $debug_trace[] = "$debug_text converts to {$source_row[$translate_fieldname]}"; }
    } else {
      // one or more of mapping values/results isn't an integer... oops... level 2 debug
      if ($debug_level >= 1) {
        $debug_text .= "converts to {$source_row[$translate_fieldname]}<br>Translate Table: $translate_table - Key: $translate_key - Field: $translate_field - Condition: $translate_condition<br>";
        $debug_trace[] = $debug_text.'<pre>SOURCE ROW '.print_r($source_row, TRUE).'</pre><pre>MAP ROW '.print_r($translate_array[$source_row[$translate_key]], TRUE).'</pre>';
      }
      if (!$map_warning_issued) {
        $map_warning_issued = TRUE;
        $warning_message = $lang_plugin_gallery_merge['w_unexpected_mapping'].'<br>'.$lang_plugin_gallery_merge['w_unexpected_mapping2'];
        gallery_merge_error('db', $plugin_table_constant, 'warning', $warning_message);
      }
    }
  }
  return $source_row;
}


// set target destination based on provided source - used for add/remove folder processing and translating and verifying pictures entries for userpics
function gallery_merge_set_target_dest($source) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute_files;
  global $source_mapping_users, $source_mapping_users_key;                      // users map needed for userpics processing
  global $source_mapping_albums, $source_mapping_albums_key;                    // albums map may be needed for userpics processing
  global $process_constant;
  static $remove_subfolder_len = '0';                                           // save length to avoid recalculating for each call
  static $utility_user_array = array();                                         // if calling from utility code - we won't have the benefit of user mapping array...
  static $utility_album_array = array();                                        // if calling from utility code - we won't have the benefit of album mapping array...
  static $userpics_length = '0';
  static $userpics_warn = FALSE;
  // initialize
  if ($userpics_length == '0') {
    $userpics_length = strlen($SOURCE['userpics']);                             // get userpics length for the SOURCE the first time through... (used by add/remove/userpics...)
  }
  // determine if we need to remove or add requested folder to source filepath - or adjust folder for userpics to match config and target userid
  // based on config variables 'plugin_gallery_merge_remove_subfolder' and 'plugin_gallery_merge_add_subfolder' and the actual directory name (edit and userpics are excluded...)
  $target = $source;

  // process remove request if specified - and not a userpics destination
  if (($CONFIG['plugin_gallery_merge_remove_subfolder'] != '') && ($SOURCE['userpics'] != substr($target, 0, $userpics_length))) {
    if ($remove_subfolder_len == '0') {                                         // set length once only
      $remove_subfolder_len = strlen($CONFIG['plugin_gallery_merge_remove_subfolder']);
    }
    if ($CONFIG['plugin_gallery_merge_remove_subfolder'] == substr($target, 0, $remove_subfolder_len)) {
      $target = str_replace($CONFIG['plugin_gallery_merge_remove_subfolder'], '', $target);     // update source filepath to remove requested folder
      if ($debug_level >= 3)  { $debug_trace[] = "Translation for 'filepath' Value: '$source' converts to '$target' (subfolder removal)"; }
    }
  }

  // process userpics destination.. a little more involved
  // the two galleries might have different values for the 'userpics' folder...
  // folders within userpics are named the numeric user_id of the user (unless 'silly_safe_mode') - the user_id value likely changed...
  // if $CONFIG['upload_create_album_directory'] (hidden option) - then subfolders for the user are named the numeric aid - which also likely changed
  if ($SOURCE['userpics'] == substr($target, 0, $userpics_length)) {
    // have a userpics destination
    // extract the userid
    $target_parts = explode("/", $target, 4);                                   // get the pieces of filepath - down to possible subfolder
    $userfolder = $target_parts[1];                                             // grab the second field - the user folder ([0] = 'userpics' (or whatever config value is)
    $albumfolder = $target_parts[2];                                            // grab the third field [0] - might be album folder...
    // see if source and target userpics match..
    if ($CONFIG['userpics'] != $SOURCE['userpics']) {
      // userpics specification different - change it..
      $target = str_replace($SOURCE['userpics'], $CONFIG['userpics'], $target);
    }
    // see if we have a folder to translate...
    if (is_numeric($userfolder)) {
      // userfolder is numeric... continue on...
      $userid = $userfolder - FIRST_USER_CAT;                                   // subtract 10000 to get to user_id (use first usercat defined as number...)
      $target_userid = FALSE;
      if (is_array($source_mapping_users)) {
        // this array will exist for merge_db.php executions... but calls from merge_utility.php will not have the mapping array
        if (array_key_exists($userid, $source_mapping_users)) {
          // a userid we've already seen... (better be the case!) take from array
          $target_userid = $source_mapping_users[$userid]['target_user_id'];
          $target_userfolder = $target_userid + FIRST_USER_CAT;
          $target_username = $source_mapping_users[$userid]['user_name'];
        }
      } else {
        // merge_utility.php executions won't have mapping_users table.. make needed sql calls and build $utility_user_array to minimize calls
        if (array_key_exists($userid, $utility_user_array)) {
          // a userid we've already seen... take from array
          $target_userid = $utility_user_array[$userid]['target_user_id'];
          $target_userfolder = $target_userid + FIRST_USER_CAT;
          $target_username = $utility_user_array[$userid]['user_name'];
        } else {
          // not in utility array - sql calls to obtain...
          $result = cpg_db_query("SELECT user_name FROM {$SOURCE['TABLE_USERS']} WHERE user_id = '$userid' LIMIT 1", $SOURCE['LINK_ID']);
          $source_row = cpg_db_fetch_row($result);
          cpg_db_free_result($result);
          $result = cpg_db_query("SELECT user_id FROM {$CONFIG['TABLE_USERS']} WHERE user_name = '{$source_row['user_name']}' LIMIT 1");
          $target_row = cpg_db_fetch_row($result);
          cpg_db_free_result($result);
          $target_userid = $utility_user_array[$userid]['target_user_id'] = $target_row['user_id'];
          $target_userfolder = $target_userid + FIRST_USER_CAT;
          $target_username = $utility_user_array[$userid]['user_name'] = $source_row['user_name'];
        }
      }
      if (!$target_userid) {
        // no match found...this shouldn't happen... as we have already processed all users earlier in merge!
        $target_userfolder = 'unknown';
        if (!$userpics_warn) {
          // if no warning yet - issue warning (will fail sim)
          $userpics_warn = TRUE;
          $warning_message = $lang_plugin_gallery_merge['w_unexpected_mapping'].'<br>'.$lang_plugin_gallery_merge['w_unexpected_mapping2'];
          gallery_merge_error('files', $process_constant, 'warn', $warn_message);
        }
        $col2 = sprintf($lang_plugin_gallery_merge['userpics_map_warning'], $userid);
        gallery_merge_output_status('blank', $col2);
      }
      // one way or another - $target_userfolder is now set... use it
      $target = str_replace($userfolder, $target_userfolder, $target);

      // see if we need to worry about albums... (based on $SOURCE setting - already reported difference earlier if set in SOURCE and not in TARGET)
      if ($SOURCE['upload_create_album_directory'] && is_numeric($albumfolder)) {
        // config option to createalbum folders... and numeric folder found... translate if we can
        // it is possible album won't exist - pics could have been moved between albums and orig album deleted - which wouldn't change folders in path
        // we won't fail if we can't remap...
        if (is_array($source_mapping_albums)) {
          // this array will exist for merge_db.php executions... but calls from merge_utility.php will not have the mapping array
          if (array_key_exists($albumfolder, $source_mapping_albums)) {
            // an album we've already seen... take from array
            $target_albumfolder = $source_mapping_albums[$albumfolder]['target_aid'];
            $target_albumname = $source_mapping_albums[$albumfolder]['title'];
          }
        } else {
          // merge_utility.php executions won't have mapping_users table.. make needed sql calls and build $utility_user_array to minimize calls
          if (array_key_exists($albumfolder, $utility_album_array)) {
            // an album we've already seen... take from array
            $target_albumfolder = $utility_album_array[$albumfolder]['target_aid'];
            $target_albumname = $utility_album_array[$albumfolder]['title'];
          } else {
            // not in utility array - sql calls to obtain...
            $result = cpg_db_query("SELECT title FROM {$SOURCE['TABLE_ALBUMS']} WHERE aid = '$albumfolder' LIMIT 1", $SOURCE['LINK_ID']);
            $source_row = cpg_db_fetch_row($result);
            cpg_db_free_result($result);
            $result = cpg_db_query("SELECT aid FROM {$CONFIG['TABLE_ALBUMS']} WHERE title = '{$source_row['title']}' AND owner = '{$utility_user_array[$userid]['target_user_id']}' LIMIT 1");
            $target_row = cpg_db_fetch_row($result);
            cpg_db_free_result($result);
            $target_albumfolder = $utility_album_array[$albumfolder]['target_aid'] = $target_row['aid'];
            $target_albumname = $utility_album_array[$albumfolder]['title'] = $source_row['title'];
          }
        }
        if ($target_albumfolder) {
          $target = str_replace($albumfolder, $target_albumfolder, $target);
        } else {
          if ($debug_level >= 3)  { $debug_trace[] = "Userpics destination: Unmapped subfolder '$albumfolder' found in userpics filepath"; }
        }
      } // endif album directories and folder is numeric
      if ($debug_level >= 3)  { $debug_trace[] = "Translation for filepath '$source' using Translate table 'users' converts to '$target' (userpics for '$target_username')"; }
    // endif userpics is_numeric
    } else {
      // we have a non-numeric folder/file in root userpics... ignore for now
      if ($debug_level >= 3)  { $debug_trace[] = "Userpics destination: Non-numeric folder/file '$userfolder' found in userpics filepath"; }
    }
  } // end userpics

  // process add request if specified - and not a userpics destination (else handles the userpics part...)
  elseif ($CONFIG['plugin_gallery_merge_add_subfolder'] != '') {
    $target = $CONFIG['plugin_gallery_merge_add_subfolder'].$target;          // add requested additional folder
    if ($debug_level >= 3)  { $debug_trace[] = "Translation for 'filepath' Value: '$source' converts to '$target' (subfolder add)"; }
  }

  return $target;
}


// process most data tables
function gallery_merge_process_table($table_constant, $orderby, $compare_fields, $compare_type, $skip_keys, $obsolete_keys, $zero_keys, $skip_keys_difference, $translations, $map_values, $restart_key = 0) {
/*
      SAMPLE input variables:
        $table_constant = 'TABLE_USERGROUPS';
        $orderby = 'ORDER BY depth, pos, cid';                                  // default order_by is key field - specify if data needs to be processed in specific order
        $compare_fields = 'group_name';                                         // field(s) - comma separated - to match source to target (identify duplicates/inserts)
        $compare_type = '';                                                     // type of compare: UNIQUE, IGNORE, MERGE
        $skip_keys = array('group_id');                                         // set keys to bypass for insert (typically auto-increment fields)
        $obsolete_keys = array('subcat_count', 'alb_count');                    // set keys that are obsolete for this table - will be bypassed if exist (with warning)
        $zero_keys = '';                                                        // set keys to 0 value on insert (typically data not yet available... to be updated later)
        $map_values = array('/name', 'pos', 'thumb', 'depth');                  // set mapping values to set fields to build array for later processing
        $skip_keys_difference = array();                                        // set keys to bypass for comparing like rows in source and target (like passwords...)
        $translations = array('user_group' => 'usergroups');                    // set translations to be done... fieldname => table to use
*/
  global $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;
  global $CONFIG, $SOURCE;
  static $count_total_inserted = 0;
  $count_inserted = 0;                                                          // initialize inserted counter
  $count_existed = 0;                                                           // initialize existed counter
  $source_table = $SOURCE[$table_constant];                                     // source table name
  $target_table = $CONFIG[$table_constant];                                     // target table name
  $table_name_uc = substr($table_constant, 6);                                  // uppercase table constant for messages
  $table_name = strtolower($table_name_uc);                                     // lowercase table constant for messages

  $table_key_var = "source_mapping_{$table_name}_key";
  $table_key = $GLOBALS[$table_key_var];                                        // key used for simulated inserts
  $map_var = "source_mapping_{$table_name}";                                    // map name used for checkpointing (contents not accessed here... with two exceptions)
  $target_array = array();                                                      // initialize target array

  $unique_warning_issued = FALSE;

  $col1 = sprintf($lang_plugin_gallery_merge['processing_table'], $table_name);
  $col1 .= gallery_merge_db_doc($table_name);
  $col2 = $lang_plugin_gallery_merge['source_merged'];
  gallery_merge_output_status($col1, $col2);

  // save our status to start...
  $data = func_get_args();
  gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:STARTED", $data);

  // compare table structures
  gallery_merge_compare_structure($table_constant, $source_table, $target_table, $obsolete_keys, &$skip_keys);

  if (!$plugin_gallery_merge_execute) {
    // for simulation - we need max value of autoincrement field to be able to simulate inserts
    if (($table_key != '*none*') && ($table_name != 'favpics')) {
      $result = cpg_db_query("SELECT MAX($table_key) AS max_value FROM {$target_table}");
      $row = cpg_db_fetch_row($result);
      cpg_db_free_result($result);
      $next_insert_id = $row['max_value'] +1;
      if ($debug_level >= 3) { $debug_trace[] = "Simulated inserts into '$table_name' will start with key '$next_insert_id'"; }
    }
  }

  // set SQL parms
  $target_select = '*';
  if ((isset($CONFIG["plugin_gallery_merge_{$table_name}_process"])) && ($CONFIG["plugin_gallery_merge_{$table_name}_process"] == '1')) {
    // special processing to minimize memory usage for a table - only select what we need to map (and key field)
    // for duplicates - we will have less info to compare - set this option to 0 for full detail
    // currently only implemented for pictures table - but adding variable to config table for desired table will be auto-detected here
    $target_select = "$compare_fields, $table_key";                             // start with compare fields and table key
    foreach ($map_values as $map_column) {                                      // process mapping fields requested
      if (substr($map_column, 0, 1) == '/') {
        continue;                                                               // fields beginning with / don't need target data
      }
      $target_select .= ", $map_column";                                        // add field to select
    }
    $col2 = sprintf($lang_plugin_gallery_merge['limited_compare'], $target_select);
    gallery_merge_output_status('blank', $col2);
  }

  $where = '';
  if (($restart_key > '0') && ($table_key != '*none*')) {
    // have a key to resume processing with - and table has a key... (only catmap does not have a single column key)
    // some tables (only categories from base) use special order by processing that make restarting at a key more interesting...
    // we don't know how many records we processed already to use LIMIT... but these tables are generally small.
    if ($orderby == '') {
      // normal order processing by key - we can restart at the last key
      $where = "WHERE $table_key > '$restart_key'";
      $skip_restart = FALSE;
    } else {
      // unique order by processing... we can't just start at the last key...
      // we'll get all records - and skip until key matches our restart key.
      $skip_restart = $restart_key;
      $debug_trace[] = "Restart processing table $table_name with 'skip restart' to key: $skip_restart";
    }
  }

  if (($orderby == '') && ($table_key != '*none*')) {
    // default orderby to key field unless specified
    $orderby = "ORDER BY $table_key";
  }

  if ($debug_level >= 3) { $debug_trace[] = "Processing table $table_name with Orderby: '$orderby' and Where: '$where'"; }

  // extract compare fields
  $compare_field =  explode(',', $compare_fields);                              // extract based on comma... $compare_field[0] is first, $compare_field[1] is second, etc
  $num_compare_fields = count($compare_field);
  if ($num_compare_fields > '4') {
    // this is an arbitrary limit.. can be raised if needed by adding options to the setting of $match_key below (for both TARGET and SOURCE processing)
    $error_message = $lang_plugin_gallery_merge['e_invalid_compare_toomany'];
    gallery_merge_error('db', $table_constant, 'error', $error_message);
    return;
  }

  // process compare types - UNIQUE, INSERT, MERGE
  if ($compare_type != 'IGNORE') {
    // Retrieve TARGET DB info
    //**Memory Usage Note
    //**     - for most tables, we process target with SELECT * and load into array... large picture tables (probably the only one of concern...) could have an issue.
    //**       For pictures, a config option exists to select and save a reduced amount of data - reduces ability to compare all fields on duplicates, but should support larger target galleries...
    //**       Normal merge should have few conflicts - but a restart after a failure (not Quiesce) could reprocess up to checkpoint interval of duplicate files...
    //**       My 19,000+ picture gallery is only 3MB of data in MySQL - so may be ok...
    $insert_ignore = '';
    // load target data into array (for all but largest tables...)
    $target_result = cpg_db_query("SELECT {$target_select} FROM {$target_table} {$orderby}");
    while ($target_row = cpg_db_fetch_row($target_result)) {
      if ($num_compare_fields == '1') {
        $match_key = $target_row[$compare_field[0]];
      } elseif ($num_compare_fields == '2') {
        $match_key = "{$target_row[$compare_field[0]]}-{$target_row[$compare_field[1]]}";
      } elseif ($num_compare_fields == '3') {
        $match_key = "{$target_row[$compare_field[0]]}-{$target_row[$compare_field[1]]}-{$target_row[$compare_field[2]]}";
      } elseif ($num_compare_fields == '4') {
        $match_key = "{$target_row[$compare_field[0]]}-{$target_row[$compare_field[1]]}-{$target_row[$compare_field[2]]}-{$target_row[$compare_field[3]]}";
      }
      if (!key_exists($match_key, $target_array)) {
        // key not seen yet - set it..
        $target_array[$match_key] = $target_row;
      } elseif ($compare_type == 'UNIQUE') {
        // key already exists.. and UNIQUE processing requested - if this key also exists in source - we can have a problem...
        $target_array[$match_key]['gm_duplicate'] = 'dup';                      // mark duplicate encountered
        // we'll see below if SOURCE contains the same info...
      }
    }
    cpg_db_free_result($target_result);
  } else {
    // set up IGNORE for insert and leave array empty - we will attempt insert of all records and let SQL handle (INSERT IGNORE)
    $insert_ignore = 'IGNORE';
  }

  // Retrieve SOURCE DB info - and compare to TARGET
  if (($message_level == 2) && ($table_name == 'pictures')) {
    // suppress individual INSERTED messages for pictures... just too many. (Debug will still show SQL if requested.)
    $col2 = sprintf($lang_plugin_gallery_merge['messages_suppressed'], $table_name_uc);
    gallery_merge_output_status('blank', $col2);
  }
  // SELECT * required from SOURCE to be able to build INSERT statements for what should be most entries...
  //** Memory Usage Note
  //**     - possibly change this SELECT to 'UNBUFFERED' to reduce memory usage. testing required (and new function to call...)
  $source_result = cpg_db_query("SELECT * FROM {$source_table} {$where} {$orderby}", $SOURCE['LINK_ID']);
  while ( ($source_row = cpg_db_fetch_row($source_result)) ) {
    if ($skip_restart) {
      // restart processing for table with unique order by...
      if ($source_row[$table_key] != $skip_restart) {
        continue;   // skip this record
      } else {
        $skip_restart = FALSE;
        $debug_trace[] = "Restart found record $skip_restart - resuming processing...";
        continue;   // start with next record
      }
    }
    $insert_id = '#';                                                           // reset insert_id for this iteration
    $inserted = TRUE;

    // assuming we will be inserting most records encountered... and because the compare fields may require mapping first - we'll process mapping for every record, then see if it needs to be inserted.
    // perform any translations needed
    $source_row_updated = gallery_merge_map_translate($source_row, $translations);

    // special processing for pictures - not handled in normal map translation (partial field replacements)
    // - if adding a subfolder was requested... add to all pics EXCEPT those in 'userpics' folders ('edit' folder completely bypassed)
    // - if removing a subfolder was requested... remove from all places it exists (shouldn't exist in userpics)
    // - if userpics entry, translate user related folder name based on remapped user
    if ($table_name == 'pictures') {
      $source_row_updated['filepath'] = gallery_merge_set_target_dest($source_row_updated['filepath']);
    }

    // We may need to compare one or more fields... check possibilities based on number of compare fields provided (target array will be built to match one of these)
    //** Memory Usage Note
    //**     - compares here need to change if altering from using an array to hold target table.
    if ($num_compare_fields == '1') {
      $match_key = $source_row_updated[$compare_field[0]];
    } elseif ($num_compare_fields == '2') {
      $match_key = "{$source_row_updated[$compare_field[0]]}-{$source_row_updated[$compare_field[1]]}";
    } elseif ($num_compare_fields == '3') {
      $match_key = "{$source_row_updated[$compare_field[0]]}-{$source_row_updated[$compare_field[1]]}-{$source_row_updated[$compare_field[2]]}";
    } elseif ($num_compare_fields == '4') {
      $match_key = "{$source_row_updated[$compare_field[0]]}-{$source_row_updated[$compare_field[1]]}-{$source_row_updated[$compare_field[2]]}-{$source_row_updated[$compare_field[3]]}";
    }

    $insert_record = FALSE;
    $report_differences = FALSE;
    if (($compare_type != 'MERGE') && (!array_key_exists($match_key, $target_array))) {
      // SOURCE info not in TARGET DB (or IGNORE requested...)
      if (($message_level == 2) && ($table_name != 'pictures')) {
        // suppress individual INSERTED messages for pictures... just too many. (Debug will still show SQL if requested.)
        $col2 = sprintf($lang_plugin_gallery_merge['row_inserted'], $table_name_uc, substr($match_key, 0, 50));  // truncate inserted match data for message
        gallery_merge_output_status('blank', $col2);
      }
      $insert_record = TRUE;
    } elseif (($compare_type == 'INSERT') && (array_key_exists($match_key, $target_array))) {
      // SOURCE info in TARGET DB but INSERT (versus UNIQUE) requested...
      $col2 = sprintf($lang_plugin_gallery_merge['row_inserted_dup'], $table_name_uc, substr($match_key, 0, 50));  // truncate inserted match data for message
      $insert_record = TRUE;
      $report_differences = TRUE;
    } elseif (($compare_type == 'UNIQUE') && ($table_name == 'albums') && (array_key_exists($match_key, $target_array))) {
      // SOURCE info in TARGET DB for UNIQUE album request.. Check to see if match on user vs non-user album...
      if (($source_row_updated['category'] < FIRST_USER_CAT) && ($target_array[$match_key]['category'] < FIRST_USER_CAT)) {
        // both are main category... let it fall through as duplicate...
      } elseif ((($source_row_updated['category'] < FIRST_USER_CAT) && ($target_array[$match_key]['category'] > FIRST_USER_CAT)) ||
                (($source_row_updated['category'] > FIRST_USER_CAT) && ($target_array[$match_key]['category'] < FIRST_USER_CAT))) {
        // one is usercat and one is main...
        if (!key_exists('gm_duplicate', $target_array[$match_key])) {
          // no duplicates in source - so the only entry is a mismatch - insert away... (otherwise fall through as dup...)
          if ($debug_level >= 3) { $debug_trace[] = "Insert allowed for unique album {$source_row_updated['title']} new cat {$source_row_updated['category']} - exists with cat {$target_array[$match_key]['category']}"; }
          $col2 = sprintf($lang_plugin_gallery_merge['row_inserted'], $table_name_uc, substr($match_key, 0, 50));  // truncate inserted match data for message
          gallery_merge_output_status('blank', $col2);
          $insert_record = TRUE;
        }
      // should be only case left - both are usercat (> FIRST_USER_CAT)... are they for same user???
      } elseif ($source_row_updated['owner'] == $target_array[$match_key]['owner']) {
        // same user owns both - let it fall through as dup...
        if (!key_exists('gm_duplicate', $target_array[$match_key])) {
          // no duplicates in source - so the only entry is a mismatch - insert away... (otherwise fall through as dup...)
          if ($debug_level >= 3) { $debug_trace[] = "Insert allowed for unique album {$source_row_updated['title']} new cat {$source_row_updated['category']} - exists with cat {$target_array[$match_key]['category']}"; }
          $col2 = sprintf($lang_plugin_gallery_merge['row_inserted'], $table_name_uc, substr($match_key, 0, 50));  // truncate inserted match data for message
          gallery_merge_output_status('blank', $col2);
          $insert_record = TRUE;
        }
      }
    }
    // see if we need to insert.. and do so
    if ($insert_record) {
      // generate sql
      $merge_sql = gallery_merge_generate_sql($target_table, $source_row_updated, $skip_keys, $zero_keys, $insert_ignore);
      gallery_merge_debug_sql($merge_sql);
      $insert_id = '*SIM*';                                                     // show new insert id to be generated
      if ($plugin_gallery_merge_execute) {
        // execution mode!
        $insert_result = cpg_db_query($merge_sql);                              // execute SQL
        $insert_affected_rows = cpg_db_affected_rows();                         // should be 1 for normal INSERT - could be 0 for INSERT IGNORE
        if ($insert_affected_rows) {
          $insert_id = cpg_db_last_insert_id();                                 // get inserted id as final value
        } else {
          // row NOT inserted - INSERT IGNORE specified
          if ($debug_level >= 3) { $debug_trace[] = "Insert bypassed (IGNORE) for table $table_name_uc with source key {$source_row_updated[$table_key]}"; }
        }
      } else {
        // simulation mode
        $insert_id = $next_insert_id;                                           // set simulated id
        $next_insert_id ++;                                                     // and increment for next row
      }
      $count_inserted ++;
      $count_total_inserted ++;

   } elseif ($compare_type != 'MERGE') {
      // SOURCE info already in TARGET DB (we didn't insert... and not MERGE)
      if (($compare_type == 'UNIQUE') && (key_exists('gm_duplicate', $target_array[$match_key]))) {
        // unique processing requested - but the TARGET system has a duplicate for this item - can't properly map...
        $col2a = sprintf($lang_plugin_gallery_merge['target_duplicate'], $compare_fields).'<br>'.$lang_plugin_gallery_merge['target_duplicate2'];
        gallery_merge_output_status('blank', $col2a);                           // col2 has message being built at moment...
        if ($plugin_gallery_merge_execute) {
          // abort if execution
          $error_message = $lang_plugin_gallery_merge['e_target_duplicate'];
          gallery_merge_error('db', $table_constant, 'error', $error_message);
        } elseif (!$unique_warning_issued) {
          // warn if simulation
          $unique_warning_issued = TRUE;
          $warning_message = $lang_plugin_gallery_merge['w_target_duplicate']."<br />- ".$lang_plugin_gallery_merge['w_resolve_duplicate']." ".$lang_plugin_gallery_merge['w_rerun_simulation'];
          gallery_merge_error('db', $table_constant, 'warning', $warning_message);
        }
      }
      // continue on...
      $col2 = sprintf($lang_plugin_gallery_merge['row_exists'], $table_name_uc, substr($match_key, 0, 50));  // truncate inserted match data for message
      $report_differences = TRUE;
      if ($table_name == 'albums') {
        // for existing picture only - merge hit, keyword (only 1 allowed) fields
        gallery_merge_process_albums_existing($source_row, $target_array[$match_key]);
        $count_total_inserted ++;                                               // count here for the QUIESCE limit
      }
      if ($table_name == 'pictures') {
        // for existing picture only - merge hit, vote, rating, and keyword  fields
        gallery_merge_process_pictures_existing($source_row, $target_array[$match_key]);
        $count_total_inserted ++;                                               // count here for the QUIESCE limit
      }
      $insert_id = '*TGT*';                                                     // not inserting data - set flag to use id from TARGET
      $inserted = FALSE;
      $count_existed ++;
   }

   if ($report_differences) {
      // if reduced memory  processing for table - we won't have full target row... only a reduced commparison (of those fields we have from target)
      if ((isset($CONFIG["plugin_gallery_merge_{$table_name}_process"])) && ($CONFIG["plugin_gallery_merge_{$table_name}_process"] == '1')) {
        $partial_data = TRUE;
      } else {
        $partial_data = FALSE;
      }
      // compare using updated source row - avoid differences that were fixed by mapping (and allows compare with added/removed folders)
      $difference = gallery_merge_compare_arrays($source_row_updated, $target_array[$match_key], $skip_keys_difference, $partial_data);
      $col2 .= "<br />$difference";
      gallery_merge_output_status('blank', $col2);
      if (!$insert_record && ($compare_type != 'MERGE')) {
        // we didn't insert and done comparing
        $source_row_updated = $target_array[$match_key];                        // show updated row as existing target
      }
   }

    if ($compare_type == 'MERGE') {
      // MERGE requested... currently only used for favpics - no key field, inbedded pid references in serialized arrays...
      // have to translate user_id before comparing to see if insert or update - so source_row_updated used in more places
      if ($table_name == 'favpics') {
        $source_row_updated = gallery_merge_map_translate($source_row, $translations);
        $favpics = unserialize(base64_decode($source_row['user_favpics']));
        foreach ($favpics AS $key => $pid) {
          $favpics[$key] = $source_mapping_pictures[$pid]['target_pid'];
          if ($debug_level >= 3) { $debug_trace[] = "Translation for 'user_favpics imbedded pid' Value: {$pid} using Translate table 'pictures' key 'pid' ($pid) - field 'pid' converts to {$favpics[$key]}"; }
        }
        if (!array_key_exists($match_key, $target_array)) {
          // insert - encode array and generate sql
          $source_row_updated['user_favpics'] = base64_encode(serialize($favpics));
          $merge_sql = gallery_merge_generate_sql($target_table, $source_row_updated, $skip_keys, $zero_keys, $insert_ignore);
          $count_inserted ++;
          $count_total_inserted ++;
        } else {
          // update - we need to combine SOURCE and TARGET information
          $favpics_target = unserialize(base64_decode($target_array[$match_key]['user_favpics']));
          $favpics = array_merge($favpics_target, $favpics);                    // merge arrays
          $favpics = array_values(array_unique($favpics));                      // eliminate duplicates and resequence
          $source_row_updated['user_favpics'] = base64_encode(serialize($favpics));
          $merge_sql = "UPDATE $target_table SET user_favpics = '{$source_row_updated['user_favpics']}' WHERE user_id = '{$source_row_updated['user_id']}'";
          $inserted = FALSE;
          $count_existed ++;
          $count_total_inserted ++;                                             // count here for the QUIESCE limit
        }
        gallery_merge_debug_sql($merge_sql);  
        if ($plugin_gallery_merge_execute) {
          // execution mode!
          $insert_result = cpg_db_query($merge_sql);                            // execute SQL
        } else {
          // simulation mode
          $insert_id = $next_insert_id;                                         // set simulated id
          $next_insert_id ++;                                                   // and increment for next row
        }
      } else {
        // add other unique tables here... but until then...
        $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_merge_table']. $table_name);
        gallery_merge_error('db', $table_constant, 'error', $error_message);
      }
    }

    // build requested mapping array(s) if map_values were provided
    // filter whether to map pictures - based on primed array showing the keys we need
    if ((count($map_values) > '0') && (($table_name != 'pictures') || (($table_name == 'pictures') && (array_key_exists($source_row[$table_key], $GLOBALS[$map_var]))))) {
      gallery_merge_map_build($source_row, $source_row_updated, $target_array[$match_key], $table_name, $map_values, $insert_id);
    }

    // checkpoint and quiesce check and processing
    if ((($count_total_inserted % $CONFIG['plugin_gallery_merge_checkpoint_quiesce']) == 0) && ($count_total_inserted > 0)) {
      // take a checkpoint and quiesce processing - limit reached
      $error_message = sprintf($lang_plugin_gallery_merge['e_processing_quiesced'], $count_total_inserted, $count_inserted, $table_name);
      gallery_merge_error('db', $table_constant, 'error', $error_message, 'nobypass');
      gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:QUIESCE:$count_inserted:LASTKEY:{$source_row[$table_key]}", $$map_var);
      $debug_trace[] = "$table_name table processing QUIESCED";
      break;                                                                    // exit the WHILE loop...
    }
    if ((($count_inserted % $CONFIG['plugin_gallery_merge_checkpoint_freq']) == 0) && ($count_inserted > 0)) {
      // take a checkpoint and continue on...
      gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:INSERTED:$count_inserted:LASTKEY:{$source_row[$table_key]}", $$map_var);
    }
  } // end while
  unset ($target_array);
  cpg_db_free_result($source_result);
  $col2 = sprintf($lang_plugin_gallery_merge['inserted_existed'], $count_inserted, $count_existed, $table_name);
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "$table_name table processing complete";
}


function gallery_merge_check_rebuild_category_tree() {
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;
  // categories inserted without setting up tree structures (lft, rgt, depth) and pos may contain gaps due to duplicates between source/target that weren't inserted...
  // cpg has a function to rebuild tree structure when lft=0 as we inserted it...
  check_rebuild_tree();
  $col2 = $lang_plugin_gallery_merge['rebuild_category'];
  gallery_merge_output_status('blank', $col2);
}


function gallery_merge_check_rebuild_album_pos() {
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;
  // scan albums for any changes needed in pos field... We added fixed increment to each pos based on number of albums in the category at the start...
  // but an already existing album would have left a gap. CPG should fix on next album add/delete/sort - but let's leave it clean.
  $lastcat = FALSE;
  $count = 0;
  $result = cpg_db_query("SELECT * FROM {$CONFIG['TABLE_ALBUMS']} ORDER BY category, pos");
  while ($target_row = cpg_db_fetch_row($result)) {
    if ($target_row['category'] != $lastcat) {
      // new category - reset pos
      $pos = '100';
      $lastcat = $target_row['category'];
    }
    if ($target_row['pos'] != $pos) {
      // mismatch between set pos and what it should be..
      $update_sql = "UPDATE {$CONFIG['TABLE_ALBUMS']} SET pos = '$pos' WHERE aid = '{$target_row['aid']}'";
      gallery_merge_debug_sql($update_sql);
      $count ++;
      if ($plugin_gallery_merge_execute) {
        // execution mode!
        $result = cpg_db_query($update_sql);
      }
    } // end if
    $pos ++;
  }
  $col2 = sprintf($lang_plugin_gallery_merge['rebuild_album'], $count);
  gallery_merge_output_status('blank', $col2);
}


// update existing pictures combining hits, keywords
function gallery_merge_process_albums_existing($source_row, $target_row) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;
  global $source_mapping_pictures,   $source_mapping_pictures_key;              // access to pictures mapping table and key
  // for existing albums already in TARGET (name matches) - we want to merge statistics concerning hits and set keyword if possible
  // hits is simple sums
  $album_hits = $source_row['alb_hits'] + $target_row['alb_hits'];
  // see if we need to update keywords...
  $keyword = $target_row['keyword'];
  if ($source_row['keyword'] != '') {
    // source has keywords...
    if ($target_row['keyword'] == '') {
      // simple case - copy source to target
      $keyword = $source_row['keyword'];
    } else {
      // target already has a keyword... album can only have one... compare differences will already point out the conflict.
    }
  } // end keyword to merge
  $update_sql = "UPDATE {$CONFIG['TABLE_ALBUMS']} SET alb_hits = '$album_hits', keyword = '$keyword' WHERE pid = '{$target_row['pid']}'";
  gallery_merge_debug_sql($update_sql);
  if ($plugin_gallery_merge_execute) {
    // execution mode!
    $result = cpg_db_query($update_sql);
  }
}


// update existing pictures combining hits, votes, calculating new pic_rating, and merging keywords
function gallery_merge_process_pictures_existing($source_row, $target_row) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;
  global $source_mapping_pictures,   $source_mapping_pictures_key;              // access to pictures mapping table and key
  // for existing pictures already in TARGET (filename AND filepath match) - we want to merge statistics concerning hits and votes
  // hits and votes are simple sums
  $pic_hits = $source_row['hits'] + $target_row['hits'];
  $pic_votes = $source_row['votes'] + $target_row['votes'];
  // pic_rating calculated based on current rating and number of votes.
  // Formula for a new single vote from ratepic.php: (adapted for our simpler case)
  //   $rate = $superCage->get->getInt('rate');
  //   $rating_stars_amount = ($CONFIG['old_style_rating']) ? 5 : $CONFIG['rating_stars_amount'];
  //   $new_rating = round(($row['votes'] * $row['pic_rating'] + ($rate * (5 / $rating_stars_amount)) * 2000) / ($row['votes'] + 1));
  //   $sql = "UPDATE {$CONFIG['TABLE_PICTURES']} SET pic_rating = $new_rating, votes = votes + 1 WHERE pid = $pic";
  //   $result = cpg_db_query($sql);
  $rating_stars_amount = ($CONFIG['old_style_rating']) ? 5 : $CONFIG['rating_stars_amount'];
  if ($pic_votes > 0) {
    $pic_rating = round((($source_row['votes'] * $source_row['pic_rating']) + ($target_row['votes'] * $target_row['pic_rating'])) / $pic_votes);
  } else {
    $pic_rating = 0;
  }
  // see if we need to update keywords...
  $keywords = $target_row['keywords'];
  if ($source_row['keywords'] != '') {
    // source has keywords...
    if ($target_row['keywords'] == '') {
      // simple case - copy source to target
      $keywords = $source_row['keywords'];
    } else {
      // extract source and target keywords and compare/merge
      $target_keyword_list = explode($CONFIG['keyword_separator'], $target_row['keywords']);
      $source_keyword_list = explode($SOURCE['keyword_separator'], $source_row['keywords']);
      foreach ($source_keyword_list AS $source_keyword) {
        if (!in_array($source_keyword, $target_keyword_list)) {
          $keywords .= $CONFIG['keyword_separator'].$source_keyword;
        }
      } // end foreach
    }
  } // end keywords to merge
  $update_sql = "UPDATE {$CONFIG['TABLE_PICTURES']} SET hits = '$pic_hits', votes = '$pic_votes', pic_rating = '$pic_rating', keywords = '$keywords' WHERE pid = '{$target_row['pid']}'";
  gallery_merge_debug_sql($update_sql);
  if ($plugin_gallery_merge_execute) {
    // execution mode!
    $result = cpg_db_query($update_sql);
  }
}


// update thumb references in categories and albums now that pictures are inserted
function gallery_merge_process_thumbs() {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;
  global $source_mapping_categories, $source_mapping_categories_key;            // access to categories mapping table and key
  global $source_mapping_albums,     $source_mapping_albums_key;                // access to albums mapping table and key
  global $source_mapping_pictures,   $source_mapping_pictures_key;              // access to pictures mapping table and key
  $count_updated_cat = 0;                                                       // initialize updated counter
  $count_updated_alb = 0;                                                       // initialize updated counter
  // pictures array now populated with target ids for the ones we care about
  // since we've already processed categories and albums - go back and update with correct reference (was set to 0)
  // Categories - thumb references > 0
  // Albums - thumb references > 0
  $col1 = $lang_plugin_gallery_merge['processing_references'];
  $col2 = $lang_plugin_gallery_merge['updating_references'];
  gallery_merge_output_status($col1, $col2);
  // process categories
  foreach ($source_mapping_categories AS $category) {
    if ($category['thumb'] > 0) {
      $update_sql = "UPDATE {$CONFIG['TABLE_CATEGORIES']} SET thumb = '{$source_mapping_pictures[$category['thumb']]['target_pid']}' WHERE cid = '{$category['target_cid']}'";
      gallery_merge_debug_sql($update_sql);
      if ($plugin_gallery_merge_execute) {
        // execution mode!
        $result = cpg_db_query($update_sql);
      }
      $count_updated_cat ++;
    } // end if
  } // end foreach
  // process albums
  foreach ($source_mapping_albums AS $album) {
    if ($album['thumb'] > 0) {
      $update_sql = "UPDATE {$CONFIG['TABLE_ALBUMS']} SET thumb = '{$source_mapping_pictures[$album['thumb']]['target_pid']}' WHERE aid = '{$album['target_aid']}'";
      gallery_merge_debug_sql($update_sql);
      if ($plugin_gallery_merge_execute) {
        // execution mode!
        $result = cpg_db_query($update_sql);
      }
      $count_updated_alb ++;
    } // end if
  } // end foreach
  $col2 = sprintf($lang_plugin_gallery_merge['references_updated'], $count_updated_cat, $count_updated_alb);
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "thumbnail reference processing complete";
}


// process compare of data tables - used for tables like CONFIG where we only compare and report values
// expects fields `name` and `value` only...
function gallery_merge_process_compare_values($table_constant, $field_name = 'name', $field_value = 'value') {
  // function to compare 'configuration' type tables - containing a column with variable name - and a column with the value.
  // defaults to 'name' and 'value' as column names
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  $source_table = $SOURCE[$table_constant];
  $target_table = $CONFIG[$table_constant];
  $table_name_uc = substr($table_constant, 6);
  $table_name = strtolower($table_name_uc);
  $col1 = sprintf($lang_plugin_gallery_merge['processing_table'], $table_name);
  $col1 .= gallery_merge_db_doc($table_name);
  $col2 = $lang_plugin_gallery_merge['no_merge'];
  gallery_merge_output_status($col1, $col2);
  // this routine might be used for other 'config like' tables (plugins)
  // if main config table - target data already loaded... otherwise - go get it
  if ($table_name == 'config') {
    // TARGET config table is contained in $CONFIG array - no need to obtain again
    $target_array = 'CONFIG';                                                   // point to CONFIG array
  } else {
    // get target data...
    $target_result = cpg_db_query("SELECT $field_name, $field_value FROM {$target_table}");
    while ( ($row = cpg_db_fetch_row($target_result)) ) {
      $target_array_loaded[$row[$field_name]] = $row[$field_value];
    }
    cpg_db_free_result($target_result);
    $target_array = 'target_array_loaded';                                      // point to loaded data
    if ($debug_level >= 3) { $debug_trace[] = "<pre>$table_constant ".print_r($target_array_loaded, TRUE).'</pre>'; }
  }

  // Retrieve SOURCE DB stored configuration - and compare to TARGET
  $result = cpg_db_query("SELECT * FROM {$SOURCE[$table_constant]}", $SOURCE['LINK_ID']);
  while ($source_row = cpg_db_fetch_row($result))  {
    if (!array_key_exists($source_row[$field_name], $$target_array)) {
      // SOURCE config value not in TARGET DB
      $col2 = sprintf($lang_plugin_gallery_merge['source_not_in_target'], $table_name_uc, "'{$source_row[$field_name]}' = '{$source_row[$field_value]}'");
      if (substr($source_row[$field_name], 0, 7) == 'plugin_') {
        $col2 .= "<br>".$lang_plugin_gallery_merge['plugin_related'];
      }
      gallery_merge_output_status('blank', $col2);
    } elseif (${$target_array}[$source_row[$field_name]] != $source_row[$field_value]) {
      // SOURCE config value not equal to TARGET setting
      $col2 = sprintf($lang_plugin_gallery_merge['source_does_not_match'], $table_name_uc, "'{$source_row[$field_name]}' = '{$source_row[$field_value]}'", ${$target_array}[$source_row[$field_name]]);
      if (substr($source_row[$field_name], 0, 7) == 'plugin_') {
        $col2 .= "<br>".$lang_plugin_gallery_merge['plugin_related'];
      }
      gallery_merge_output_status('blank', $col2);
    }
  } // while
  cpg_db_free_result($result);
  unset($target_rowset);
  $col2 = sprintf($lang_plugin_gallery_merge['table_complete'], $table_name);
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "$table_name table processing complete";
}


// process plugin table - a table (or rather process) like no other table...
function gallery_merge_process_plugin($table_constant) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;
  global $plugin_array, $source_mapping_plugins;
  $source_table = $SOURCE[$table_constant];
  $target_table = $CONFIG[$table_constant];
  $table_name_uc = substr($table_constant, 6);
  $table_name = strtolower($table_name_uc);
  $col1 = sprintf($lang_plugin_gallery_merge['processing_table'], $table_name);
  $col1 .= gallery_merge_db_doc($table_name);
  $col2 = $lang_plugin_gallery_merge['no_merge'];
  $col2 .= "<br>".$lang_plugin_gallery_merge['no_merge_plugin_note']."<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;".$lang_plugin_gallery_merge['no_merge_plugin_note2']."<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;".$lang_plugin_gallery_merge['no_merge_plugin_note2a'];
  gallery_merge_output_status($col1, $col2);
  $plugin_alter_error_issued = FALSE;
  $plugin_alter_warning_issued = FALSE;
  $plugin_addtable_error_issued = FALSE;
  $plugin_addtable_warning_issued = FALSE;
  // Retrieve TARGET DB info
  // plugin names may vary in database by language used at time of install.. compare based on path (foldername)
  $result = cpg_db_query("SELECT name, path FROM {$CONFIG[$table_constant]}");
  while ($target_row = cpg_db_fetch_row($result)) {
    $target_array[$target_row['path']] = $target_row;
  }
  cpg_db_free_result($result);
  // Retrieve SOURCE DB info - and compare to TARGET
  $result = cpg_db_query("SELECT name, path FROM {$SOURCE[$table_constant]}", $SOURCE['LINK_ID']);
  while ( ($source_row = cpg_db_fetch_row($result)) ) {
    // see what we know about this plugin...
    $comment_not_installed = '';
    if (array_key_exists($source_row['path'], $plugin_array)) {
      // see if it alters cpg core tables
      if ($plugin_array[$source_row['path']]['alters_cpg_tables'] == TRUE) {
        $comment_not_installed = $lang_plugin_gallery_merge['plugin_alters_cpg_tables'].'<br>';
      }
      // see if it adds its own tables
      $merge_script = FALSE;
      if ($plugin_array[$source_row['path']]['adds_tables'] == TRUE) {
        if (($plugin_array[$source_row['path']]['merge_script'] == '') || ($plugin_array[$source_row['path']]['merge_script'] == 'needed')) {
          $comment_key = 'plugin_adds_data_warning';                            // no merge script available...
        } elseif ($plugin_array[$source_row['path']]['merge_script'] == 'n/a') {
          $comment_key = 'plugin_data_merge_not_required';                      // no merge script needed (shouldn't be used in this case - define a script even if tables transient - avoids warning messages later for unknown table)
        } else {
          $comment_not_installed .= $lang_plugin_gallery_merge['plugin_adds_data_not_installed'].'<br>';
          $comment_key = 'plugin_data_merge_supported';                         // merge script provided
          $merge_script = TRUE;
        }
      } else {
        $comment_key = 'plugin_no_data_impact';
      }
    } else {
      // unknown plugin (not in our array - see data/plugin_data.php to add...)
      $comment_key = 'plugin_unknown';
      $comment_not_installed_key = 'plugin_unknown_alter';
    }

    // see if it is installed in target
    if (!array_key_exists($source_row['path'], $target_array)) {
      // SOURCE info not in TARGET DB
      $col2 = sprintf($lang_plugin_gallery_merge['plugin_not_installed'], "{$source_row['name']} ({$source_row['path']})");
      $col2 .= "<br>{$comment_not_installed}{$lang_plugin_gallery_merge[$comment_key]}";
      gallery_merge_output_status('blank', $col2);
      // issue errors/warning if missing plugin alters CPG tables
      if (($plugin_gallery_merge_execute) && ($plugin_array[$source_row['path']]['alters_cpg_tables'] == TRUE) && (!$plugin_alter_error_issued)) {
        // execution mode - abort
        $plugin_alter_error_issued = TRUE;                                      // avoid issuing error more han once for this error
        $error_message = $lang_plugin_gallery_merge['e_missing_plugin_alter'];
        gallery_merge_error('db', $table_constant, 'error', $error_message);
      } elseif (($plugin_array[$source_row['path']]['alters_cpg_tables'] == TRUE) && (!$plugin_alter_warning_issued)) {
        // simulation mode - warn
        $plugin_alter_warning_issued = TRUE;                                    // avoid issuing warning more than once for this error
        $warning_message = $lang_plugin_gallery_merge['w_missing_plugin_alter']."<br />- <a href='pluginmgr.php'>".$lang_plugin_gallery_merge['w_use_pluginmgr']."</a> ".$lang_plugin_gallery_merge['w_rerun_simulation'];
        gallery_merge_error('db', $table_constant, 'warning', $warning_message);
      }
      // issue errors/warning if missing plugin adds tables and merge is supported
      if (($plugin_gallery_merge_execute) && ($plugin_array[$source_row['path']]['adds_tables'] == TRUE) && $merge_script && (!$plugin_addtable_error_issued)) {
        // execution mode - abort
        $plugin_addtable_error_issued = TRUE;                                      // avoid issuing error more han once for this error
        $error_message = $lang_plugin_gallery_merge['e_missing_plugin_addtable'];
        gallery_merge_error('db', $table_constant, 'error', $error_message);
      } elseif (($plugin_array[$source_row['path']]['adds_tables'] == TRUE) && $merge_script && (!$plugin_addtable_warning_issued)) {
        // simulation mode - warn
        $plugin_addtable_warning_issued = TRUE;                                    // avoid issuing warning more than once for this error
        $warning_message = $lang_plugin_gallery_merge['w_missing_plugin_addtable']."<br />- <a href='pluginmgr.php'>".$lang_plugin_gallery_merge['w_use_pluginmgr']."</a> ".$lang_plugin_gallery_merge['w_rerun_simulation'];
        gallery_merge_error('db', $table_constant, 'warning', $warning_message);
      }
    } else {
      // SOURCE info already in TARGET DB
      // add to mapping array as an installed plugin - capture merge_script and pid_sql to allow use later
      $source_mapping_plugins[$source_row['path']]['name'] = $plugin_array[$source_row['path']]['name'];
      $source_mapping_plugins[$source_row['path']]['merge_script'] = $plugin_array[$source_row['path']]['merge_script'];
      $source_mapping_plugins[$source_row['path']]['pid_sql'] = $plugin_array[$source_row['path']]['pid_sql'];
      //(should be low volume - so write out all regardless of msglevel - also allows writing plugin specific comment
      $col2 = sprintf($lang_plugin_gallery_merge['plugin_already_installed'], "{$source_row['name']} ({$source_row['path']})");
      $col2 .= "<br>$lang_plugin_gallery_merge[$comment_key]";
      gallery_merge_output_status('blank', $col2);
    }
  } // while
  cpg_db_free_result($result);
  unset($target_array);
  $col2 = sprintf($lang_plugin_gallery_merge['table_complete'], $table_name);
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "$table_name table processing complete";
}


// process static data tables
function gallery_merge_process_static($table_constant) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  $source_table = $SOURCE[$table_constant];
  $target_table = $CONFIG[$table_constant];
  $table_name = strtolower(substr($table_constant, 6));
  $col1 = sprintf($lang_plugin_gallery_merge['processing_table'], $table_name);
  $col1 .= gallery_merge_db_doc($table_name);
  $col2 = $lang_plugin_gallery_merge['static_table'];
  // Retrieve TARGET DB info
  $target_result = cpg_db_query("SELECT COUNT( * ) FROM {$target_table}");
  $target_count = cpg_db_result($target_result, 0, 0);
  cpg_db_free_result($target_result);
  // Retrieve SOURCE DB stored configuration - and compare to TARGET
  $source_result = cpg_db_query("SELECT COUNT( * ) FROM {$source_table}", $SOURCE['LINK_ID']);
  $source_count = cpg_db_result($source_result, 0, 0);
  cpg_db_free_result($source_result);
  if ($target_count == $source_count) {
    // source and target row counts equal
    if ($message_level == 2) {
      $col2 .= "<br>".sprintf($lang_plugin_gallery_merge['static_rows_match'], $table_name, $source_count);
    }
  } else {
    // mismatch in row counts
    $col2 .= "<br>".sprintf($lang_plugin_gallery_merge['static_rows_mismatch'], $table_name, $source_count, $target_count);
  }
  gallery_merge_output_status($col1, $col2);
  $col2 = sprintf($lang_plugin_gallery_merge['table_complete'], $table_name);
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "$table_name table processing complete";
}


// process optional data tables
function gallery_merge_process_optional($table_constant) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute;
  // unique case - currently used for DICT table - if used - rebuild.
  // code will need to be added for each table calling this to properly rebuild
  $source_table = $SOURCE[$table_constant];
  $target_table = $CONFIG[$table_constant];
  $table_name = strtolower(substr($table_constant, 6));
  $col1 = sprintf($lang_plugin_gallery_merge['processing_table'], $table_name);
  $col1 .= gallery_merge_db_doc($table_name);
  $col2 = $lang_plugin_gallery_merge['optional_table'];
  // Retrieve TARGET DB stored configuration - and display counts
  $result = cpg_db_query("SELECT COUNT( * ) FROM {$target_table}");
  $target_count = cpg_db_result($result, 0, 0);
  cpg_db_free_result($result);
  // Retrieve SOURCE DB stored configuration - and display counts
  $result = cpg_db_query("SELECT COUNT( * ) FROM {$source_table}", $SOURCE['LINK_ID']);
  $source_count = cpg_db_result($result, 0, 0);
  cpg_db_free_result($result);
  if (($source_count == 0) && ($target_count == 0)) {
    $action = $lang_plugin_gallery_merge['optional_no_action'];
  } else {
    $action = $lang_plugin_gallery_merge['optional_rebuild'];
  }
  if (($source_count > 0) || ($target_count > 0) || ($message_level == 2)) {
    $col2 .= "<br>".sprintf($lang_plugin_gallery_merge['optional_counts'], $table_name, $source_count, $target_count)." - $action";
  }
  gallery_merge_output_status($col1, $col2);
  if (($source_count > 0) || ($target_count > 0)) {
    switch($table_name) {
      case 'dict':
        $code_file = './keyword_create_dict.php';
        if (file_exists($code_file)) {
          // found code..
          // cpg script has include for include/init.inc.php that will cause duplicate function definitions
          // delete that reference... and execute
          $source_dict_rebuild = file_get_contents($code_file);
          $source_dict_rebuild = str_replace("require('include/init.inc.php');", '/* removed by gallery merge */', $source_dict_rebuild);
          $source_dict_rebuild = str_replace('<?php', '', $source_dict_rebuild);
          $source_dict_rebuild = str_replace('?>', '', $source_dict_rebuild);
          if ($plugin_gallery_merge_execute) {
            ob_start();
            eval ($source_dict_rebuild);
            $debug_trace[] = 'DICT rebuild - rows: '.ob_end_clean();            // direct output to debug
          } else {
            $debug_trace[] = 'DICT rebuild - bypassed for simulation';          // simulation only...
          }
          unset ($source_dict_rebuild);
          // Retrieve TARGET DB stored configuration - and display counts
          $result = cpg_db_query("SELECT COUNT( * ) FROM {$target_table}");
          $target_count = cpg_db_result($result, 0, 0);
          cpg_db_free_result($result);
          $col2 = sprintf($lang_plugin_gallery_merge['optional_rebuild_complete'], $target_count);
          gallery_merge_output_status('blank', $col2);
        } else {
          $error_message = sprintf($lang_plugin_gallery_merge['e_optional_rebuild_failed'], $table_name);
          gallery_merge_error('db', $table_constant, 'error', $error_message);
          $debug_trace[] = "$table_name rebuild processing failed - unable to locate support file";
        }
      break;

      default:
        $error_message = sprintf($lang_plugin_gallery_merge['e_optional_rebuild_failed'], $table_name);
        gallery_merge_error('db', $table_constant, 'error', $error_message);
        $debug_trace[] = "$table_name invalid for rebuild processing";
      break;
    }
  }
  $col2 = sprintf($lang_plugin_gallery_merge['table_complete'], $table_name);
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "$table_name table processing complete";
}


// process transient data tables
function gallery_merge_process_transient($table_constant) {
  global $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  $source_table = $SOURCE[$table_constant];
  $table_name = strtolower(substr($table_constant, 6));
  $col1 = sprintf($lang_plugin_gallery_merge['processing_table'], $table_name);
  $col1 .= gallery_merge_db_doc($table_name);
  $col2 = $lang_plugin_gallery_merge['transient_table'];
  // No TARGET DB info needed
  // Retrieve SOURCE DB stored configuration - and display counts
  $result = cpg_db_query("SELECT COUNT( * ) FROM {$source_table}", $SOURCE['LINK_ID']);
  $source_count = cpg_db_result($result, 0, 0);
  cpg_db_free_result($result);
  if (($source_count > 0) || ($message_level == 2)) {
    $col2 .= "<br>".sprintf($lang_plugin_gallery_merge['table_count'], $table_name, $source_count);
  }
  gallery_merge_output_status($col1, $col2);
  $col2 = sprintf($lang_plugin_gallery_merge['table_complete'], $table_name);
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "$table_name table processing complete";
}


// process unused data tables
function gallery_merge_process_unused($table_constant) {
  global $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  $source_table = $SOURCE[$table_constant];
  $table_name = strtolower(substr($table_constant, 6));
  $col1 = sprintf($lang_plugin_gallery_merge['processing_table'], $table_name);
  $col1 .= gallery_merge_db_doc($table_name);
  $col2 = $lang_plugin_gallery_merge['unused_table'];
  // No TARGET DB info needed
  // Retrieve SOURCE DB stored configuration - and display counts
  $result = cpg_db_query("SELECT COUNT( * ) FROM {$source_table}", $SOURCE['LINK_ID']);
  $source_count = cpg_db_result($result, 0, 0);
  cpg_db_free_result($result);
  if (($source_count > 0) || ($message_level == 2)) {
    $col2 .= "<br>".sprintf($lang_plugin_gallery_merge['table_count'], $table_name, $source_count);
  }
  gallery_merge_output_status($col1, $col2);
  $col2 = sprintf($lang_plugin_gallery_merge['table_complete'], $table_name);
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "$table_name table processing complete";
}


// process unsupported data tables
function gallery_merge_process_unsupported($table_constant) {
  global $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  $source_table = $SOURCE[$table_constant];
  $table_name = strtolower(substr($table_constant, 6));
  $col1 = sprintf($lang_plugin_gallery_merge['processing_table'], $table_name);
  $col1 .= gallery_merge_db_doc($table_name);
  $col2 = $lang_plugin_gallery_merge['unsupported_table'];
  // No TARGET DB info needed
  // Retrieve SOURCE DB stored configuration - and display counts
  $result = cpg_db_query("SELECT COUNT( * ) FROM {$source_table}", $SOURCE['LINK_ID']);
  $source_count = cpg_db_result($result, 0, 0);
  cpg_db_free_result($result);
  if (($source_count > 0) || ($message_level == 2)) {
    $col2 .= "<br>".sprintf($lang_plugin_gallery_merge['table_count'], $table_name, $source_count);
  }
  gallery_merge_output_status($col1, $col2);
  $col2 = sprintf($lang_plugin_gallery_merge['table_complete'], $table_name);
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "$table_name table processing complete";
}


// process unknown data tables - see if any exist that we don't know about... and report
function gallery_merge_process_unknown() {
  global $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_array, $source_mapping_plugins;
  $col1 = sprintf($lang_plugin_gallery_merge['processing_table'], 'unknown tables');
  $col2 = $lang_plugin_gallery_merge['unknown_tables'];
  gallery_merge_output_status($col1, $col2);
  // build pattern_array for additional known table masks (that follow CPG table prefix) from main plugin_array
  // faster seaching with just data we need vs multideminsional array
  $pattern_array = array();
  $pattern_array[$CONFIG['plugin_gallery_merge_backout_prefix']] = 'Gallery Merge:unknown_plugin_gallery_merge_bkup';  // add our own backup files to handle properly
  foreach ($plugin_array AS $plugin_folder => $plugin_detail_array) {
    $plugin_name = $plugin_detail_array['name'];
    if ($plugin_detail_array['adds_tables']) {
      // plugin has tables we care about - let's set up the right message keys...
      if (($plugin_detail_array['merge_script'] == '') || ($plugin_detail_array['merge_script'] == 'needed')) {
        $merge = 'needed';                                                      // plugin has no merge script...
      } elseif ($plugin_detail_array['merge_script'] == 'n/a') {
        $merge = 'nomerge';                                                     // plugin data says no merge required
      } else {
        $merge = 'merge';                                                       // plugin has merge script
      }
      if (key_exists($plugin_folder, $source_mapping_plugins)) {
        $install = 'install';                                                   // plugin is installed
      } else {
        $install = 'uninst';                                                    // plugin not installed
      }
      $message_key = "unknown_plugin_{$install}_{$merge}";
      $pattern_array[$plugin_detail_array['table_pattern']] = "$plugin_name:$message_key";
    }
  }
  if ($debug_level >= 3) { $debug_trace[] = '<pre>PLUGIN TABLES PATTERN '.print_r($pattern_array, TRUE).'</pre>'; }

  $result = cpg_db_query("SHOW FULL TABLES LIKE '{$SOURCE['TABLE_PREFIX']}%'", $SOURCE['LINK_ID']);
  $rowset = cpg_db_fetch_rowset($result);
  cpg_db_free_result($result);
  foreach ($rowset AS $row) {
    $tablearray = each ($row );                                                 // don't know the key - contains DBNAME and LIKE criteria - so use each to extract value
    $tabletype = $row['Table_type'];                                            // we know the key for the type... BASE TABLE or VIEW
    if (!in_array($tablearray['value'], $SOURCE)) {
      // a table we haven't processed...
      // see if it matches a known name pattern
      $known = FALSE;
      foreach ($pattern_array AS $pattern_suffix => $description) {
        $pattern = $SOURCE['TABLE_PREFIX'].$pattern_suffix;
        $patternlength = strlen($pattern);
        if ($pattern == substr($tablearray['value'], 0, $patternlength)) {
          $known = TRUE;
          $comment =  explode(':', $description);                               // extract plugin name and language key
          break;
        } // end if
      } // end foreach pattern_array
      if ($known) {
        $col2 = sprintf($lang_plugin_gallery_merge['unknown_plugin_table'], $tablearray['value']).'<br>'.sprintf($lang_plugin_gallery_merge[$comment[1]], $comment[0]);
      } else {
        $col2 = sprintf($lang_plugin_gallery_merge['unknown_table'], $tablearray['value']);
        $debug_trace[] = "Unknown table {$tablearray['value']} found";
      }
      // Retrieve SOURCE DB stored configuration - and display counts
      $result = cpg_db_query("SELECT COUNT( * ) FROM `{$tablearray['value']}`", $SOURCE['LINK_ID']);
      $source_count = cpg_db_result($result, 0, 0);
      cpg_db_free_result($result);
      if (($source_count > 0) || ($message_level == 2)) {
        $col2 .= "<br>".sprintf($lang_plugin_gallery_merge['table_count'], $tablearray['value'], $source_count);
      }
      gallery_merge_output_status('blank', $col2);
    } // end if !in_array
  } // end foreach rowset
  unset($pattern_array);
  $col2 = sprintf($lang_plugin_gallery_merge['table_complete'], 'UNKNOWN');
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "unknown table processing complete";
}


// verify table contents
function gallery_merge_verify_table($table_constant, $compare_field, $compare_field2 = '') {
/*
      SAMPLE input variables:
        $table_constant = 'TABLE_USERGROUPS';
        $compare_field = 'group_name';                                          // field to match source to target (identify duplicates/inserts)
        $compare_field2 = '';                                                   // second field to match source to target (identify duplicates/inserts)
*/
  global $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $CONFIG, $SOURCE;
  $count_missing = '0';                                                         // initialize inserted counter
  $count_existed = '0';                                                         // initialize existed counter
  $count_ignored = '0';                                                         // initialize existed counter
  $compare_error_issued = FALSE;
  $source_table = $SOURCE[$table_constant];                                     // source table name
  $target_table = $CONFIG[$table_constant];                                     // target table name
  $table_name_uc = substr($table_constant, 6);                                  // uppercase table constant for messages
  $table_name = strtolower($table_name_uc);                                     // lowercase table constant for messages
  $map_array = array();                                                         // array if needed for mapping (used by pictures for userpics)

  $col1 = sprintf($lang_plugin_gallery_merge['verifying_table'], $table_name);
  $col1 .= gallery_merge_db_doc($table_name);
  $col2 = $lang_plugin_gallery_merge['verifying_table_comment'];
  gallery_merge_output_status($col1, $col2);

  // set SQL parms
  $select = "$compare_field";
  if ($compare_field2 != '') {
    $select .= ", $compare_field2";
  }

  // Retrieve TARGET DB info
  // load target data into array - only one or two fields - so much smaller arrays here
  $target_result = cpg_db_query("SELECT {$select} FROM {$target_table}");
  while ($target_row = cpg_db_fetch_row($target_result)) {
    if ($compare_field2 != '') {
      $target_array["{$target_row[$compare_field]}-{$target_row[$compare_field2]}"] = $target_row[$compare_field];
    } else {
      $target_array[$target_row[$compare_field]] = $target_row[$compare_field];
    }
  }
  cpg_db_free_result($target_result);

  if (($message_level == 2) && ($table_name == 'pictures')) {
    // suppress individual MATCHED messages for pictures... just too many.
    $col2 = sprintf($lang_plugin_gallery_merge['match_messages_suppressed'], $table_name_uc);
    gallery_merge_output_status('blank', $col2);
  }
  // Retrieve SOURCE DB info - and compare to TARGET
  $source_result = cpg_db_query("SELECT {$select} FROM {$source_table} {$orderby}", $SOURCE['LINK_ID']);
  while ( ($source_row = cpg_db_fetch_row($source_result)) ) {

    // special processing for pictures - not handled in normal map translation (partial field replacements)
    // - if adding a subfolder was requested... add to all pics EXCEPT those in 'userpics' folders ('edit' folder completely bypassed)
    // - if removing a subfolder was requested... remove from all places it exists (shouldn't exist in userpics)
    // - if userpics entry, translate user related folder name based on remapped user
    if ($table_name == 'pictures') {
      $source_row['filepath'] = gallery_merge_set_target_dest($source_row['filepath']);
    }

    // We may need to compare one field or two... check both possibilities (target array will be built to match one of these)
    if ((!array_key_exists($source_row[$compare_field], $target_array)) && (!array_key_exists("{$source_row[$compare_field]}-{$source_row[$compare_field2]}", $target_array))) {
      // SOURCE info not in TARGET DB
      // Oops - missing data...)
      $col2 = sprintf($lang_plugin_gallery_merge['verify_missing'], $table_name_uc, $source_row[$compare_field]);
      gallery_merge_output_status('blank', $col2);
      if (!$compare_error_issued) {
        $compare_error_issued = TRUE;
        $error_message = sprintf($lang_plugin_gallery_merge['e_verify_table_missing_data'], $table_name);
        gallery_merge_error('db', "COMPARE_$table_constant", 'error', $error_message);
      }
      $count_missing ++;
      if ($count_missing > '99') {
        $error_message = sprintf($lang_plugin_gallery_merge['e_verify_table_aborted'], $count_missing);
        gallery_merge_error('db', "COMPARE_$table_constant", 'error', $error_message);
        break;
      }
    } else {
      // SOURCE info already in TARGET DB
      if (($message_level == 2) && ($table_name != 'pictures')) {
        $col2 = sprintf($lang_plugin_gallery_merge['verify_match'], $table_name_uc, $source_row[$compare_field]);
        gallery_merge_output_status('blank', $col2);
      }
      $count_existed ++;
    }
  } // end while
  unset ($target_array);
  cpg_db_free_result($source_result);
  $col2 = sprintf($lang_plugin_gallery_merge['verify_counts'], $table_name, $count_missing, $count_existed);
  gallery_merge_output_status('blank', $col2);
  $col2 = sprintf($lang_plugin_gallery_merge['table_complete'], $table_name);
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "$table_name table processing complete";
}


// *********************
// Merge Files Functions
// *********************

//traverse directory structure returning array(s) with directories, files, or both.
function gallery_merge_traverse($dir, &$dir_array, &$file_array, $mode, $lev) {
  global $CONFIG, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  // mode = dir | files | all  -  indicates what types to write to array
  // lev  = current | all      -  show current level only or all lower levels
  // filter = '' | string      -  string that must appear in filename to include
  // function is recursive to traverse all lower level directories
  if (is_dir($dir))
    {
    $fd = @opendir($dir);
    if($fd) {
      while (($part = @readdir($fd)) == true) {
        if ($part != "." && $part != "..") {
          if (is_dir($dir."/".$part)) {
            if ($debug_level == 4)  {  $debug_trace[] =  "Traverse: Found Directory ".$dir."/".$part;  }  //diag
            if (($mode == "dir") | ($mode == "all")) {
              // if directories requested
              //   exclude FP directories
              $pos = strpos($part, "_vti_");
              if ($pos === false) {
                // write out
                $dir_array[]=$dir."/".$part;
                if ($lev == "all") {
                  // and continue down the line
                  gallery_merge_traverse($dir."/".$part, $dir_array, $file_array, $mode, $lev, $filter);
                }
              }
            } // end if directories requested
          } // end if found lower directory
          else {
            if (($mode == "files") | ($mode == "all")){
              // if files requested
              $file_array[]=$dir."/".$part;    // write out - and return...
            }
          }
        } // end if not . or ..
      } // end while readdir
    } // end if $fd (directory opened)
  } else {
    // no directory found
    if ($debug_level == 4)  {  $debug_trace[] =  "Traverse: No directory found at ".$dir;  }  //diag
  }
  if ($dir_array){
    sort($dir_array);
    reset($dir_array);
  }
  if ($file_array){
    sort($file_array);
    reset($file_array);
  }
} //end function


// compare source and target directory - no recursion to lower levels (use gallery_merge_compare_directories for that) - report missing files/folders in target
function gallery_merge_compare_directory($compare) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute_files;
  global $plugin_array;
  $count = '0';
  // process the directory corresponding to the compare function requested - and report items missing from target
  // filepaths obtained from $CONFIG (our additions) and $SOURCE arrays.
  // $source_directory_array starts empty. Will contain list of directories in the requested directory
  // $source_file_array starts empty. Will contain list of files in the requested directory
  $source_directory_array = array();
  $source_file_array = array();
  // set standard variables based on type
  if ($compare == 'Language') {
    $source_filepath = $SOURCE['language_filepath'];
    $target_filepath = $CONFIG['language_filepath'];
    $mode = 'files';
    $copy_request = $CONFIG['plugin_gallery_merge_copy_languages'];
  } elseif ($compare == 'Plugin') {
    $source_filepath = $SOURCE['plugins_filepath'];
    $target_filepath = $CONFIG['plugins_filepath'];
    $mode = 'dir';
    $copy_request = $CONFIG['plugin_gallery_merge_copy_plugins'];
    // additional processing for plugins...
    // build pattern_array for additional known folder masks from main plugin_array
    // faster seaching with just data we need vs multideminsional array
    $pattern_array = array();
    foreach ($plugin_array AS $plugin_folder => $plugin_detail_array) {
      // all plugins add a folder...
      $pattern_array[$plugin_folder] = $plugin_detail_array['name'];
    }
  } elseif ($compare == 'Theme') {
    $source_filepath = $SOURCE['themes_filepath'];
    $target_filepath = $CONFIG['themes_filepath'];
    $mode = 'dir';
    $copy_request = $CONFIG['plugin_gallery_merge_copy_themes'];
  }
  if ($copy_request == '1') {
    $action = sprintf($lang_plugin_gallery_merge['compare_directory_copy'], $compare);
  } else {
    $action = sprintf($lang_plugin_gallery_merge['compare_directory_nocopy'], $compare);
  }
  gallery_merge_traverse($source_filepath, $source_directory_array, $source_file_array, $mode, 'current');

  if ($mode == 'dir') {
    $process_array = 'source_directory_array';
  } else {
    $process_array = 'source_file_array';
  }
  foreach ($$process_array AS $source) {
    $target = str_replace($source_filepath, $target_filepath, $source);         // replace source filepath with target
    $name = str_replace("$target_filepath/", '', $target);                      // extract file/folder name
    if (file_exists($target)) {
      // exists in both source and target
      if ($debug_level >= 3)  {  $debug_trace[] =  "$compare $name already exists in TARGET gallery"; }
    } else {
      $col2 = sprintf($lang_plugin_gallery_merge['compare_directory_item'], $compare, $name).' - '.$action;
      if ($compare == 'Plugin') {
        if (array_key_exists($name, $pattern_array)) {
          $plugin =  $pattern_array[$name];                                     // extract plugin name
        } else {
          $plugin = 'UNKNOWN';
        }
        $col2 .= "<br>".sprintf($lang_plugin_gallery_merge['compare_plugin_adds_folder'], $plugin);
      }
      gallery_merge_output_status('blank', $col2);
      if ($debug_level >= 3)  {  $debug_trace[] =  "$compare $name missing from TARGET gallery"; }
      // see if we need to execute
      if (($plugin_gallery_merge_execute_files) && ($copy_request == '1')) {
        // missing file... execute mode... copy requested... go do it
        if ($mode == 'files') {
          gallery_merge_file_copy($source,$target);
        } else {
          gallery_merge_recurse_copy($source,$target);
        }
        $count ++;
        if ($debug_level >= 3)  {  $debug_trace[] =  "$compare $name copied to TARGET gallery"; }
      }
    }
  }
  if (($plugin_gallery_merge_execute_files) && ($copy_request == '1')) {
    $col2 = sprintf($lang_plugin_gallery_merge['compare_directory_copied'], $count, $compare);
    gallery_merge_output_status('blank', $col2);
  }
}


// compare cpg source and target filesystems recursively - report directory conflicts
// exclude directories otherwise processed (albums, themes, etc...)
function gallery_merge_compare_cpg($source_filepath) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $process_constant, $plugin_array;
  // process through all  non-excluded CPG folders/files and report any folders/files in source that are not in target
  // initial source_filepath is in $CONFIG array - but this routine is caled recursively following the directory structure
  // conflicting types (name is directory in one and file in other) flags error. Other conflicts reported only.
  // $source_directory_array starts empty. Will contain all directories/subdirectories searched
  // $source_file_array starts empty. Will contain all files found in common directories that need to be verified for conflicts
  static $exclude_array = array();
  static $pattern_array = array();

  if (empty($exclude_array)) {
    // initialize first time through...
    $exclude_array = array($SOURCE['language_filepath'],
                           $SOURCE['plugins_filepath'],
                           $SOURCE['themes_filepath'],
                           $SOURCE['album_filepath']);
    // build pattern_array for additional known folders (that don't use plugins folder) from main plugin_array
    // faster seaching with just data we need vs multideminsional array
    foreach ($plugin_array AS $plugin_name => $plugin_detail_array) {
      if ($plugin_detail_array['standard_folders'] == FALSE) {
        // plugin has folder we care about - doesn't use std folders in plugin directory
        $pattern_array[$plugin_detail_array['folder_name']] = $plugin_name;
      }
    }
    if ($debug_level >= 3)  {  $debug_trace[] = '<pre>PLUGINS FOLDERS PATTERN '.print_r($pattern_array, TRUE).'</pre>'; }
    // compare anycontent.php size...
    $source_anycontent_size = filesize("{$CONFIG['plugin_gallery_merge_source_filepath']}/anycontent.php");
    $target_anycontent_size = filesize("{$CONFIG['cpg_filepath']}/anycontent.php");
    if ($source_anycontent_size != $target_anycontent_size) {
        $col2 = sprintf($lang_plugin_gallery_merge['compare_anycontent_mismatch'], $source_anycontent_size, $target_anycontent_size);
        gallery_merge_output_status('blank', $col2);
    } else {
      if ($debug_level >= 3)  {  $debug_trace[] =  "Compare CPG: /anycontent.php is $source_anycontent_size in SOURCE and TARGET"; }
    }
  }
  $source_directory_array = array();
  $source_file_array = array();
  gallery_merge_traverse($source_filepath, $source_directory_array, $source_file_array, 'all', 'current');
  foreach ($source_file_array AS $source_file) {
    $target_file = str_replace($CONFIG['plugin_gallery_merge_source_filepath'], $CONFIG['cpg_filepath'], $source_file); // replace source filepath with target
    if ($debug_level == 4)  {  $debug_trace[] =  "Compare CPG: Source File: $source_file => Target File: $target_file"; }
    if (file_exists($target_file)) {
      // exists in both source and target
      if (!is_dir($target_file)) {
        // a file in both - we're done
      } else {
        // directory in source - and file in target... flag error
        $col2 = sprintf($lang_plugin_gallery_merge['compare_filetype_conflict'], $target_file);
        gallery_merge_output_status('blank', $col2);
        $error_message = sprintf($lang_plugin_gallery_merge['e_file_dir_mismatch'], $target_file);
        gallery_merge_error('files', $process_constant, 'error', $error_message);
      }
    } else {
      $subfile = str_replace($CONFIG['cpg_filepath'], '', $target_file);
      $col2 = sprintf($lang_plugin_gallery_merge['compare_file_missing'], $subfile);
      gallery_merge_output_status('blank', $col2);
    }
  }
  foreach ($source_directory_array AS $source_directory) {
    $target_directory = str_replace($CONFIG['plugin_gallery_merge_source_filepath'], $CONFIG['cpg_filepath'], $source_directory); // replace source filepath with target
    if ($debug_level >= 3)  {  $debug_trace[] =  "Compare CPG: Source Directory: $source_directory => Target Directory: $target_directory"; }
    if (!in_array($source_directory, $exclude_array)) {
      // directory not excluded...
      if (file_exists($target_directory)) {
        // exists in both source and target
        if (is_dir($target_directory)) {
          // a directory in both - continue recursively down the tree
          if ($debug_level == 4)  {  $debug_trace[] =  "Compare CPG: recursive call using directory: $source_directory"; }
          gallery_merge_compare_cpg($source_directory);
        } else {
          // directory in source - and file in target... flag error
          $col2 = sprintf($lang_plugin_gallery_merge['compare_dirtype_conflict'], $target_directory);
          gallery_merge_output_status('blank', $col2);
          $error_message = sprintf($lang_plugin_gallery_merge['e_file_dir_mismatch'], $target_directory);
          gallery_merge_error('files', $process_constant, 'error', $error_message);
        }
      } else {
        $subdirectory = str_replace("{$CONFIG['cpg_filepath']}/", '', $target_directory);
        $col2 = sprintf($lang_plugin_gallery_merge['compare_dir_missing'], $subdirectory);
        if (array_key_exists($subdirectory, $pattern_array)) {
          $col2 .= "<br>".sprintf($lang_plugin_gallery_merge['plugin_adds_folder'], $pattern_array[$subdirectory]);
       }
        gallery_merge_output_status('blank', $col2);
      }
    }
  }
}


// set target destination based on provided source - used for recursive copies of albums directory and sub-directory
function gallery_merge_set_target_dest_files($source) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute_files;
  global $process_constant;
  global $source_mapping_users, $source_mapping_albums;
  global $source_mapping_users_key, $source_mapping_albums_key;
  static $userpics_warn = FALSE;
  static $userpics_error = FALSE;
  // determine if we need to remove or add requested folder to source directory - or change a userpics directory/folder/subfolder
  // based on config variables 'plugin_gallery_merge_remove_subfolder' and 'plugin_gallery_merge_add_subfolder' and the actual directory name (edit and userpics are excluded...)
  // comparison names and lengths set in $CONFIG to speed processing - as we may get called a LOT.
  $target = $source;

  // process remove request - and not a userpics destination
  if (($CONFIG['plugin_gallery_merge_remove_subfolder'] != '') && ($SOURCE['compare_userpics'] != substr($target, 0, $SOURCE['compare_userpics_len']))) {
    if (($SOURCE['compare_remove'] == substr($target, 0, $SOURCE['compare_remove_len'])) || ($SOURCE['compare_remove2'] == $target)) {
      $target = str_replace($SOURCE['compare_remove2'], $SOURCE['album_filepath'], $target);     // update source filepath to remove requested folder
    }
  }

  // process userpics destination.. a little more involved
  if ($SOURCE['compare_userpics2'] == substr($source, 0, $SOURCE['compare_userpics2_len'])) {
    // have a userpics destination (may be just directory - may be directory + file...)
    $debug_text = "Userpics destination: Source folder: '$source' ";
    // extract the userid
    $source_extract = substr($source, $SOURCE['compare_userpics_len']);         // strip off path/albums/userpics/ (userpics vs userpics 2 intentional to remove /)
    if ($source_extract) {
      // was there anything after the prefix??
      $source_parts = explode("/", $source_extract, 3);                         // get the pieces of filepath - down to possible subfolder
      $userfolder = $source_parts[0];                                           // grab the first field - the user folder (info before already removed...)
      $albumfolder = $source_parts[1];                                          // grab the third field [0] - might be album folder...
    } else {
      $userfolder = FALSE;
    }
    // see if source and target userpics match..
    if ($CONFIG['userpics'] != $SOURCE['userpics']) {
      // userpics specification different - change it..
      $target = str_replace($SOURCE['userpics'], $CONFIG['userpics'], $target);
    }
    if (is_numeric($userfolder)) {
      // userfolder is numeric... continue on...
      $userid = $userfolder - 10000;                                            // subtract 10000 to get to user_id (unlike category - doesn't appear to be a constant for this value...)
      $debug_text .= "User: '$userid' ";
      if (key_exists($userid, $source_mapping_users)) {
        // seen this value before - use it..
        if ($source_mapping_users[$userid]['target_user_folder'] > 0) {
          $target_userfolder = $source_mapping_users[$userid]['target_user_folder'];
        } else {
          $target_userfolder = 0;
        }
      } else {
        // new user - go get translation
        $source_result = cpg_db_query("SELECT user_name FROM {$SOURCE['TABLE_USERS']} WHERE user_id = '$userid' LIMIT 1", $SOURCE['LINK_ID']);
        $source_row = cpg_db_fetch_row($source_result);                         // get user name from source
        $debug_text .= "Username: '{$source_row['user_name']}' ";
        cpg_db_free_result($source_result);
        $target_result = cpg_db_query("SELECT user_id FROM {$CONFIG['TABLE_USERS']} WHERE user_name = '{$source_row['user_name']}' LIMIT 1");
        if ($target_row = cpg_db_fetch_row($target_result)) {                   // get matching user id from target
          $source_mapping_users[$userid]['target_user_id'] = $target_row['user_id'];
          $source_mapping_users[$userid]['target_user_folder'] = $target_userfolder = $target_row['user_id'] + 10000;
        } else {
          $source_mapping_users[$userid]['target_user_folder'] = $target_userfolder = 0;
        }
      }
      if ($target_userfolder > 0) {
        $target = str_replace("{$SOURCE['compare_userpics']}$userfolder", "{$SOURCE['compare_userpics']}$target_userfolder", $target);
        if ($debug_level >= 3)  { $debug_trace[] = "$debug_text translates to target '$target'"; }
      } else {
        // no match found...
        if ($plugin_gallery_merge_execute_files && !$userpics_error) {
          $userpics_error = TRUE;
          // if execution - and no error flagged yet - flag an error...
          $error_message = $lang_plugin_gallery_merge['e_userpics_mapping_needs_db'];
          gallery_merge_error('files', $process_constant, 'error', $error_message);
        }
        if (!$plugin_gallery_merge_execute_files && !$userpics_warn) {
          // if simulation and no warning yet - issue warning (do not fail sim - as we can execute to this point and restart later)
          $userpics_warn = TRUE;
          $info_message = $lang_plugin_gallery_merge['i_userpics_mapping_needs_db'].'<br>'.$lang_plugin_gallery_merge['i_userpics_mapping_needs_db2'];
          gallery_merge_error('files', $process_constant, 'info', $info_message);
        }
        $col2 = sprintf($lang_plugin_gallery_merge['userpics_nomatch'], "{$source_row['user_name']} ($userid)");
        gallery_merge_output_status('blank', $col2);
        $target = str_replace("{$SOURCE['compare_userpics']}$userfolder", "{$SOURCE['compare_userpics']}unknown", $target);
        $debug_trace[] = "$debug_text - Unable to map to target - set to $target";
      }
      // see if we need to worry about albums... (based on $SOURCE setting - DB merge reports difference if set in SOURCE and not in TARGET)
      if ($SOURCE['upload_create_album_directory'] && ($target_userfolder > 0) && is_numeric($albumfolder)) {
        // config option to createalbum folders... and numeric folder found... translate if we can
        // it is possible album won't exist - pics could have been moved between albums and orig album deleted - which wouldn't change folders in path
        // we won't fail if we can't remap...
        if (key_exists($albumfolder, $source_mapping_albums)) {
          // seen this value before - use it..
          if ($source_mapping_albums[$albumfolder]['target_album_folder'] > 0) {
            $target_albumfolder = $source_mapping_users[$albumfolder]['target_album_folder'];
          } else {
            $target_albumfolder = 0;
          }
        } else {
          // new album - go get translation
          $source_result = cpg_db_query("SELECT user_name FROM {$SOURCE['TABLE_ALBUMS']} WHERE aid = '$albumfolder' LIMIT 1", $SOURCE['LINK_ID']);
          $source_row = cpg_db_fetch_row($source_result);                             // get user name from source
          $debug_text .= "Album Title: '{$source_row['title']}' ";
          cpg_db_free_result($source_result);
          $target_result = cpg_db_query("SELECT user_id FROM {$CONFIG['TABLE_ALBUMS']} WHERE title = '{$source_row['title']}' AND owner = '{$source_mapping_users[$userid]['target_user_id']}' LIMIT 1");
          if ($target_row = cpg_db_fetch_row($target_result)) {                     // get matching user id from target
            $source_mapping_users[$userid]['target_album_folder'] = $target_albumfolder = $target_row['aid'];
          } else {
            $source_mapping_users[$userid]['target_album_folder'] = $target_albumfolder = 0;
          }
        }
        if ($target_albumfolder > 0) {
          $target = str_replace("{$SOURCE['compare_userpics']}$target_userfolder/$albumfolder", "{$SOURCE['compare_userpics']}$target_userfolder/$target_albumfolder", $target);
        }
      } // endif is_numeric (albumfolder)
      if ($debug_level >= 3)  { $debug_trace[] = "$debug_text translates to target '$target'"; }
    // endif is_numeric (userfolder)
    } else {
      // we have a non-numeric folder/file in root userpics... ignore for now
      if ($debug_level >= 3)  { $debug_trace[] = "Userpics destination: Non-numeric folder/file '$userfolder' found in {$CONFIG['userpics']} directory"; }
    }
  }

  // process add request if set and not a userpics destination
  if (($CONFIG['plugin_gallery_merge_add_subfolder'] != '') && ($SOURCE['compare_userpics2'] != substr($target, 0, $SOURCE['compare_userpics2_len']))) {
    $target = str_replace($SOURCE['album_filepath'], $CONFIG['album_filepath_add'], $target);   // replace source filepath with target including additional folder
  } else {
    // default replacement if no folders added or userpics
    $target = str_replace($SOURCE['album_filepath'], $CONFIG['album_filepath'], $target);       // replace source filepath with target
  }

  return $target;
}


// compare source and target album filesystems recursively - report directory conflicts
function gallery_merge_compare_directories($source_album_filepath, &$source_directory_array, &$source_file_array, $bypass_array = array()) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $process_constant;
  // process through $source_album_filepath as deep as necessary to identify all conflicts
  // initial source_album _filepath is in $SOURCE array - but this routine is caled recursively following the directory structure
  // conflicting types (name is directory in one and file in other) flags error. Other conflicts reported only.
  // $source_directory_array starts empty. Will contain all directories/subdirectories searched
  // $source_file_array starts empty. Will contain all files found in common directories that need to be verified for conflicts
  $recursive_directory_array = array();
  gallery_merge_traverse($source_album_filepath, $source_directory_array, $source_file_array, 'all', 'current');
  foreach ($source_directory_array AS $source_directory) {
    if (in_array($source_directory, $bypass_array)) {
      $debug_trace[] = "Compare Directories: Directory $source_directory bypassed - in bypass array";
      continue;                                                                 // move on to next entry
    }
    $target_directory = gallery_merge_set_target_dest_files($source_directory);       // replace source filepath with target
    if ($debug_level >= 3)  {  $debug_trace[] =  "Compare Directories: Source Directory: $source_directory - looking for Target Directory: $target_directory"; }
    if (file_exists($target_directory)) {
      // exists in both source and target
      if (is_dir($target_directory)) {
        $subdirectory = str_replace($CONFIG['album_filepath'], '', $target_directory);
        $col2 = sprintf($lang_plugin_gallery_merge['compare_directory_merge'], $subdirectory);
        gallery_merge_output_status('blank', $col2);
        // a directory in both - continue recursively down the tree
        if ($debug_level >= 3)  {  $debug_trace[] =  "Compare Directories: recursive call using directory: $source_directory"; }
        gallery_merge_compare_directories($source_directory, $recursive_directory_array, $source_file_array);
      } else {
        // directory in source - and file in target... flag error
        $col2 = sprintf($lang_plugin_gallery_merge['compare_dirtype_conflict'], $target_directory);
        gallery_merge_output_status('blank', $col2);
        $error_message = sprintf($lang_plugin_gallery_merge['e_file_dir_mismatch'], $target_directory);
        gallery_merge_error('files', $process_constant, 'error', $error_message);
      }
    } else {
      if ($message_level == 2) {
        $subdirectory = str_replace($SOURCE['album_filepath'], '', $source_directory);
        $col2 = sprintf($lang_plugin_gallery_merge['compare_directory_copy'], $subdirectory);
        gallery_merge_output_status('blank', $col2);
      }
    }
  }
  $source_directory_array = array_merge($source_directory_array, $recursive_directory_array);
}


// compare source and target album filesystems - report file conflicts
function gallery_merge_compare_files($source_file_array) {
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $process_constant;
  // process through $source_file_array to identify all conflicts
  // conflicting types (name is directory in one and file in other) flags error. Other conflicts reported only.
  $bypass_files = array("/{$CONFIG['userpics']}index.php", "/{$CONFIG['userpics']}no_FTP-uploads_into_this_folder!", "/{$CONFIG['userpics']}no_FTP-uploads_into_this_folder!.txt");
  foreach ($source_file_array AS $source_file) {
    $target_file = gallery_merge_set_target_dest_files($source_file);                 // replace source filepath with target
    if (file_exists($target_file)) {
      // exists in both source and target
      if (!is_dir($target_file)) {
        $subfile = str_replace($SOURCE['album_filepath'], '', $source_file);
        if (in_array($subfile, $bypass_files)) {
          if ($debug_level >= 3) { $debug_trace[] =  "$subfile in SOURCE gallery already exists in TARGET gallery albums directory (file part of Coppermine install)."; }
        } else {
          $col2 = sprintf($lang_plugin_gallery_merge['compare_file_conflict'], $subfile);
          gallery_merge_output_status('blank', $col2);
        }
      } else {
        // directory in target - and file in source... flag error
        $col2 = sprintf($lang_plugin_gallery_merge['compare_filetype_conflict'], $source_file);
        gallery_merge_output_status('blank', $col2);
        $error_message = sprintf($lang_plugin_gallery_merge['e_file_dir_mismatch'], $source_file);
        gallery_merge_error('files', $process_constant, 'error', $error_message);
      }
    } else {
      if ($message_level == 2) {
        $subfile = str_replace($SOURCE['album_filepath'], '', $source_file);
        $col2 = sprintf($lang_plugin_gallery_merge['compare_file_copy'], $subfile);
        gallery_merge_output_status('blank', $col2);
      } elseif ($debug_level >= 3) { $debug_trace[] =  "Compare Files: Source File: $source_file - looking for Target File: $target_file"; }
    }
  }
}


function gallery_merge_file_copy($src,$dst, $replace = FALSE) {
  // simple file copy with global counts
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute_files;
  global $process_constant, $quiesce_copy, $quiesce_count, $quiesce_directory;
  if (!$plugin_gallery_merge_execute_files) {
    // safety net...
    $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_copy_invocation'], 'file_copy');
    gallery_merge_error('files', $process_constant, 'error', $error_message);
    return;
  }
  if (($replace == '1') || (!file_exists($dst))) {                              // honor replace option
    copy($src,$dst);
    // Update file based on file permission settings
    //   chmod($dst . '/' . $file, 0775);
    if ($CONFIG['plugin_gallery_merge_file_permissions'] == '2') {
      // use specified value from config
      @chmod($dst, octdec($CONFIG['plugin_gallery_merge_file_perm_value']));    // quietly
    } elseif ($CONFIG['plugin_gallery_merge_file_permissions'] == '1') {
      // use source permissions
      @chmod($dst, fileperms($src));                                            // quietly
    } elseif ($CONFIG['plugin_gallery_merge_file_permissions'] == '0') {
      // use CPG config default
      @chmod($dst, octdec($CONFIG['default_file_mode']));                       // quietly
    }
    $quiesce_count ++;
  }
}


function gallery_merge_recurse_copy($src,$dst, $replace = FALSE) {
  // simple recursive copy with global counts
  global $CONFIG, $SOURCE, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute_files;
  global $process_constant, $quiesce_copy, $quiesce_count, $quiesce_directory;
  if (!$plugin_gallery_merge_execute_files) {
    // safety net...
    $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_copy_invocation'], 'recurse_copy');
    gallery_merge_error('files', $process_constant, 'error', $error_message);
    return;
  }
  $dir = opendir($src);
  if (!is_dir($dst)) {
    // Create directory based on folder permission settings (windows ignores mode in mkdir - so chmod if we can...)
    // set mkdir recursive (, TRUE) to allow nested directory creation - first iteration may have that case if we are adding folders
    if ($CONFIG['plugin_gallery_merge_folder_permissions'] == '2') {
      // use specified value from config
      mkdir($dst, octdec($CONFIG['plugin_gallery_merge_folder_perm_value']), TRUE);
      @chmod($dst, octdec($CONFIG['plugin_gallery_merge_folder_perm_value']));  // quietly...
    } elseif ($CONFIG['plugin_gallery_merge_folder_permissions'] == '1') {
      // use source permissions
      mkdir($dst, fileperms($src), TRUE);
      @chmod($dst, fileperms($src));                                            // quietly
    } elseif ($CONFIG['plugin_gallery_merge_folder_permissions'] == '0') {
      // default... use CPG config option
      mkdir($dst, octdec($CONFIG['default_dir_mode']), TRUE);
      @chmod($dst, octdec($CONFIG['default_dir_mode']));                        // quietly
    } else {
      // just mkdir - no perms
      mkdir($dst, '', TRUE);
    }
  }
  while(false !== ( $file = readdir($dir)) ) {
    if (( $file != '.' ) && ( $file != '..' )) {
      if ( is_dir($src . '/' . $file) ) {
        gallery_merge_recurse_copy($src.'/'.$file,$dst.'/'.$file);
      }
      else {
        if (($replace == '1') || (!file_exists($dst . '/' . $file))) {          // honor replace option
          copy($src . '/' . $file,$dst . '/' . $file);
          // Update file based on file permission settings
          //   chmod($dst . '/' . $file, 0775);
          if ($CONFIG['plugin_gallery_merge_file_permissions'] == '2') {
            // use specified value from config
            @chmod($dst . '/' . $file, octdec($CONFIG['plugin_gallery_merge_file_perm_value']));
          } elseif ($CONFIG['plugin_gallery_merge_file_permissions'] == '1') {
            // use source permissions
            @chmod($dst . '/' . $file, fileperms($src . '/' . $file));
          } elseif ($CONFIG['plugin_gallery_merge_file_permissions'] == '0') {
            // use CPG config default
            @chmod($dst . '/' . $file, octdec($CONFIG['default_file_mode']));
          }
          $quiesce_count ++;
        }
      }
    }
  }
  closedir($dir);
}


function gallery_merge_recurse_copy_chkpt($src, $dst, $replace = FALSE, $restart_flag = FALSE, $bypass_array = array()) {
  // recursive copy with counts and checkpoint/restart capability
  // while $dst is specified, to implement 'remove' and 'add' support - it will be overridden here in all cases
  // checkpoints taken after each subdirectory is completed (end of each recursion level)
  // process quiesces at end of current directory if threshold exceeded (directories can be LARGE - so may be several thousand records past...
  // if $restart_flag = TRUE, compare each directory with $RESTART_DIR array to see if it already completed.
  global $CONFIG, $SOURCE, $RESTART_DIR, $BYPASS_DIR, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $plugin_gallery_merge_execute_files;
  global $process_constant, $quiesce_copy, $quiesce_count, $quiesce_directory;
  // with multiple levels of recursion - a checkpoint will be taken just before return from each level
  // total calls checked after each file copy - we will stop immediately on reaching the quiesce threshold
  if (!$plugin_gallery_merge_execute_files) {
    $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_copy_invocation'], 'recurse_copy_chkpt');
    gallery_merge_error('files', $process_constant, 'error', $error_message);
    return;
  }
  // process 'normal' bypass - mainline code
  if (in_array($src, $bypass_array)) {
    $debug_trace[] = "File Copy: Directory $src bypassed - in bypass array";
    return;
  }
  // process restart
  if ($restart_flag == 'RESTART') {
    if (in_array($src, $RESTART_DIR)) {
      // if $src is VALUE in array - it has already been processed
      $debug_trace[] = "File Copy RESTART: Directory $src bypassed - already processed";
      return;
    }
  }
  // process 'override' bypass
  if (in_array($src, $BYPASS_DIR)) {
      $debug_trace[] = "File Copy BYPASS: Directory $src will be BYPASSED";
      return;
  }
  // if we are still here...
  $debug_trace[] = "File Copy: Directory $src will be processed";

  $dir = opendir($src);
  $dst = gallery_merge_set_target_dest_files($src);                                   // set destination from source based on remove/add folder settings
  if (!is_dir($dst)) {
    // Create directory based on folder permission settings (windows ignores mode in mkdir - so chmod if we can...)
    // set mkdir recursive (, TRUE) to allow nested directory creation - first iteration may have that case if we are adding folders
    if ($CONFIG['plugin_gallery_merge_folder_permissions'] == '2') {
      // use specified value from config
      mkdir($dst, octdec($CONFIG['plugin_gallery_merge_folder_perm_value']), TRUE);
      @chmod($dst, octdec($CONFIG['plugin_gallery_merge_folder_perm_value']));  // quietly
    } elseif ($CONFIG['plugin_gallery_merge_folder_permissions'] == '1') {
      // use source permissions
      mkdir($dst, fileperms($src), TRUE);
      @chmod($dst, fileperms($src));                                            // quietly
    } elseif ($CONFIG['plugin_gallery_merge_folder_permissions'] == '0') {
      // default...
      mkdir($dst, octdec($CONFIG['default_dir_mode']), TRUE);
      @chmod($dst, octdec($CONFIG['default_dir_mode']));                        // quietly
    } else {
      // just mkdir - no perms
      mkdir($dst, '', TRUE);
    }
  }
  while(false !== ( $file = readdir($dir)) ) {
    if (( $file != '.' ) && ( $file != '..' )) {
      if ( is_dir($src . '/' . $file) ) {
        gallery_merge_recurse_copy_chkpt($src.'/'.$file, $dst.'/'.$file, $replace, $restart_flag, $bypass_array);
       if ($quiesce_copy) break;                                                // exit while loop if quiesce is set
      }
      else {
        if (($replace == '1') || (!file_exists($dst . '/' . $file))) {          // honor replace option
          copy($src . '/' . $file,$dst . '/' . $file);
          // Update file based on file permission settings
          if ($CONFIG['plugin_gallery_merge_file_permissions'] == '2') {
            // use specified value from config
            @chmod($dst . '/' . $file, octdec($CONFIG['plugin_gallery_merge_file_perm_value']));
          } elseif ($CONFIG['plugin_gallery_merge_file_permissions'] == '1') {
            // use source permissions
            @chmod($dst . '/' . $file, fileperms($src . '/' . $file));
          } elseif ($CONFIG['plugin_gallery_merge_file_permissions'] == '0') {
            @chmod($dst . '/' . $file, octdec($CONFIG['default_file_mode']));
          }
          $quiesce_count ++;
          if ($quiesce_count >= $CONFIG['plugin_gallery_merge_chkpt_quiesce_files']) {
            // set quiesce processing - limit reached - and not already set
            $debug_trace[] = "Setting QUIESCE Flag - Total count copied: $count_total_copied - Limit: {$CONFIG['plugin_gallery_merge_chkpt_quiesce_files']}";
            $quiesce_copy = TRUE;                                               // save to global var
            $quiesce_directory = $dst.'/'.$file;                                // save to global var
            $chkpt_dir = str_replace($SOURCE['album_filepath'], '', $quiesce_directory);
            $error_message = sprintf($lang_plugin_gallery_merge['e_processing_quiesced_files'], $quiesce_count);
            gallery_merge_error('files', $process_constant, 'error', $error_message, 'nobypass');
            gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:QUIESCE:$quiesce_count:LASTKEY:{$chkpt_dir}");
            $col2 = sprintf($lang_plugin_gallery_merge['recurse_copy_quiesce'], $quiesce_count);
            gallery_merge_output_status('blank', $col2);
            $debug_trace[] = "$process_name copy processing QUIESCED";
            break;
          }
        }
      }
    }
  }
  closedir($dir);
  if (!$quiesce_copy) {
    // take a checkpoint and continue on...
    $chkpt_dir = str_replace($SOURCE['album_filepath'], '', $src);
    gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COPIED:$quiesce_count:LASTKEY:{$chkpt_dir}");
    }
  return;
}


// compare source and target album filesystems - report missing items - used POST merge
function gallery_merge_verify_files($type = 'albums') {
  global $CONFIG, $SOURCE, $BYPASS_DIR, $lang_plugin_gallery_merge, $message_level, $debug_level, $debug_trace;
  global $process_constant;
  // different approach from above... above method is efficient for small intersects... only obtaining detail from common directories
  // everything in source SHOULD now exist in target - arrays would be huge.
  // in this case:
  // - obtain complete directory list from source
  // - for each:
  //   - verify directory exists in target
  //   - obtain file list for this directory
  //   - verify each file exists in target
  // - report only missing files - and set error flag if discrepancy found.
  $compare_directory_array = array();
  $temp_directory_array = array();
  $count_missing_dir = '0';                                                     // initialize missing counter
  $count_existed_dir = '0';                                                     // initialize existed counter
  $count_bypassed_dir = '0';                                                    // initialize bypassed counter
  $count_missing_files = '0';                                                   // initialize missing counter
  $count_existed_files = '0';                                                   // initialize existed counter
  $verify_dir_error = FALSE;                                                    // no error issued
  $verify__file_error = FALSE;                                                  // no error issued
  // set starting directory based on verify type passed
  if ($type == 'userpics') {
    $startdir = $SOURCE['userpics_filepath'];
  } else {
    $startdir = $SOURCE['album_filepath'];
  }
  $debug_trace[] =  "Verify Files: comparing $type source and target files";
  gallery_merge_traverse($startdir, $compare_directory_array, $compare_file_array, 'dir', 'all');
  if ($debug_level >= 3) { $debug_trace[] = '<pre>VERIFY DIRECTORIES '.print_r($compare_directory_array, TRUE).'</pre>'; }
  foreach ($compare_directory_array AS $source_directory) {
    if ($type == 'albums') {
      // extra checks needs for exclusions...
      if (($source_directory == "{$SOURCE['album_filepath']}/edit") ||
          ($source_directory == $SOURCE['compare_userpics2']) || ($SOURCE['compare_userpics'] == substr($source_directory, 0, $SOURCE['compare_userpics_len']))) {
        continue;                                                               // bypass edit and userpics directory and files...
      }
    } elseif ($type == 'all') {
      // extra checks needs for exclusions...
      if ($source_directory == "{$SOURCE['album_filepath']}/edit") {
        continue;                                                               // bypass edit directory and files...
      }
    }
    $target_directory = gallery_merge_set_target_dest_files($source_directory); // set target directory
    if (is_dir($target_directory)) {
      // directory exists in target... get files in this directory (current level only to keep arrays small)
      $count_existed_dir ++;
      $compare_file_array = array();
      gallery_merge_traverse($source_directory, $temp_directory_array, $compare_file_array, 'files', 'current');
      // process through $compare_file_array to identify any missing files
      foreach ($compare_file_array AS $source_file) {
        $target_file = gallery_merge_set_target_dest_files($source_file);       // set target file from SOURCE
        if (file_exists($target_file)) {
          // exists in both source and target
          $count_existed_files ++;
        } else {
          // Oops - missing file...
          $subfile = str_replace($CONFIG['album_filepath'], '', $target_file);
          $col2 = sprintf($lang_plugin_gallery_merge['compare_file_missing'], $subfile);
          gallery_merge_output_status('blank', $col2);
          if (!$verify_file_error) {
            $verify_file_error = TRUE;
            // if no error flagged yet - flag an error...
            $error_message = $lang_plugin_gallery_merge['e_verify_files_missing'];
            gallery_merge_error('files', $process_constant, 'error', $error_message);
          }
          $count_missing_files ++;
        }
      } // end foreach file
    } else {
      // Oops - missing directory...
      $subdir = str_replace($CONFIG['album_filepath'], '', $target_directory);
      if (in_array($source_directory, $BYPASS_DIR)) {
        $col2 = sprintf($lang_plugin_gallery_merge['compare_dir_bypassed'], $subdir);
        gallery_merge_output_status('blank', $col2);
        $count_bypassed_dir ++;
      } else {
        $col2 = sprintf($lang_plugin_gallery_merge['compare_dir_missing'], $subdir);
        gallery_merge_output_status('blank', $col2);
        if (!$verify_dir_error) {
          $verify_dir_error = TRUE;
          // if no error flagged yet - flag an error...
          $error_message = $lang_plugin_gallery_merge['e_verify_directories_missing'];
          gallery_merge_error('files', $process_constant, 'error', $error_message);
        }
        $count_missing_dir ++;
      }
    }
  } // end foreach directory
  unset ($compare_directory_array);
  unset ($compare_file_array);
  $col2 = sprintf($lang_plugin_gallery_merge['verify_directories_counts'], $count_missing_dir, $count_existed_dir).' - '.sprintf($lang_plugin_gallery_merge['verify_files_counts'], $count_missing_files, $count_existed_files);
  if ($count_bypassed_dir > 0) {
    $col2 .= sprintf($lang_plugin_gallery_merge['verify_directories_bypassed'], $count_bypassed_dir);
  }
  gallery_merge_output_status('blank', $col2);
  $debug_trace[] = "$type file verification processing complete";
}
?>