<?php
/**************************************************
  Coppermine 1.5.x Plugin - gallery_merge
**************************************************/

/* page included into admin.php to handle simulate and execution of database backout
*/

// Make sure that this file can't be accessed directly, but only from within the Coppermine user interface
if (!defined('IN_COPPERMINE')) {
    die('Not in Coppermine...');
}

if (!GALLERY_ADMIN_MODE) {
    cpg_die(ERROR, $lang_errors['access_denied'], __FILE__, __LINE__);
}

      // ******************************
      // Backout DB code starts here...
      // ******************************
      if ($plugin_gallery_merge_backout_simulate) {
        $plugin_gallery_merge_modetext = 'simulation';
        starttable('100%', sprintf($lang_plugin_gallery_merge['simulate_results'], $lang_plugin_gallery_merge['plugin_name']), 3, 'cpg_zebra');
      } elseif ($plugin_gallery_merge_backout_execute || $plugin_gallery_merge_backout_emergency) {
        $plugin_gallery_merge_modetext = 'execution';
        starttable('100%', sprintf($lang_plugin_gallery_merge['execute_results'], $lang_plugin_gallery_merge['plugin_name']), 3, 'cpg_zebra');
      }

      // ****************
      // Validate Backout
      // ****************
      $backout_db_timestamp = $superCage->post->getRaw('backout_db_timestamp');       // will be validated in gallery_merge_checkpoint
      if ($plugin_gallery_merge_backout_emergency) {
        // emergency backout - override timestamp from CONFIG
        $backout_db_timestamp = $CONFIG['plugin_gallery_merge_ts_emer_backout'];
      }
      $starttime = gallery_merge_checkpoint('dbbko', 'timestamp', 'backout', $backout_db_timestamp);  // validate request and show we started BACKOUT DB processing

      // ****************
      // Backout Warnings
      // ****************
      $col2 = "{$lang_plugin_gallery_merge['backout_execute_warning']}<br>{$lang_plugin_gallery_merge['backout_execute_warning2'] }";
      gallery_merge_output_status('blank', $col2);
      if ($plugin_gallery_merge_backout_emergency) {
        // emergency backout - additional warning
        $col2 = sprintf($lang_plugin_gallery_merge['backout_emergency_warning'], $backout_db_timestamp);
        gallery_merge_output_status('blank', $col2);
      }

      // *******************
      // Set Gallery Offline
      // *******************
      if (!$plugin_gallery_merge_error) {
        if ($plugin_gallery_merge_backout_execute || $plugin_gallery_merge_backout_emergency) {
          // set offline for execution
          gallery_merge_offline('offline');
        } else {
          // warn if online for simulation
          gallery_merge_offline('warnonline');
        }
      }

      // **********************
      // Process table restores
      // **********************
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        $col1 = $lang_plugin_gallery_merge['processing_restore'];
        $col2 = sprintf($lang_plugin_gallery_merge['processing_restore_comment'], $starttime);
        gallery_merge_output_status($col1, $col2);
        // list of core tables - we must at least have backups for these to proceed with restore
        $core_tables = array('TABLE_FAVPICS', 'TABLE_DICT', 'TABLE_COMMENTS', 'TABLE_CATMAP', 'TABLE_PICTURES', 'TABLE_ALBUMS', 'TABLE_CATEGORIES', 'TABLE_USERS', 'TABLE_USERGROUPS');
        $BACKOUT = array();
        // see what we have for database backups
        $timestamp = str_replace(' ', '_', $starttime);                         // remove spaces in timestamp
        $backup_table_prefix = "{$CONFIG['TABLE_PREFIX']}{$CONFIG['plugin_gallery_merge_backout_prefix']}{$timestamp}_";
        $result = cpg_db_query("SHOW FULL TABLES LIKE '{$backup_table_prefix}%'");
        $rowset = cpg_db_fetch_rowset($result);
        cpg_db_free_result($result);
        foreach ($rowset AS $row) {
          $tablearray = each ($row );                                           // don't know the key - contains DBNAME and LIKE criteria - so use each to extract value
          $tablename = str_replace($backup_table_prefix, '', $tablearray['value']); // extract table name
          $table_constant = 'TABLE_'.strtoupper($tablename);
          $BACKOUT[] = $table_constant;
          if (!$CONFIG[$table_constant]) {
            $CONFIG[$table_constant] = $CONFIG['TABLE_PREFIX'].$tablename;
          }
        }
        // be sure we have backups for all core tables...
        foreach ($core_tables AS $table_constant) {
          if (!in_array($table_constant, $BACKOUT)) {
            // missing backup for core table... not good...
            $col2 = sprintf($lang_plugin_gallery_merge['processing_restore_missing'], $table_constant);
            gallery_merge_output_status('blank', $col2);
            $error_message = $col2;
            gallery_merge_error('init', 'PROCESS_RESTORES', 'error', $error_message);
          }
        }
        // process backouts
        foreach ($BACKOUT AS $table_constant) {
          gallery_merge_restore_table($table_constant, $starttime);
        }
        if (!$plugin_gallery_merge_error) {
          $col2 = $lang_plugin_gallery_merge['processing_restore_complete'];
          gallery_merge_output_status('blank', $col2);
          if ($plugin_gallery_merge_backout_emergency) {
            // emergency backout completed - clear CONFIG setting
            $result = cpg_db_query("UPDATE {$CONFIG['TABLE_CONFIG']} SET `value` = '0000-00-00 00:00:00' WHERE `name` = 'plugin_gallery_merge_ts_emer_backout'");
            $col2 = $lang_plugin_gallery_merge['backout_emergency_complete'];
            gallery_merge_output_status('blank', $col2);
      }
        }
      }

      // ****************************
      // Reset Gallery Offline Status
      // ****************************
      if (!$plugin_gallery_merge_error) {
        if ($plugin_gallery_merge_backout_execute || $plugin_gallery_merge_backout_emergency) {
          // reset galery status to saved value
          gallery_merge_offline('reset');
        }
      } elseif ($plugin_gallery_merge_backout_execute || $plugin_gallery_merge_backout_emergency) {
        // we have an error in execution mode - warn if gallery is offline (it should be...)
        gallery_merge_offline('warnoffline');
      }

      // *************
      // Output Status
      // *************
      ob_start();
      echo <<< EOT
        <tr>
          <td class="tablef" colspan="2" >
EOT;
      if ($plugin_gallery_merge_error) {
        gallery_merge_error_output('error');
      } else {
        gallery_merge_checkpoint('dbbko', 'timestamp', 'end');                  // show we ended DB BACKOUT processing successfully
        gallery_merge_error_output('success');
        if ($plugin_gallery_merge_backout_execute) {
          // no checkpoint for backout sim - running with last execution timestamp - and will hurt later restart processing...
          gallery_merge_checkpoint('db', 'checkpoint', '', "BACKOUT:COMPLETED");
        }
      }
      echo <<< EOT
          </td>
        </tr>
EOT;
      $results_html = ob_get_flush();                                           // flush (send) buffer and retain
      // all done!!!
      endtable();
?>