<?php
/**************************************************
  Coppermine 1.5.x Plugin - gallery_merge
**************************************************/

/* page included into admin.php/merge.php to handle simulate and execution of database merges
*/

// Make sure that this file can't be accessed directly, but only from within the Coppermine user interface
if (!defined('IN_COPPERMINE')) {
    die('Not in Coppermine...');
}

if (!GALLERY_ADMIN_MODE) {
    cpg_die(ERROR, $lang_errors['access_denied'], __FILE__, __LINE__);
}

      // *********************************************
      // Simulate and Execution DB code starts here...
      // *********************************************
      if ($plugin_gallery_merge_simulate) {
        $checkpoint_text = "DB_SIMULATION";
        $plugin_gallery_merge_modetext = 'simulation';
        starttable('100%', sprintf($lang_plugin_gallery_merge['simulate_results'], $lang_plugin_gallery_merge['plugin_name']), 3, 'cpg_zebra');
      } elseif ($plugin_gallery_merge_execute) {
        $checkpoint_text = "DB_EXECUTION";
        $plugin_gallery_merge_modetext = 'execution';
        starttable('100%', sprintf($lang_plugin_gallery_merge['execute_results'], $lang_plugin_gallery_merge['plugin_name']), 3, 'cpg_zebra');
      }

      // ***************************
      // Validate Checkpoint Restart
      // ***************************
      if ($plugin_gallery_merge_restart_simulate || $plugin_gallery_merge_restart_execute) {
        if ($plugin_gallery_merge_restart_execute) {
          $restart_db_timestamp = $superCage->post->getRaw('restart_db_timestamp');       // will be validated in gallery_merge_checkpoint
        } else {
          $restart_db_timestamp = $superCage->post->getRaw('sim_restart_db_timestamp');   // will be validated in gallery_merge_checkpoint
        }
        $starttime = gallery_merge_checkpoint('db', 'timestamp', 'restart', $restart_db_timestamp);  // validate request and show we RESTARTED DB processing
        $restart_flag = 'RESTART';
      } else {
        $starttime = gallery_merge_checkpoint('db', 'timestamp', 'start');                   // show we started DB processing
        $restart_flag = '';
      }

      // *****************************
      // Obtain Source Connection Info
      // *****************************
      if (!$plugin_gallery_merge_error) {
        $SOURCE = array();
        gallery_merge_obtain_source_config();
      }

      if (!$plugin_gallery_merge_error) {
        if ($restart_flag == 'RESTART') {
          $col1 = "{$lang_plugin_gallery_merge["$plugin_gallery_merge_modetext"]} {$lang_plugin_gallery_merge['restarted']}";
        } else {
          $col1 = "{$lang_plugin_gallery_merge["$plugin_gallery_merge_modetext"]} {$lang_plugin_gallery_merge['started']}";
        }
        gallery_merge_output_status($col1, 'blank');
      }

      // ***********************************
      // Validate/Connect to SOURCE Database
      // ***********************************
      if (!$plugin_gallery_merge_error) {
        gallery_merge_connect_source();
      }

      // ********************
      // Initialize Variables
      // ********************
      // Set up all SOURCE table names we may need (adapted from init.inc.php)
      // plugin gives us target names in the $CONFIG array - set the equivalent for source names using the source table prefix
      // CPG database schema found at http://documentation.coppermine-gallery.net/en/dev_database.htm#db_reference%20within_code
      // array already started above
      $SOURCE['TABLE_PICTURES']      = $SOURCE['TABLE_PREFIX'].'pictures';
      $SOURCE['TABLE_ALBUMS']        = $SOURCE['TABLE_PREFIX'].'albums';
      $SOURCE['TABLE_COMMENTS']      = $SOURCE['TABLE_PREFIX'].'comments';
      $SOURCE['TABLE_CATEGORIES']    = $SOURCE['TABLE_PREFIX'].'categories';
      $SOURCE['TABLE_CONFIG']        = $SOURCE['TABLE_PREFIX'].'config';
      $SOURCE['TABLE_USERGROUPS']    = $SOURCE['TABLE_PREFIX'].'usergroups';
      $SOURCE['TABLE_VOTES']         = $SOURCE['TABLE_PREFIX'].'votes';
      $SOURCE['TABLE_USERS']         = $SOURCE['TABLE_PREFIX'].'users';
      $SOURCE['TABLE_BANNED']        = $SOURCE['TABLE_PREFIX'].'banned';
      $SOURCE['TABLE_EXIF']          = $SOURCE['TABLE_PREFIX'].'exif';
      $SOURCE['TABLE_FILETYPES']     = $SOURCE['TABLE_PREFIX'].'filetypes';
      $SOURCE['TABLE_ECARDS']        = $SOURCE['TABLE_PREFIX'].'ecards';
      $SOURCE['TABLE_FAVPICS']       = $SOURCE['TABLE_PREFIX'].'favpics';
      $SOURCE['TABLE_BRIDGE']        = $SOURCE['TABLE_PREFIX'].'bridge';
      $SOURCE['TABLE_VOTE_STATS']    = $SOURCE['TABLE_PREFIX'].'vote_stats';
      $SOURCE['TABLE_HIT_STATS']     = $SOURCE['TABLE_PREFIX'].'hit_stats';
      $SOURCE['TABLE_TEMP_MESSAGES'] = $SOURCE['TABLE_PREFIX'].'temp_messages';
      $SOURCE['TABLE_CATMAP']        = $SOURCE['TABLE_PREFIX'].'categorymap';
      $SOURCE['TABLE_LANGUAGE']      = $SOURCE['TABLE_PREFIX'].'languages';
      $SOURCE['TABLE_DICT']          = $SOURCE['TABLE_PREFIX'].'dict';
      // add table names CPG doesn't have as variables...
      $CONFIG['TABLE_PLUGINS']       = $CONFIG['TABLE_PREFIX'].'plugins';
      $SOURCE['TABLE_PLUGINS']       = $SOURCE['TABLE_PREFIX'].'plugins';
      $CONFIG['TABLE_SESSIONS']      = $CONFIG['TABLE_PREFIX'].'sessions';
      $SOURCE['TABLE_SESSIONS']      = $SOURCE['TABLE_PREFIX'].'sessions';

      // initialize restart array to hold statuses by table - populated during restart for TABLE_ and PROCESS_ constants.
      // any other names to be tracked should be added here with initial value of '0'.
      $RESTART = array();

      // initialize bypass arrays to hold status by table/process
      $BYPASS = array();                                                        // step bypasses - merged with RESTART data later
      $BYPASS_WARNING = array();                                                // warning bypasses
      $BYPASS_ERROR = array();                                                  // error bypasses
      $bypass_type = 'db';
      include "./plugins/gallery_merge/data/bypass_data.php";

      // initialize all arrays and key fields (some initial values set on first pass through... need all arrays ready
      $mapping_arrays = array('plugins', 'usergroups', 'users', 'categories', 'albums', 'pictures');
      $source_mapping_plugins = array();                                // initialize plugins mapping table (unique usage)
      $source_mapping_usergroups = array();                             // initialize usergroups mapping table
      $source_mapping_users = array();                                  // initialize users mapping table
      $source_mapping_categories = array();                             // initialize categories mapping table
      $source_mapping_albums = array();                                 // initialize albums mapping table
      $source_mapping_pictures = array();                               // initialize pictures mapping table
      $source_mapping_usergroups_key = 'group_id';                      // set key field for mapping array generation
      $source_mapping_users_key = 'user_id';                            // set key field for mapping array generation
      $source_mapping_categories_key = 'cid';                           // set key field for mapping array generation
      $source_mapping_albums_key = 'aid';                               // set key field for mapping array generation
      $source_mapping_pictures_key = 'pid';                             // set key field for mapping array generation
      $source_mapping_catmap_key = '*none*';                            // set key field for simulated inserts (no mapping on this table.. and no autoincrement field)
      $source_mapping_comments_key = 'msg_id';                          // set key field for simulated inserts (no mapping on this table..)
      $source_mapping_favpics_key = 'user_id';                          // set key field for simulated inserts (no mapping on this table..)

      // Retrieve SOURCE DB config fields we need  - and save them in $SOURCE
      $result = cpg_db_query("SELECT name, value FROM {$SOURCE['TABLE_CONFIG']}
                              WHERE name IN ('userpics', 'keyword_separator', 'upload_create_album_directory')", $SOURCE['LINK_ID']);
      while ( ($source_row = cpg_db_fetch_row($result)) ) {
        $SOURCE[$source_row['name']] = $source_row['value'];
      } // while
      cpg_db_free_result($result);

      // *************************
      // Warn if Gallery is Online
      // *************************
      if (!$plugin_gallery_merge_error) {
        gallery_merge_offline('warnonline');
      }

      // **************************
      // Process Checkpoint Restart
      // **************************
      if ((!$plugin_gallery_merge_error) && ($restart_flag == 'RESTART')) {
        // if this is restart - reload mapping arrays and set statuses
        gallery_merge_checkpoint_restart_db($restart_db_timestamp);
        gallery_merge_checkpoint('db', 'checkpoint', '', "$checkpoint_text:RESTARTED", $RESTART);
      } elseif ((!$plugin_gallery_merge_error) && (!$restart_flag == 'RESTART')) {
        gallery_merge_checkpoint('db', 'checkpoint', '', "$checkpoint_text:INITIATED");
      }

      if (!$plugin_gallery_merge_error && (!empty($BYPASS) || !empty($BYPASS_WARNING) || !empty($BYPASS_ERROR))) {
        // process bypass requests if any exist...
        $table_constant = 'PROCESS_BYPASS';
        gallery_merge_checkpoint_bypass();
        gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE", $BYPASS);
      }

      // Most of the work happens here... Processing each table and comparing source and target.
      // INSERT SQL generated (and optionally executed) for each item being merged.
      // Mapping tables built along way to allow translations for future inserts - maintaining relationships between items
      // ie: usersgroups table used to map users to groups... categories table used to map other categories (subcats) and albums to categories...

      // ************************
      // Verify Coppermine Levels
      // ************************
      // both source and target galleries should be at same version/maintenance level...
      $table_constant = 'PROCESS_VERSION_CHECK';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        $col1 = $lang_plugin_gallery_merge['processing_version_check'];
        $col2 = $lang_plugin_gallery_merge['processing_version_check_comment'];
        gallery_merge_output_status($col1, $col2);
        gallery_merge_version_check($table_constant, COPPERMINE_VERSION);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
          $col2 = $lang_plugin_gallery_merge['processing_version_check_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // *********************
      // Process table backups
      // *********************
      // yes - we warned the user to back up tables first... but back 'em up anyway if we can...
      $table_constant = 'PROCESS_BACKUP';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        $col1 = $lang_plugin_gallery_merge['processing_backup'];
        $col2 = $lang_plugin_gallery_merge['processing_backup_comment'];
        gallery_merge_output_status($col1, $col2);
        // for each table being UPDATED - process a backup to file containing starttime to match to execution
        gallery_merge_backup_table('TABLE_USERGROUPS', $starttime);
        gallery_merge_backup_table('TABLE_USERS', $starttime);
        gallery_merge_backup_table('TABLE_CATEGORIES', $starttime);
        gallery_merge_backup_table('TABLE_ALBUMS', $starttime);
        gallery_merge_backup_table('TABLE_PICTURES', $starttime);
        gallery_merge_backup_table('TABLE_CATMAP', $starttime);
        gallery_merge_backup_table('TABLE_COMMENTS', $starttime);
        gallery_merge_backup_table('TABLE_DICT', $starttime);
        gallery_merge_backup_table('TABLE_FAVPICS', $starttime);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
          $col2 = $lang_plugin_gallery_merge['processing_backup_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // ********************
      // Process config table
      // ********************
      // query source config table, and compare keys/values to those in target $CONFIG array
      $table_constant = 'TABLE_CONFIG';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        gallery_merge_process_compare_values($table_constant);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ********************
      // Process plugin table
      // ********************
      // query source plugin table, and compare names to those in target table
      $table_constant = 'TABLE_PLUGINS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        gallery_merge_process_plugin($table_constant);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ************************
      // Process usergroups table
      // ************************
      $table_constant = 'TABLE_USERGROUPS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        $orderby = '';                                                          // order by key field is default to assist restart - override here if needed
        $compare_fields = 'group_name';                                         // field(s) - comma separated - to match source to target (identify duplicates/inserts)
        $compare_type = 'UNIQUE';                                               // type of compare: UNIQUE, INSERT, IGNORE, MERGE
        $skip_keys = array('group_id');                                         // set keys to bypass for insert (typically auto-increment fields)
        $obsolete_keys = array();                                               // set keys that are obsolete for this table - will be bypassed if exist (with warning)
        $zero_keys = '';                                                        // set keys to 0 value on insert (typically data not yet available... to be updated later)
        $map_values = array('/group_name');                                     // set mapping values to set fields to build array for later processing
        $skip_keys_difference = array();                                        // set keys to bypass for comparing like rows in source and target (like passwords...)
        $translations = array();                                                // set translations to be done... fieldname => table to use
        gallery_merge_process_table($table_constant, $orderby, $compare_fields, $compare_type, $skip_keys, $obsolete_keys, $zero_keys, $skip_keys_difference, $translations, $map_values, $RESTART[$table_constant]);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // *******************
      // Process users table
      // *******************
      $table_constant = 'TABLE_USERS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        $orderby = '';                                                          // order by key field is default to assist restart - override here if needed
        $compare_fields = 'user_name';                                          // field(s) - comma separated - to match source to target (identify duplicates/inserts)
        $compare_type = 'UNIQUE';                                               // type of compare: UNIQUE, INSERT, IGNORE, MERGE
        $skip_keys = array('user_id');                                          // set keys to bypass for insert (typically auto-increment fields)
        $obsolete_keys = array();                                               // set keys that are obsolete for this table - will be bypassed if exist (with warning)
        $zero_keys = '';                                                        // set keys to 0 value on insert (typically data not yet available... to be updated later)
        $map_values = array('/user_name');                                      // set mapping values to set fields to build array for later processing
        $skip_keys_difference = array('user_id', 'user_group', 'user_lastvisit', 'user_regdate');  // set keys to bypass for comparing like rows in source and target (like passwords...)
        $translations = array('user_group' => 'usergroups.user_group.group_id'); // set translations to be done... fieldname => table.keyvalue.field to use
        gallery_merge_process_table($table_constant, $orderby, $compare_fields, $compare_type, $skip_keys, $obsolete_keys, $zero_keys, $skip_keys_difference, $translations, $map_values, $RESTART[$table_constant]);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ************************
      // Process categories table
      // ************************
      $table_constant = 'TABLE_CATEGORIES';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
                                                                                // order by key field is default to assist restart - override here if needed
        $orderby = 'ORDER BY depth, pos, cid';                                  // to insure parents process before children so we can properly map (and then key field to assist restart)
        $compare_fields = 'name';                                               // field(s) - comma separated - to match source to target (identify duplicates/inserts)
        if ($CONFIG['plugin_gallery_merge_merge_categories'] == '1') {          // type of compare: UNIQUE, INSERT, IGNORE, MERGE - (set in $CONFIG for categories table)
          $compare_type = 'UNIQUE';
        } else {
          $compare_type = 'INSERT';
        }
        $skip_keys = array('cid', 'rgt', 'depth');                              // set keys to bypass for insert (typically auto-increment fields)
        $obsolete_keys = array('subcat_count', 'alb_count', 'pic_count', 'stat_uptodate');  // set keys that are obsolete for this table - will be bypassed if exist (with warning)
        $zero_keys = array('thumb', 'lft');                                     // set keys to 0 value on insert (typically data not yet available... to be updated later)
        $map_values = array('/name', 'pos', 'thumb', 'depth');                  // set mapping values to set fields to build array for later processing
        $skip_keys_difference = array('lft', 'rgt', 'pos');                     // set keys to bypass for comparing like rows in source and target (like passwords...)
        $translations = array('owner_id' => 'users.owner_id.user_id',
                              'depth' => 'categories.parent.subcat_depth',
                              'pos' => 'categories.parent.incr_subcat_pos',
                              'parent' => 'categories.parent.cid');             // set translations to be done... fieldname => table.keyvalue.field to use
        gallery_merge_map_prime_categories($RESTART[$table_constant]);          // initialize array with possible user categories
        gallery_merge_process_table($table_constant, $orderby, $compare_fields, $compare_type, $skip_keys, $obsolete_keys, $zero_keys, $skip_keys_difference, $translations, $map_values, $RESTART[$table_constant]);
        if ($plugin_gallery_merge_execute) {
          gallery_merge_check_rebuild_category_tree();                          // after all inserted, correct any pos entries, and set lft, rgt, depth
        }
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ********************
      // Process albums table
      // ********************
      $table_constant = 'TABLE_ALBUMS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        $orderby = '';                                                          // order by key field is default to assist restart - override here if needed
        $compare_fields = 'title';                                              // field(s) - comma separated - to match source to target (identify duplicates/inserts)
        if ($CONFIG['plugin_gallery_merge_merge_albums'] == '1') {              // type of compare: UNIQUE, INSERT, IGNORE, MERGE - (set in $CONFIG for albums table)
          $compare_type = 'UNIQUE';
        } else {
          $compare_type = 'INSERT';
        }
        $skip_keys = array('aid');                                              // set keys to bypass for insert (typically auto-increment fields)
        $obsolete_keys = array('pic_count', 'last_addition', 'stat_uptodate');  // set keys that are obsolete for this table - will be bypassed if exist (with warning)
        $zero_keys = array('thumb');                                            // set keys to 0 value on insert (typically data not yet available... to be updated later)
        $map_values = array('/title', 'thumb');                                 // set mapping values to set fields to build array for later processing
        $skip_keys_difference = array('aid', 'thumb');                          // set keys to bypass for comparing like rows in source and target (like passwords...)
        $translations = array('visibility' => 'usergroups.visibility.group_id',
                              'owner' => 'users.owner.user_id',
                              'category' => 'categories.category.cid',
                              'pos' => 'categories.category.incr_subalbum_pos'); // set translations to be done... fieldname => table.keyvalue.field to use
        gallery_merge_process_table($table_constant, $orderby, $compare_fields, $compare_type, $skip_keys, $obsolete_keys, $zero_keys, $skip_keys_difference, $translations, $map_values, $RESTART[$table_constant]);
        if ($plugin_gallery_merge_execute) {
          gallery_merge_check_rebuild_album_pos();                              // after all inserted, scan and correct any pos entries to account for already existing albums
        }
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // **********************
      // Process pictures table
      // **********************
      $table_constant = 'TABLE_PICTURES';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        $orderby = '';                                                          // order by key field is default to assist restart - override here if needed
        $compare_fields = 'filename,filepath';                                  // field(s) - comma separated - to match source to target (identify duplicates/inserts)
        $compare_type = 'UNIQUE';                                               // type of compare: UNIQUE, INSERT, IGNORE, MERGE
        $skip_keys = array('pid');                                              // set keys to bypass for insert (typically auto-increment fields)
        $obsolete_keys = array('owner_name');                                   // set keys that are obsolete for this table - will be bypassed if exist (with warning)
        $zero_keys = '';                                                        // set keys to 0 value on insert (typically data not yet available... to be updated later)
        $map_values = array('/filename');                                       // set mapping values to set fields to build array for later processing
        $skip_keys_difference = array('pid', 'total_filesize', 'pwidth', 'pheight', 'hits', 'mtime', 'ctime', 'owner_id', 'pic_rating', 'votes');
                                                                                // set keys to bypass for comparing like rows in source and target (like passwords...)
        $translations = array('aid' => 'albums.aid.aid',
                              'owner_id' => 'users.owner_id.user_id');          // set translations to be done... fieldname => table.keyvalue.field to use
        gallery_merge_map_prime_pictures($RESTART[$table_constant]);            // initialize array with key values we need to track
        gallery_merge_process_table($table_constant, $orderby, $compare_fields, $compare_type, $skip_keys, $obsolete_keys, $zero_keys, $skip_keys_difference, $translations, $map_values, $RESTART[$table_constant]);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ***********************
      // Update thumb references
      // ***********************
      $table_constant = 'PROCESS_THUMBS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        gallery_merge_process_thumbs();                                         // resolve thumb references we can now map
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // *************************
      // Process categorymap table
      // *************************
      $table_constant = 'TABLE_CATMAP';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // no mapping array or key needed
        $orderby = '';                                                          // order by key field is default to assist restart - override here if needed
                                                                                // no key on this table... processed with INSERT IGNORE to handle duplicates
        $compare_fields = 'cid';                                                // field(s) - comma separated - to match source to target (identify duplicates/inserts)
        $compare_type = 'IGNORE';                                               // type of compare: UNIQUE, INSERT, IGNORE, MERGE
        $skip_keys = array();                                                   // set keys to bypass for insert (typically auto-increment fields)
        $obsolete_keys = array();                                               // set keys that are obsolete for this table - will be bypassed if exist (with warning)
        $zero_keys = array();                                                   // set keys to 0 value on insert (typically data not yet available... to be updated later)
        $map_values = array();                                                  // set mapping values to set fields to build array for later processing
        $skip_keys_difference = array();                                        // set keys to bypass for comparing like rows in source and target (like passwords...)
        $translations = array('cid' => 'categories.cid.cid',
                              'group_id' => 'usergroups.group_id.group_id');    // set translations to be done... fieldname => table.keyvalue.field to use
        gallery_merge_process_table($table_constant, $orderby, $compare_fields, $compare_type, $skip_keys, $obsolete_keys, $zero_keys, $skip_keys_difference, $translations, $map_values, $RESTART[$table_constant]);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // **********************
      // Process comments table
      // **********************
      $table_constant = 'TABLE_COMMENTS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // no mapping array needed
        $orderby = '';                                                          // order by key field is default to assist restart - override here if needed
        $compare_fields = 'pid,msg_body';                                       // field(s) - comma separated - to match source to target (identify duplicates/inserts)
        $compare_type = 'UNIQUE';                                               // type of compare: UNIQUE, INSERT, IGNORE, MERGE
        $skip_keys = array('msg_id');                                           // set keys to bypass for insert (typically auto-increment fields)
        $obsolete_keys = array();                                               // set keys that are obsolete for this table - will be bypassed if exist (with warning)
        $zero_keys = array();                                                   // set keys to 0 value on insert (typically data not yet available... to be updated later)
        $map_values = array();                                                  // set mapping values to set fields to build array for later processing
        $skip_keys_difference = array();                                        // set keys to bypass for comparing like rows in source and target (like passwords...)
        $translations = array('pid' => 'pictures.pid.pid',
                              'author_id' => 'users.author_id.user_id');    // set translations to be done... fieldname => table.keyvalue.field to use
        gallery_merge_process_table($table_constant, $orderby, $compare_fields, $compare_type, $skip_keys, $obsolete_keys, $zero_keys, $skip_keys_difference, $translations, $map_values, $RESTART[$table_constant]);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // *********************
      // Process favpics table
      // *********************
      $table_constant = 'TABLE_FAVPICS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // special processing needed - references to pictures (pid) are imbedded in serialized array in table...
        // no mapping array needed
        $orderby = '';                                                          // order by key field is default to assist restart - override here if needed
                                                                                // no key on this table... special processing with to handle duplicates
        $compare_fields = 'user_id';                                            // field(s) - comma separated - to match source to target (identify duplicates/inserts)
        $compare_type = 'MERGE';                                                // type of compare: UNIQUE, INSERT, IGNORE, MERGE
        $skip_keys = array();                                                   // set keys to bypass for insert (typically auto-increment fields)
        $obsolete_keys = array();                                               // set keys that are obsolete for this table - will be bypassed if exist (with warning)
        $zero_keys = array();                                                   // set keys to 0 value on insert (typically data not yet available... to be updated later)
        $map_values = array();                                                  // set mapping values to set fields to build array for later processing
        $skip_keys_difference = array('user_id', 'user_favpics');               // set keys to bypass for comparing like rows in source and target (like passwords...)
        $translations = array('user_id' => 'users.user_id.user_id');            // set translations to be done... fieldname => table.keyvalue.field to use
        gallery_merge_process_table($table_constant, $orderby, $compare_fields, $compare_type, $skip_keys, $obsolete_keys, $zero_keys, $skip_keys_difference, $translations, $map_values, $RESTART[$table_constant]);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ******************
      // Process dict table
      // ******************
      $table_constant = 'TABLE_DICT';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // A list of all keywords that have been applied to files.
        // optional table - action taken only if used... appears to only be updated manually via script - could find no automatic updates...
        // Using this is more efficient that loading the keywords from the pictures table every time it's needed. It can be regenerated using the keyword_create_dict.php script.
        gallery_merge_process_optional($table_constant);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // **************************
      // Process bridge_stats table
      // **************************
      $table_constant = 'TABLE_BRIDGE';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // unsupported table - no action taken... let them know how many rows...
        gallery_merge_process_unsupported($table_constant);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ********************
      // Process ecards table
      // ********************
      $table_constant = 'TABLE_ECARDS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // unsupported table - no action taken... let them know how many rows...
        // - table only used if CONFIG 'Logging and Statistics' option 'Log ecards' is on...
        // - history/log data only
        if ($CONFIG['log_ecards']) {
          gallery_merge_process_unsupported($table_constant);
        } else {
          gallery_merge_process_unused($table_constant);
        }
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ***********************
      // Process hit_stats table
      // ***********************
      $table_constant = 'TABLE_HIT_STATS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // transient table - no action taken... let them know how many rows...
        // - table only used if CONFIG 'Logging and Statistics' option 'Keep detailed hit statistics' is on...
        // - data can be reset from Admin menu 'Overall Stats' (visible only if stats are being kept.)
        if ($CONFIG['hit_details']) {
          gallery_merge_process_transient($table_constant);
        } else {
          gallery_merge_process_unused($table_constant);
        }
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ********************
      // Process banned table
      // ********************
      $table_constant = 'TABLE_BANNED';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // transient table - no action taken... let them know how many rows...
        // data regularly purged at expiration
        gallery_merge_process_transient($table_constant);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // *******************
      // Process votes table
      // *******************
      $table_constant = 'TABLE_VOTES';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // transient table - no action taken... let them know how many rows...
        // - data only retained for $CONFIG['keep_votes_time'] days - default 30 (pic rating itself saved in pictures table)
        gallery_merge_process_transient($table_constant);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ************************
      // Process vote_stats table
      // ************************
      $table_constant = 'TABLE_VOTE_STATS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // unsupported table - no action taken... let them know how many rows...
        // - table only used if CONFIG 'Logging and Statistics' option 'Keep detailed vote statistics' is on...
        if ($CONFIG['vote_details']) {
          gallery_merge_process_unsupported($table_constant);
        } else {
          gallery_merge_process_unused($table_constant);
        }
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ******************
      // Process exif table
      // ******************
      $table_constant = 'TABLE_EXIF';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // transient table - no action taken... let them know how many rows...
        // - table populated as files are viewed with config option turned on. table can be emptied at any time - file data will be re-read on next access.
        // - table only used if CONFIG 'File Settings' option 'Read EXIF data from JPEG files' is on...
        if ($CONFIG['read_exif_data']) {
          gallery_merge_process_transient($table_constant);
        } else {
          gallery_merge_process_unused($table_constant);
        }
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // **********************
      // Process sessions table
      // **********************
      $table_constant = 'TABLE_SESSIONS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // transient table - no action taken... let them know how many rows...
        gallery_merge_process_transient($table_constant);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ***************************
      // Process temp_messages table
      // ***************************
      $table_constant = 'TABLE_TEMP_MESSAGES';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // transient table - no action taken... let them know how many rows...
        gallery_merge_process_transient($table_constant);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ***********************
      // Process filetypes table
      // ***********************
      $table_constant = 'TABLE_FILETYPES';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // static table - no action taken... compare rows to be sure.
        gallery_merge_process_static($table_constant);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ***********************
      // Process languages table
      // ***********************
      $table_constant = 'TABLE_LANGUAGE';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // static table - no action taken... compare rows to be sure.
        gallery_merge_process_static($table_constant);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ***********************************
      // Process any Plugin Tables supported
      // ***********************************
      $table_constant = 'PROCESS_PLUGINS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        $col1 = $lang_plugin_gallery_merge['processing_plugin_tables'];
        $col2 = $lang_plugin_gallery_merge['processing_plugin_tables_comment'];
        gallery_merge_output_status($col1, $col2);
        // for each installed plugin - see if a script exists to support gallery merge..
        // needs to run as mainline code to allow global variables and functions to be defined/included
        foreach ($source_mapping_plugins AS $plugin_folder => $plugin_data) {
          if ($debug_level >= 3) { $debug_trace[] = "Process Plugins: processing {$plugin_data['name']}<br><pre>{$plugin_folder} ".print_r($plugin_data, TRUE).'</pre>'; }
          if (($plugin_data['merge_script'] != '') && ($plugin_data['merge_script'] != 'n/a') && ($plugin_data['merge_script'] != 'needed')) {
            // we have an installed plugin with a merge script provided
            $col2 = sprintf($lang_plugin_gallery_merge['processing_plugin'], $plugin_data['name']);
            gallery_merge_output_status('blank', $col2);
            // include merge script with plugin specific functions and variable definitions
            require "./plugins/gallery_merge/data/{$plugin_data['merge_script']}";
            // include base process merge code
            require "./plugins/gallery_merge/merge_plugin.inc.php";
            $col2 = sprintf($lang_plugin_gallery_merge['processing_plugin_complete'], $plugin_data['name']);
            gallery_merge_output_status('blank', $col2);
            }
        }
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // *********************
      // Look for other tables
      // *********************
      $table_constant = 'PROCESS_UNKNOWN';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($table_constant))) {
        // might be other non-stanard (or new) tables - look and report based on Coppermine table prefix..
        // Obviously no action for those we don't know
        gallery_merge_process_unknown();
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('db', 'checkpoint', '', "$table_constant:COMPLETE");
        }
      }

      // ********************
      // Housekeeping Cleanup
      // ********************
      // done with mapping arrays - free up some memory before final compare
      if (!$plugin_gallery_merge_error) {
        foreach ($mapping_arrays AS $mapped_table) {
          $mapname = "source_mapping_$mapped_table";
          if ($mapname != 'source_mapping_users') {
            // unset all BUT users map - needed for validation/verification below
            unset($$mapname);
          }
        }
      }

      // **************************
      // Compare tables to validate
      // **************************
      // execution mode only... nothing different than above if simulation
      $table_constant = 'PROCESS_COMPARE';
      if ((!$plugin_gallery_merge_error) && ($plugin_gallery_merge_execute)) {
        $col1 = $lang_plugin_gallery_merge['processing_compare_tables'];
        $col2 = $lang_plugin_gallery_merge['processing_compare_tables_comment'];
        gallery_merge_output_status($col1, $col2);
        // for each table updated earlier - and containing a value we can compare... compare to see that all in SOURCE made it to TARGET
        // can process in any order EXCEPT PICTURES must follow USERS to properly validate USERPICS.
        gallery_merge_verify_table('TABLE_USERGROUPS', 'group_name');
        gallery_merge_verify_table('TABLE_USERS', 'user_name');
        gallery_merge_verify_table('TABLE_CATEGORIES', 'name');
        gallery_merge_verify_table('TABLE_ALBUMS', 'title');
        gallery_merge_verify_table('TABLE_PICTURES', 'filename', 'filepath');
        // gallery_merge_verify_table('TABLE_CATMAP');                          // no value to directly compare
        // gallery_merge_verify_table('TABLE_COMMENTS');                        // no value to directly compare
        // gallery_merge_verify_table('TABLE_DICT');                            // no value to directly compare
        // gallery_merge_verify_table('TABLE_FAVPICS');                         // no value to directly compare
        if (!$plugin_gallery_merge_error) {
          $col2 = $lang_plugin_gallery_merge['processing_compare_tables_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // ********************
      // Capture Debug Output
      // ********************
      gallery_merge_debug_sql('# SQL Log file closed', 'close');                // close log file if used...
      $debug_trace[] = 'Peak memory used: '.memory_get_peak_usage();

      // *************
      // Output Status
      // *************
      ob_start();
      echo <<< EOT
        <tr>
          <td class="tablef" colspan="2" >
EOT;
      if ($plugin_gallery_merge_info) {
        gallery_merge_error_output('info');
      }
      if ($plugin_gallery_merge_warning) {
        gallery_merge_error_output('warning');
      }
      if ($plugin_gallery_merge_error) {
        gallery_merge_error_output('error');
      } else {
        if ($plugin_gallery_merge_warning) {
          gallery_merge_checkpoint('db', 'timestamp', 'warn');                  // show we ended DB processing ended with warning
          gallery_merge_error_output('success_warn');
        } else {
          gallery_merge_checkpoint('db', 'timestamp', 'end');                   // show we ended DB processing successfully
          gallery_merge_error_output('success');
        }
        gallery_merge_checkpoint('db', 'checkpoint', '', "$checkpoint_text:COMPLETED");  // show processing complete
      }
      echo <<< EOT
          </td>
        </tr>
EOT;
      $results_html = ob_get_flush();                                           // flush (send) buffer and retain
      // all done!!!

      endtable();
?>