<?php
/**************************************************
  Coppermine 1.5.x Plugin - gallery_merge
**************************************************/

/* page included into admin.php to handle simulate and execution of file merges
*/

// Make sure that this file can't be accessed directly, but only from within the Coppermine user interface
if (!defined('IN_COPPERMINE')) {
    die('Not in Coppermine...');
}

if (!GALLERY_ADMIN_MODE) {
    cpg_die(ERROR, $lang_errors['access_denied'], __FILE__, __LINE__);
}

      // ************************************************
      // Simulate and Execution FILES code starts here...
      // ************************************************
      if ($plugin_gallery_merge_simulate_files) {
        $checkpoint_text = "FILES_SIMULATION";
        $plugin_gallery_merge_modetext = 'simulation';
        starttable('100%', sprintf($lang_plugin_gallery_merge['simulate_results'], $lang_plugin_gallery_merge['plugin_name']), 3, 'cpg_zebra');
      } elseif ($plugin_gallery_merge_execute_files) {
        $checkpoint_text = "FILES_EXECUTION";
        $plugin_gallery_merge_modetext = 'execution';
        starttable('100%', sprintf($lang_plugin_gallery_merge['execute_results'], $lang_plugin_gallery_merge['plugin_name']), 3, 'cpg_zebra');
      }

      // ***************************
      // Validate Checkpoint Restart
      // ***************************
      if ($plugin_gallery_merge_restart_simulate_files || $plugin_gallery_merge_restart_execute_files) {
        $restart_files_timestamp = $superCage->post->getRaw('restart_files_timestamp');       // will be validated in gallery_merge_checkpoint
        $starttime = gallery_merge_checkpoint('files', 'timestamp', 'restart', $restart_files_timestamp);  // validate request and show we RESTARTED FILES processing
        $restart_flag = 'RESTART';
      } else {
        $starttime = gallery_merge_checkpoint('files', 'timestamp', 'start');                   // show we started FILES processing
        $restart_flag = '';
      }

      // *****************************
      // Obtain Source Connection Info
      // *****************************
      if (!$plugin_gallery_merge_error) {
        $SOURCE = array();
        gallery_merge_obtain_source_config();
      }

      if (!$plugin_gallery_merge_error) {
        if ($restart_flag == 'RESTART') {
          $col1 = "{$lang_plugin_gallery_merge["$plugin_gallery_merge_modetext"]} {$lang_plugin_gallery_merge['restarted']}";
        } else {
          $col1 = "{$lang_plugin_gallery_merge["$plugin_gallery_merge_modetext"]} {$lang_plugin_gallery_merge['started']}";
        }
        gallery_merge_output_status($col1, 'blank');
      }

      // ***********************************
      // Validate/Connect to SOURCE Database
      // ***********************************
      if (!$plugin_gallery_merge_error) {
        gallery_merge_connect_source();
      }

      // ********************
      // Initialize Variables
      // ********************
      // Set up all SOURCE table names we may need (adapted from init.inc.php)
      // plugin gives us target names in the $CONFIG array - set the equivalent for source names using the source table prefix
      // CPG database schema found at http://documentation.coppermine-gallery.net/en/dev_database.htm#db_reference%20within_code
      $SOURCE['TABLE_CONFIG'] = $SOURCE['TABLE_PREFIX'].'config';
      $SOURCE['TABLE_USERS'] = $SOURCE['TABLE_PREFIX'].'users';

      // initialize restart array to hold statuses by table - populated during restart for TABLE_ and PROCESS_ constants.
      // any other names to be tracked should be added here with initial value of '0'.
      $RESTART = array();

      // RESTART_DIR array populated during checkpoint restart
      $RESTART_DIR = array();

      // initialize bypass arrays to hold status by table/process
      $BYPASS = array();                                                        // step bypasses - merged with RESTART data later
      $BYPASS_WARNING = array();                                                // warning bypasses
      $BYPASS_ERROR = array();                                                  // error bypasses
      $BYPASS_DIR = array();                                                    // file directory bypasses
      $bypass_type = 'files';
      include "./plugins/gallery_merge/data/bypass_data.php";

      // initialize all arrays and variables
      $source_directory_array = array();
      $source_file_array = array();
      $quiesce_copy = FALSE;                                                    // global var used to terminate copy part way through...
      $quiesce_count = 0;                                                       // global var used to terminate copy part way through...
      $quiesce_directory = '';                                                  // global var will hold last directory processed
      // mapping array's used by userpics processing to translate user and possibly album ids...
      $mapping_arrays = array('users', 'albums');
      $source_mapping_users = array();                                  // initialize users mapping table
      $source_mapping_albums = array();                                 // initialize albums mapping table
      $source_mapping_users_key = 'user_id';                            // set key field for mapping array generation
      $source_mapping_albums_key = 'aid';                               // set key field for mapping array generation
  
      // *************************
      // Warn if Gallery is Online
      // *************************
      // for file merge - not necessary. readme.txt recommends merging files first - and (especially without replace) that merge will have
      // no impact on the current gallery - just adding to albums directory what we will reference later.

      // **************************
      // Process Checkpoint Restart
      // **************************
      if ((!$plugin_gallery_merge_error) && ($restart_flag == 'RESTART')) {
        // if this is restart - re-establish file position and set statuses
        gallery_merge_checkpoint_restart_files($restart_files_timestamp);
        gallery_merge_checkpoint('files', 'checkpoint', '', "$checkpoint_text:RESTARTED", $RESTART);
      } elseif ((!$plugin_gallery_merge_error) && (!$restart_flag == 'RESTART')) {
        gallery_merge_checkpoint('files', 'checkpoint', '', "$checkpoint_text:INITIATED");
      }

      if (!$plugin_gallery_merge_error && (!empty($BYPASS) || !empty($BYPASS_WARNING) || !empty($BYPASS_ERROR) || !empty($BYPASS_DIR))) {
        // process bypass requests if any exist...
        $process_constant = 'PROCESS_BYPASS';
        $bypass_list = gallery_merge_checkpoint_bypass();
        gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COMPLETE", $bypass_list);
      }

      // Different processing here for simulation versus execution
      // Simulation will compare album folders and files - traversing the source directory tree and seeing if the files exist in target.
      //   Conflicts will be reported for resolution

      // Execution will perform mass copy without replace from SOURCE to TARGET
      //   OS copy command will be used for Linux and Windows


      // **************************************
      // Obtain filepaths for Source and Target
      // **************************************
      if (!$plugin_gallery_merge_error) {
        // get albums folder filepath for SOURCE and TARGET...
        // Retrieve SOURCE DB config fields we need  - and save them in $SOURCE
        $result = cpg_db_query("SELECT name, value FROM {$SOURCE['TABLE_CONFIG']}
                                WHERE name IN ('fullpath', 'userpics', 'upload_create_album_directory')", $SOURCE['LINK_ID']);
        while ( ($source_row = cpg_db_fetch_row($result)) ) {
          $SOURCE[$source_row['name']] = $source_row['value'];
        } // while
        cpg_db_free_result($result);
        $SOURCE['language_filepath'] = $CONFIG['plugin_gallery_merge_source_filepath'].'/lang';
        $SOURCE['plugins_filepath'] = $CONFIG['plugin_gallery_merge_source_filepath'].'/plugins';
        $SOURCE['themes_filepath'] = $CONFIG['plugin_gallery_merge_source_filepath'].'/themes';
        $SOURCE['album_filepath'] = $CONFIG['plugin_gallery_merge_source_filepath'].'/'.rtrim($SOURCE['fullpath'], '/');
        $SOURCE['editpics_filepath'] = $CONFIG['plugin_gallery_merge_source_filepath'].'/'.$SOURCE['fullpath'].'edit';
        $SOURCE['userpics_filepath'] = $CONFIG['plugin_gallery_merge_source_filepath'].'/'.$SOURCE['fullpath'].rtrim($CONFIG['userpics'], '/');
        $bypass_array = array($SOURCE['editpics_filepath'], $SOURCE['userpics_filepath']);
        // $CONFIG['cpg_filepath'] set in merge.php
        $CONFIG['language_filepath'] = $CONFIG['cpg_filepath'].'/lang';
        $CONFIG['plugins_filepath'] = $CONFIG['cpg_filepath'].'/plugins';
        $CONFIG['themes_filepath'] = $CONFIG['cpg_filepath'].'/themes';
        // album filepath options set based on whether additional subfolder was requested...
        $CONFIG['album_filepath'] = $CONFIG['cpg_filepath'].'/'.rtrim($CONFIG['fullpath'], '/');
        $CONFIG['album_filepath_add'] = $CONFIG['cpg_filepath'].'/'.$CONFIG['fullpath'].rtrim($CONFIG['plugin_gallery_merge_add_subfolder'], '/');
        $CONFIG['userpics_filepath'] = $CONFIG['cpg_filepath'].'/'.$CONFIG['fullpath'].rtrim($CONFIG['userpics'], '/');
        // compare values needed for processing remove/add folder requests
        $SOURCE['compare_remove'] = "{$SOURCE['album_filepath']}/{$CONFIG['plugin_gallery_merge_remove_subfolder']}";              // substring match for directories beginning with...
        $SOURCE['compare_remove_len'] = strlen($SOURCE['compare_remove']);      // length for compare
        $SOURCE['compare_remove2'] = "{$SOURCE['album_filepath']}/".rtrim($CONFIG['plugin_gallery_merge_remove_subfolder'], '/');  // exact match for directory
        $SOURCE['compare_userpics'] = "{$SOURCE['album_filepath']}/{$SOURCE['userpics']}";  // substring match for directories beginning with...
        $SOURCE['compare_userpics_len'] = strlen($SOURCE['compare_userpics']);  // length for compare
        $SOURCE['compare_userpics2'] = "{$SOURCE['album_filepath']}/".rtrim($SOURCE['userpics'], '/');  // exact match for directory
        $SOURCE['compare_userpics2_len'] = strlen($SOURCE['compare_userpics2']);  // length for compare
        // update directory arrays with filepath
        if ($restart_flag == 'RESTART') {
          array_walk($RESTART_DIR, 'gallery_merge_checkpoint_restart_table_update', $SOURCE['album_filepath']);
        }
        array_walk($BYPASS_DIR, 'gallery_merge_checkpoint_restart_table_update', $SOURCE['album_filepath']);

        if ($debug_level >= 3) { $debug_trace[] = "Source Language Filepath: {$SOURCE['language_filepath']} - Target Language Filepath: {$CONFIG['language_filepath']}";
                                 $debug_trace[] = "Source Plugin Filepath: {$SOURCE['plugins_filepath']} - Target Plugin Filepath: {$CONFIG['plugins_filepath']}";
                                 $debug_trace[] = "Source Theme Filepath: {$SOURCE['themes_filepath']} - Target Theme Filepath: {$CONFIG['themes_filepath']}";
                                 $debug_trace[] = "Source Album Filepath: {$SOURCE['album_filepath']} - Target Album Filepath: {$CONFIG['album_filepath']}"; }
      }

      // ***********************************
      // Process Languages for Missing Files
      // ***********************************
      $process_constant = 'PROCESS_LANGUAGES';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($process_constant))) {
        // compare directories - outputting information on missing target files
        $col1 = $lang_plugin_gallery_merge['compare_languages'];
        $col2 = $lang_plugin_gallery_merge['compare_languages_comment'];
        gallery_merge_output_status($col1, $col2);
        gallery_merge_compare_directory('Language');
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COMPLETE");
          $col2 = $lang_plugin_gallery_merge['compare_languages_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // ***************************************
      // Process Plugins for Missing Directories
      // ***************************************
      $process_constant = 'PROCESS_PLUGINS';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($process_constant))) {
        // compare directories - outputting information on missing target directories
        $col1 = $lang_plugin_gallery_merge['compare_plugins'];
        $col2 = $lang_plugin_gallery_merge['compare_plugins_comment'];
        gallery_merge_output_status($col1, $col2);
        gallery_merge_compare_directory('Plugin');
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COMPLETE");
          $col2 = $lang_plugin_gallery_merge['compare_plugins_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // **************************************
      // Process Themes for Missing Directories
      // **************************************
      $process_constant = 'PROCESS_THEMES';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($process_constant))) {
        // compare directories - outputting information on missing target directories
        $col1 = $lang_plugin_gallery_merge['compare_themes'];
        $col2 = $lang_plugin_gallery_merge['compare_themes_comment'];
        gallery_merge_output_status($col1, $col2);
        gallery_merge_compare_directory('Theme');
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COMPLETE");
          $col2 = $lang_plugin_gallery_merge['compare_themes_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // ***************************************
      // Process Album Directories for Conflicts
      // ***************************************
      $process_constant = 'PROCESS_ALBUM_COMPARE';                              // intentionally same process constant for these steps
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($process_constant))) {
        // compare directories - returning array of directories search, and files found in common drectories
        $col1 = $lang_plugin_gallery_merge['compare_directories'];
        $col2 = $lang_plugin_gallery_merge['compare_directories_comment'];
        gallery_merge_output_status($col1, $col2);
        if ($CONFIG['plugin_gallery_merge_remove_subfolder'] != '') {
          $col2 = sprintf($lang_plugin_gallery_merge['compare_directories_remove'], $CONFIG['plugin_gallery_merge_remove_subfolder']);
          gallery_merge_output_status('blank', $col2);
        }
        if ($CONFIG['plugin_gallery_merge_add_subfolder'] != '') {
          $col2 = sprintf($lang_plugin_gallery_merge['compare_directories_add'], $CONFIG['plugin_gallery_merge_add_subfolder']);
          gallery_merge_output_status('blank', $col2);
        }
        gallery_merge_compare_directories($SOURCE['album_filepath'], $source_directory_array, $source_file_array, $bypass_array);
        if (!$plugin_gallery_merge_error) {
          $col2 = $lang_plugin_gallery_merge['compare_directories_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // *********************************
      // Process Album Files for Conflicts
      // *********************************
      $process_constant = 'PROCESS_ALBUM_COMPARE';                              // intentionally same process constant for these steps
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($process_constant))) {
        // compare files existing in common directories to identify conflicts
        $col1 = $lang_plugin_gallery_merge['compare_files'];
        $col2 = $lang_plugin_gallery_merge['compare_files_comment'];
        gallery_merge_output_status($col1, $col2);
        gallery_merge_compare_files($source_file_array);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COMPLETE");
          $col2 = $lang_plugin_gallery_merge['compare_files_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // *******************
      // Execute ALBUM Copy!
      // *******************
      $process_constant = 'PROCESS_ALBUM_COPY';
      if ((!$plugin_gallery_merge_error) && ($plugin_gallery_merge_execute_files) && (gallery_merge_checkpoint_restart_table($process_constant))) {
        $col1 = $lang_plugin_gallery_merge['copy_files'];
        $col2 = $lang_plugin_gallery_merge['copy_files_comment'].'<br>'.$lang_plugin_gallery_merge['copy_files_comment2'];
        gallery_merge_output_status($col1, $col2);
        $col2 = sprintf($lang_plugin_gallery_merge['copy_files_detail'], $SOURCE['album_filepath'], $CONFIG['album_filepath']);
        gallery_merge_output_status('blank', $col2);
        $debug_trace[] = "Albums File Copy initiated.";
        // process all folders except edit and userpics (in bypass array)
        gallery_merge_recurse_copy_chkpt($SOURCE['album_filepath'], $CONFIG['album_filepath'], $CONFIG['plugin_gallery_merge_replace_files'], $restart_flag, $bypass_array);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COMPLETE");
          $col2 = sprintf($lang_plugin_gallery_merge['copy_files_complete'], $quiesce_count);
          gallery_merge_output_status('blank', $col2);
        }
        $debug_trace[] = "Albums File Copy completed.";
      }

      // *********************************************
      // Compare Directories/Files to Validate Results
      // *********************************************
      $process_constant = 'PROCESS_VALIDATE_ALBUMS';
      if ((!$plugin_gallery_merge_error) && ($plugin_gallery_merge_execute_files) && (gallery_merge_checkpoint_restart_table($process_constant))) {
        // execute only...
        // compare directories/files - report exceptions only
        $col1 = $lang_plugin_gallery_merge['verify_files'];
        $col2 = $lang_plugin_gallery_merge['verify_files_comment'];
        gallery_merge_output_status($col1, $col2);
        gallery_merge_verify_files();
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COMPLETE");
          $col2 = $lang_plugin_gallery_merge['verify_files_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // *********************
      // Execute USERPICS Copy
      // *********************
      $process_constant = 'PROCESS_USERPICS_COPY';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($process_constant))) {
        $col1 = $lang_plugin_gallery_merge['copy_userpics'];
        $col2 = $lang_plugin_gallery_merge['copy_userpics_comment'];
        gallery_merge_output_status($col1, $col2);
        $col2 = sprintf($lang_plugin_gallery_merge['copy_userpics_detail'], $SOURCE['userpics_filepath'],$CONFIG['userpics_filepath']);
        gallery_merge_output_status('blank', $col2);
        $debug_trace[] = "Userpics File Compare initiated.";
        // first see if we have all info needed to copy...
        gallery_merge_compare_directories($SOURCE['userpics_filepath'], $source_userpics_directory_array, $source_userpics_file_array);
        gallery_merge_compare_files($source_userpics_file_array);
        // proceed with copy if info is available (error would be set if not...) and this is execution mode...
        if ((!$plugin_gallery_merge_error) && ($plugin_gallery_merge_execute_files)) {
          $debug_trace[] = "Userpics File Copy initiated.";
          gallery_merge_recurse_copy_chkpt($SOURCE['userpics_filepath'], $CONFIG['userpics_filepath'], $CONFIG['plugin_gallery_merge_replace_files'], $restart_flag);
        }
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COMPLETE");
          $col2 = sprintf($lang_plugin_gallery_merge['copy_userpics_complete'], $quiesce_count);
          gallery_merge_output_status('blank', $col2);
        }
        $debug_trace[] = "Userpics File Copy completed.";
      }

      // *********************************************
      // Compare Directories/Files to Validate Results
      // *********************************************
      $process_constant = 'PROCESS_VALIDATE_USERPICS';
      if ((!$plugin_gallery_merge_error) && ($plugin_gallery_merge_execute_files) && (gallery_merge_checkpoint_restart_table($process_constant))) {
        // execute only...
        // compare directories/files - report exceptions only
        $col1 = $lang_plugin_gallery_merge['verify_userpics'];
        $col2 = $lang_plugin_gallery_merge['verify_userpics_comment'];
        gallery_merge_output_status($col1, $col2);
        gallery_merge_verify_files('userpics');
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COMPLETE");
          $col2 = $lang_plugin_gallery_merge['verify_userpics_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // *********************************************
      // Process Remaining CPG Files for Discrepancies
      // *********************************************
      $process_constant = 'PROCESS_COMPARE_CPG';
      if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($process_constant))) {
        // compare directories - outputting information on missing target directories
        $col1 = $lang_plugin_gallery_merge['compare_cpg'];
        $col2 = $lang_plugin_gallery_merge['compare_cpg_comment'];
        gallery_merge_output_status($col1, $col2);
        $col2 = $lang_plugin_gallery_merge['compare_cpg_comment2'];
        gallery_merge_output_status('blank', $col2);
        gallery_merge_compare_cpg($CONFIG['plugin_gallery_merge_source_filepath']);
        if (!$plugin_gallery_merge_error) {
          gallery_merge_checkpoint('files', 'checkpoint', '', "$process_constant:COMPLETE");
          $col2 = $lang_plugin_gallery_merge['compare_cpg_complete'];
          gallery_merge_output_status('blank', $col2);
        }
      }

      // ********************
      // Capture Debug Output
      // ********************
      if ($debug_level >= 3) {
        if ($restart_flag == 'RESTART') {
          $debug_trace[] = '<pre>RESTART '.print_r($RESTART, TRUE).'</pre>';
          $debug_trace[] = '<pre>RESTART DIRECTORY '.print_r($RESTART_DIR, TRUE).'</pre>';
        }
        $debug_trace[] = '<pre>SOURCE DIRECTORIES (to point of uniqueness) '.print_r($source_directory_array, TRUE).'</pre>';
        $debug_trace[] = '<pre>FILES in COMMON DIRECTORIES '.print_r($source_file_array, TRUE).'</pre>';
        $debug_trace[] = 'Peak memory used: '.memory_get_peak_usage();
      }

      // *************
      // Output Status
      // *************
      ob_start();
      echo <<< EOT
        <tr>
          <td class="tablef" colspan="2" >
EOT;
      if ($plugin_gallery_merge_info) {
        gallery_merge_error_output('info');
      }
      if ($plugin_gallery_merge_error) {
        gallery_merge_error_output('error');
      } else {
        gallery_merge_checkpoint('files', 'timestamp', 'end');                  // show we ended FILE processing successfully
        gallery_merge_error_output('success');
        gallery_merge_checkpoint('files', 'checkpoint', '', "$checkpoint_text:COMPLETED");        // show processing complete
      }
      echo <<< EOT
          </td>
        </tr>
EOT;
      $results_html = ob_get_flush();                                           // flush (send) buffer and retain
      // all done!!!
      endtable();
?>