<?php
/**************************************************
  Coppermine 1.5.x Plugin - gallery_merge
  Plugin Support for {plugin_name}
**************************************************/

/*
   page included into merge_db.php after merge_plugin_{plugin_folder}.php (array and functions) is included for each plugin
   processes the tables defined to handle simulate and execution of merges for a specific plugin
   variables should be all set - we just need to execute
   We are running in mainline - all global variables available to us
*/

// Make sure that this file can't be accessed directly, but only from within the Coppermine user interface
if (!defined('IN_COPPERMINE')) {
    die('Not in Coppermine...');
}

if (!GALLERY_ADMIN_MODE) {
    cpg_die(ERROR, $lang_errors['access_denied'], __FILE__, __LINE__);
}

      // ********************
      // Initialize Variables
      // ********************

      // validate plugin folder with plugin_array
      if (array_key_exists($plugin_folder, $plugin_array)) {
        foreach ($plugin_table_array AS $plugin_table => $plugin_table_data) {
          $plugin_table_uc = strtoupper($plugin_table);
          $plugin_table_constant = "TABLE_$plugin_table_uc";
          // add table constant to CONFIG array if not already there (Plugin may have defined it's own tables)
          if (!$CONFIG[$plugin_table_constant]) {
            $CONFIG[$plugin_table_constant] = $CONFIG['TABLE_PREFIX'].$plugin_table;
          }
          // add table constant to SOURCE array
          $SOURCE[$plugin_table_constant]   = $SOURCE['TABLE_PREFIX'].$plugin_table;
          // determine process type

          if ($plugin_table_data['process_type'] == 'process_table') {
            // use gallery_merge_process_table routine... extract additional parms
            if (!empty($plugin_table_data['map_values'])) {
              // mapping requested construct map array name
              $map_array = "source_mapping_{$plugin_table}";                    // construct array name
              $mapping_arrays[] = $plugin_table;                                // save to our list of arrays
              if (!is_array($$map_array)) {
                // array doesn't exist yet... (on checkpoint restart, could already be reloaded from prior run..)
                $$map_array = array();
              }
            } // end mapping requested
            // set key field for mapping array generation and simulated insert support
            $map_array_key = "source_mapping_{$plugin_table}_key";
            $$map_array_key = $plugin_table_data['table_key'];

            // main processing for table.. backup comes first
            $plugin_process_constant = "PROCESS_PLUGIN_{$plugin_table_constant}_BACKUP";
            // checkpoint restart point...
            if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($plugin_process_constant))) {
              // if no error - and restart says proceed...
                $col1 = sprintf($lang_plugin_gallery_merge['processing_plugins_backup'], $plugin_table);
                $col2 = $lang_plugin_gallery_merge['processing_plugins_backup_comment'];
                gallery_merge_output_status($col1, $col2);
                gallery_merge_backup_table($plugin_table_constant, $starttime);
                if (!$plugin_gallery_merge_error) {
                  gallery_merge_checkpoint('db', 'checkpoint', '', "$plugin_process_constant:COMPLETE");
                  $col2 = sprintf($lang_plugin_gallery_merge['processing_plugins_backup_completed'], $plugin_table_uc);
                  gallery_merge_output_status('blank', $col2);
                }
            }

            // checkpoint restart point...
            if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($plugin_table_constant))) {
              // see if pre merge function call is requested
              if ($plugin_table_data['pre_merge_function'] != '') {
                $plugin_table_data['pre_merge_function']();                     // call the specified function
              }
              if (!$plugin_gallery_merge_error) {
                // main merge process
                gallery_merge_process_table($plugin_table_constant, $plugin_table_data['orderby'], $plugin_table_data['compare_fields'], $plugin_table_data['compare_type'], $plugin_table_data['skip_keys'], $plugin_table_data['obsolete_keys'], $plugin_table_data['zero_keys'], $plugin_table_data['skip_keys_difference'], $plugin_table_data['translations'], $plugin_table_data['map_values'], $RESTART[$plugin_table_constant]);
              }
              // see if post merge function call is requested
              if ((!$plugin_gallery_merge_error) && ($plugin_table_data['post_merge_function'] != '')) {
                $plugin_table_data['post_merge_function']();                    // call the specified function
              }
              if (!$plugin_gallery_merge_error) {
                gallery_merge_checkpoint('db', 'checkpoint', '', "$plugin_table_constant:COMPLETE");
              }
            }
          } // endif 'process_table'


          elseif ($plugin_table_data['process_type'] == 'compare_values') {
            if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($plugin_table_constant))) {
              // compare values only within this table
              gallery_merge_process_compare_values($plugin_table_constant, $plugin_table_data['compare_name_column'], $plugin_table_data['compare_value_column']);
              if (!$plugin_gallery_merge_error) {
                gallery_merge_checkpoint('db', 'checkpoint', '', "$plugin_table_constant:COMPLETE");
              }
            }
          } // endif 'compare_values'


          elseif ($plugin_table_data['process_type'] == 'transient') {
            if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($plugin_table_constant))) {
              // transient table - no action taken... let them know how many rows...
              gallery_merge_process_transient($plugin_table_constant);
              if (!$plugin_gallery_merge_error) {
                gallery_merge_checkpoint('db', 'checkpoint', '', "$plugin_table_constant:COMPLETE");
              }
            }
          } // endif 'transient'


          elseif ($plugin_table_data['process_type'] == 'static') {
            if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($plugin_table_constant))) {
              // static table - no action taken... compare rows to be sure.
              gallery_merge_process_static($plugin_table_constant);
              if (!$plugin_gallery_merge_error) {
                gallery_merge_checkpoint('db', 'checkpoint', '', "$plugin_table_constant:COMPLETE");
              }
            }
          } // endif 'static'


          elseif ($plugin_table_data['process_type'] == 'unsupported') {
            if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($plugin_table_constant))) {
              // unsupported table - no action taken... let them know how many rows...
              gallery_merge_process_unsupported($plugin_table_constant);
              if (!$plugin_gallery_merge_error) {
                gallery_merge_checkpoint('db', 'checkpoint', '', "$plugin_table_constant:COMPLETE");
              }
            }
          } // endif 'static'


          if ($plugin_table_data['process_type'] == 'custom') {
            // use specified custom function...
            // still set up mapping arrays if requested - and BACKUP
            // extract additional parms
            if (!empty($plugin_table_data['map_values'])) {
              // mapping requested construct map array name
              $map_array = "source_mapping_{$plugin_table}";
              if (!is_array($$map_array)) {
                // array doesn't exist yet... (on checkpoint restart, could already be reloaded from prior run..)
                $$map_array = array();
              // set key field for mapping array generation
              $map_array_key = "source_mapping_{$plugin_table}_key";
              $$map_array_key = $plugin_table_data['table_key'];
              }
            } // end mapping requested

            // main processing for table.. backup comes first
            $plugin_process_constant = "PROCESS_{$plugin_table_constant}_BACKUP";
            // checkpoint restart point...
            if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($plugin_process_constant))) {
              // if no error - and restart says proceed...
                gallery_merge_backup_table('$plugin_table_constant', $starttime);
                if (!$plugin_gallery_merge_error) {
                  gallery_merge_checkpoint('db', 'checkpoint', '', "$plugin_process_constant:COMPLETE");
                  $col2 = sprintf($lang_plugin_gallery_merge['processing_plugins_backup_completed'], $plugin_table_uc);
                  gallery_merge_output_status('blank', $col2);
                }
            }

            // checkpoint restart point...
            if ((!$plugin_gallery_merge_error) && (gallery_merge_checkpoint_restart_table($plugin_table_constant))) {
              // see if additional function call(s) is/are requested - it should be - as custom was requested
              if (($plugin_table_data['pre_merge_function'] != '') || ($plugin_table_data['post_merge_function'] != '')) {
                if ($plugin_table_data['pre_merge_function'] != '') {
                  $plugin_table_data['pre_merge_function']();                   // call the specified function
                }
                if ($plugin_table_data['post_merge_function'] != '') {
                  $plugin_table_data['post_merge_function']();                  // call the specified function
                }
              } else {
                $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_plugin_custom'], $plugin_table_constant);
                gallery_merge_error('db', $plugin_table_constant, 'error', $error_message);
              }
              if (!$plugin_gallery_merge_error) {
                gallery_merge_checkpoint('db', 'checkpoint', '', "$plugin_table_constant:COMPLETE");
              }
            }
          } // endif 'custom'

        } // end foreach plugin table
      // endif plugin in array
      } else {
        // not found in array
        $error_message = sprintf($lang_plugin_gallery_merge['e_invalid_plugin_detail'], $plugin_folder);
        gallery_merge_error('db', $table_constant, 'error', $error_message);
        $debug_trace[] = "Plugin $plugin_folder not found in detail array";
      }
?>